package com.otk.application;

import java.io.IOException;
import java.io.Serializable;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import com.otk.application.Configuration.GlobalConfiguration;
import com.otk.application.error.UnexpectedError;
import com.otk.application.util.AbstractLazilyInitialized;
import com.otk.application.util.MiscUtils;
import com.otk.application.util.TextUtils;

public abstract class AbstractAnalytics extends AbstractLazilyInitialized {

	protected AbstractApplication application;
	protected Thread regularSender;
	protected List<Tracking> trackings = Collections.synchronizedList(new ArrayList<Tracking>());

	public abstract String getDeliveryUrl();

	public AbstractAnalytics(AbstractApplication application) {
		this.application = application;
	}

	@Override
	protected boolean initialize() {
		startRegularSender();
		return true;
	}

	@Override
	protected boolean shutDown() {
		stopRegularSender();
		try {
			sendAllTrackings();
		} catch (Throwable t) {
			getApplication().getLog().error(t);
		}
		return true;
	}

	public AbstractApplication getApplication() {
		return application;
	}

	protected void startRegularSender() {
		regularSender = new Thread(MiscUtils.formatThreadName(getClass(), "RegularSender")) {
			@Override
			public synchronized void run() {
				while (true) {
					MiscUtils.sleepSafely(getApplication().getConfiguration()
							.getCategory(GlobalConfiguration.class).trackingsDeliveryPeriodSeconds * 1000);
					if (isInterrupted()) {
						break;
					}
					sendAllTrackings();
				}
			}
		};
		regularSender.setDaemon(true);
		regularSender.setPriority(Thread.MIN_PRIORITY);
		regularSender.start();
	}

	protected void stopRegularSender() {
		if (regularSender != null) {
			regularSender.interrupt();
			try {
				regularSender.join();
			} catch (InterruptedException e) {
				throw new UnexpectedError(e);
			}
		}
	}

	public void track(String event, String... details) {
		if (!ensureIsInitialized()) {
			return;
		}
		int maximumStoredTrackingCount = getApplication().getConfiguration()
				.getCategory(GlobalConfiguration.class).maximumStoredTrackingCount;
		if (trackings.size() >= (maximumStoredTrackingCount - 1)) {
			if (getApplication().getConfiguration().getCategory(GlobalConfiguration.class).logVerbose) {
				getApplication().getLog().info("Skipping tracking: maximum stored tracking count reached");
			}
			if (trackings.size() == (maximumStoredTrackingCount - 1)) {
				trackings.add(new Tracking(new Date(), "MaximumTrackingCountReached", new String[0]));
			}
			return;
		}
		for (int i = 0; i < details.length; i++) {
			details[i] = TextUtils.escapeNewLines(details[i]);
		}
		if (getApplication().getConfiguration().getCategory(GlobalConfiguration.class).logVerbose) {
			getApplication().getLog().info("Tracking: " + event);
		}
		trackings.add(new Tracking(new Date(), event, details));
	}

	public void sendAllTrackings() {
		if (getApplication().getConfiguration().getCategory(GlobalConfiguration.class).logVerbose) {
			getApplication().getLog().info("Delivering trackings");
		}
		try {
			for (Tracking tracking : trackings) {
				sendTracking(tracking.getDateTime(), tracking.getUsed(), tracking.getDetails());
			}
		} catch (Exception e) {
			getApplication().getLog().error(e);
		} finally {
			trackings.clear();
		}
	}

	public void sendTracking(Date when, String used, String... details) {
		if (!getApplication().getConfiguration().getCategory(GlobalConfiguration.class).usageTrackingEnabled) {
			return;
		}
		if (getDeliveryUrl() == null) {
			return;
		}
		used = anonymize(used);
		for (int i = 0; i < details.length; i++) {
			details[i] = anonymize(details[i]);
		}
		Map<String, String> arguments = new HashMap<String, String>();
		StringBuilder categoryId = new StringBuilder();
		{
			categoryId.append("PhoyoId");
			categoryId.append("-");
			categoryId.append("b" + getApplication().getBuildProperties().getBuildNumber());
			arguments.put("ec", categoryId.toString());
		}
		arguments.put("ea", used);
		arguments.put("qt", Long.toString((new Date().getTime() - when.getTime())));
		if (details.length > 0) {
			arguments.put("el", TextUtils.stringJoin(Arrays.asList(details), " - "));
		}
		try {
			arguments.put("cid", MiscUtils.hexEncode(System.getProperty("user.name")));
		} catch (Exception e) {
			arguments.put("cid", "UnknownHost");
		}
		try {
			MiscUtils.sendHttpPost(new URL(getDeliveryUrl()), arguments);
		} catch (IOException e) {
			throw new UnexpectedError(e);
		}
	}

	public String anonymize(String message) {
		message = message.replaceAll("([Pp]assword)=.*", "$1=**********");
		message = message.replaceAll("([Ll]ogin)=.*", "$1=**********");
		message = message.replaceAll("(user\\..*)=.*", "$1=**********");
		return message;
	}

	public static class Tracking implements Serializable {

		private static final long serialVersionUID = 1L;

		private long id;

		private Date dateTime;

		private String used;

		private String[] details;

		public Tracking() {
		}

		public Tracking(Date dateTime, String used, String[] details) {
			super();
			this.dateTime = dateTime;
			this.used = used;
			this.details = details;
		}

		public long getId() {
			return id;
		}

		public void setId(long id) {
			this.id = id;
		}

		public Date getDateTime() {
			return dateTime;
		}

		public void setDateTime(Date dateTime) {
			this.dateTime = dateTime;
		}

		public String getUsed() {
			return used;
		}

		public void setUsed(String used) {
			this.used = used;
		}

		public String[] getDetails() {
			return details;
		}

		public void setDetails(String[] details) {
			this.details = details;
		}

		@Override
		public String toString() {
			return "Tracking [dateTime=" + dateTime + ", used=" + used + ", details=" + Arrays.toString(details) + "]";
		}

	}

}