package com.otk.application;

import java.awt.Image;
import java.io.File;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.net.URL;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Locale;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.SwingUtilities;

import com.otk.application.Configuration.GlobalConfiguration;
import com.otk.application.error.AbstractApplicationError;
import com.otk.application.error.ErrorManager;
import com.otk.application.error.StandardError;
import com.otk.application.translation.T;
import com.otk.application.translation.TranslationHandler;
import com.otk.application.util.ImageUtils;
import com.otk.application.util.TextUtils;
import com.otk.application.window.ApplicationFrame;

public abstract class AbstractApplication {

	public static AbstractApplication createTestApplication() {
		return new AbstractApplication(new String[0]) {

			@Override
			public String getName() {
				return "Test App";
			}

			@Override
			public Image getIconImage() {
				return ImageUtils.NULL_IMAGE;
			}

			@Override
			protected AbstractAnalytics createAnalytics() {
				return new AbstractAnalytics(this) {

					@Override
					public String getDeliveryUrl() {
						return null;
					}
				};
			}

			@Override
			protected AbstractBuildProperties createBuildProperties() {
				return new AbstractBuildProperties() {

					@Override
					protected URL getURL() {
						return null;
					}
				};
			}


		};
	}

	public static void main(String[] args) {
		createTestApplication().launch();
	}

	private static final String RUN_CLASS_COMMAND_LINE_OPTION = "--run-class";
	private static final String OVERRIDE_CONFIG_COMMAND_LINE_OPTION = "--override-config";
	private static final String DEFINE_SYSTEM_PROPERTY_COMMAND_LINE_OPTION = "--define";
	private static final boolean IN_DEBUG_MODE = Boolean
			.valueOf(System.getProperty(AbstractApplication.class.getPackage().getName() + ".debug", "false"));

	private boolean active = false;
	private String[] args;
	private AbstractBuildProperties buildProperties;
	private ErrorManager errorManager;
	private JFrame mainWindow;
	private AbstractAnalytics analytics;
	private Configuration configuration;
	private Log log;
	private TranslationHandler translationHandler;
	
	public AbstractApplication(String[] args) {
		this.args = args;
		setUp();
	}

	@Override
	public void finalize() {
		try {
			cleanUp();
		} catch (Throwable t) {
			try {
				getLog().error(t);
			} catch (Throwable ignore) {
			}
		}
	}

	public boolean isActive() {
		return active;
	}

	public boolean isInDebugMode() {
		return IN_DEBUG_MODE;
	}

	public abstract String getName();

	public abstract Image getIconImage();

	protected abstract AbstractAnalytics createAnalytics();

	protected abstract AbstractBuildProperties createBuildProperties();

	public AbstractBuildProperties getBuildProperties() {
		return buildProperties;
	}

	
	public TranslationHandler getTranslationHandler() {
		return translationHandler;
	}

	public Configuration getConfiguration() {
		return configuration;
	}

	
	public AbstractAnalytics getAnalytics() {
		return analytics;
	}

	public ErrorManager getErrorManager() {
		return errorManager;
	}

	public Log getLog() {
		return log;
	}

	protected TranslationHandler createTranslationHandler() {
		return new TranslationHandler(this);
	}

	protected Log createLog() {
		return new Log(this);
	}

	protected Configuration createConfiguration() {
		return new Configuration(this);
	}

	protected ErrorManager createErrorManager() {
		return new ErrorManager(this);
	}

	protected JFrame createMainWindow() {
		return new ApplicationFrame(this);
	}

	public void resetMainWindow() {
		boolean wasVisible;
		if (mainWindow == null) {
			wasVisible = false;
		} else {
			wasVisible = mainWindow.isVisible();
			mainWindow.dispose();
		}
		mainWindow = createMainWindow();
		if (wasVisible) {
			mainWindow.setVisible(true);
		}
	}

	public synchronized JFrame getMainWindow() {
		return mainWindow;
	}

	protected void checkCompatibility() {
		String compatibleOSNamesString = getCompatibleOSNames();
		if (compatibleOSNamesString == null) {
			return;
		}
		List<String> compatibleOSNames = new ArrayList<String>();
		for (String osName : compatibleOSNamesString.split(",")) {
			osName = osName.trim();
			if (osName.length() == 0) {
				continue;
			}
			compatibleOSNames.add(osName);
		}
		if (compatibleOSNames.size() == 0) {
			return;
		}
		String currentOSName = System.getProperty("os.name");
		if (currentOSName != null) {
			if (!compatibleOSNames.contains(currentOSName)) {
				throw new StandardError(
						MessageFormat.format(T.t("{0} is not compatible with {1}\nCompatible systems: {2}"), getName(),
								currentOSName, TextUtils.stringJoin(compatibleOSNames, ", ")));
			}
		}
	}

	protected String getCompatibleOSNames() {
		return null;
	}

	public String getInvalidCommandLineErrorMessage(String[] args) {
		return "Invalid command line parameters: " + Arrays.toString(args) + "\nExpected: " + "["
				+ OVERRIDE_CONFIG_COMMAND_LINE_OPTION + " <field_name> <field_value>] ... ["
				+ DEFINE_SYSTEM_PROPERTY_COMMAND_LINE_OPTION + " <property_key> <property_value>] ...["
				+ RUN_CLASS_COMMAND_LINE_OPTION + " <class_name> [<argument> ...]]";
	}

	protected synchronized void setUp() {
		if (active) {
			return;
		}
		active = true;
		this.translationHandler = createTranslationHandler();
		this.configuration = createConfiguration();
		this.log = createLog();
		this.errorManager = createErrorManager();
		this.analytics = createAnalytics();
		this.buildProperties = createBuildProperties();
		initializeSystem();
		initializeConfiguration();
		applyConfiguredLocale();
	}

	protected synchronized void cleanUp() {
		if (!active) {
			return;
		}
		getLog().info("Stopping application");
		getAnalytics().track("Stop");
		try {
			getAnalytics().ensureIsShutDown();
		} catch (Throwable ignore) {
		}
		active = false;
	}

	protected void initializeSystem() {
		System.setProperty("sun.net.client.defaultConnectTimeout", "10000");
		System.setProperty("com.otk.application.buildId", Integer.toString(getBuildProperties().getBuildNumber()));
	}

	protected void initializeConfiguration() {
		if (new File(configuration.getFilePath()).exists()) {
			configuration.load();
		}
	}

	public void applyConfiguredLocale() {
		synchronized (T.TRANSLATION_CHANGE_MUTEX) {
			getLog().info("Applying configuration locale: '"
					+ getConfiguration().getCategory(GlobalConfiguration.class).language + "'");
			Locale locale = new Locale(getConfiguration().getCategory(GlobalConfiguration.class).language);
			Locale.setDefault(locale);
			JComponent.setDefaultLocale(locale);
			getTranslationHandler().loadTranslations();
			T.setTranslationHandler(getTranslationHandler());
		}
	}

	public void launch() {
		getLog().info("Starting application");
		getLog().info("Build Id: " + getBuildProperties().getBuildNumber());
		getAnalytics().track("Start", "os.name=" + System.getProperty("os.name") + ", os.arch="
				+ System.getProperty("os.arch") + ", java.home=" + System.getProperty("java.home"));
		SwingUtilities.invokeLater(new Runnable() {
			@Override
			public void run() {
				checkCompatibility();
				if (getConfiguration().getCategory(GlobalConfiguration.class).logVerbose) {
					getConfiguration().log();
				}
				if (!processArguments()) {
					return;
				}
				display();
			}
		});
	}

	protected boolean processArguments() {
		try {
			int iArg = 0;
			while (iArg < args.length) {
				if (args[iArg].equals(DEFINE_SYSTEM_PROPERTY_COMMAND_LINE_OPTION)) {
					if (args.length < (iArg + 3)) {
						throw new StandardError(getInvalidCommandLineErrorMessage(args));
					}
					iArg++;
					String propertyName = args[iArg];
					iArg++;
					String propertyValueString = args[iArg];
					getLog().info("Defining system property: " + propertyName + "=" + propertyValueString);
					System.setProperty(propertyName, propertyValueString);
				} else if (args[iArg].equals(OVERRIDE_CONFIG_COMMAND_LINE_OPTION)) {
					if (args.length < (iArg + 3)) {
						throw new StandardError(getInvalidCommandLineErrorMessage(args));
					}
					iArg++;
					String fieldId = args[iArg];
					iArg++;
					String fieldValueString = args[iArg];
					getLog().info("Overriding configuration value: " + fieldId + "=" + fieldValueString);
					Field field = getConfiguration().getFieldFromIdentifier(fieldId);
					Object filedValue = Configuration.deserialize(fieldValueString, field);
					getConfiguration().setFieldValue(field, filedValue);
				} else if (args[iArg].equals(RUN_CLASS_COMMAND_LINE_OPTION)) {
					if (args.length < (iArg + 2)) {
						throw new StandardError(getInvalidCommandLineErrorMessage(args));
					}
					String className = args[iArg + 1];
					Method method = Class.forName(className).getMethod("main", String[].class);
					String[] classArgs = Arrays.copyOfRange(args, iArg + 2, args.length);
					method.invoke(null, new Object[] { classArgs });
					return false;
				} else {
					throw new StandardError(getInvalidCommandLineErrorMessage(args));
				}
				iArg++;
			}
			return true;
		} catch (Throwable t) {
			throw AbstractApplicationError.getToRethrow(t);
		}
	}

	protected void display() {
		mainWindow = createMainWindow();
		getMainWindow().setVisible(true);

	}

}