package com.otk.application;

import java.awt.Color;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import com.otk.application.error.UnexpectedError;
import com.otk.application.util.AbstractLazilyInitialized;
import com.otk.application.util.ImageUtils;
import com.otk.application.util.MiscUtils;
import com.otk.application.util.TextUtils;

import javassist.Modifier;

public class Configuration extends AbstractLazilyInitialized {

	public static final String CLASSES_SUFFIX = "Configuration";
	public static final String FILE_NAME_EXTENSION = "properties";

	private AbstractApplication application;
	private Map<Class<?>, Object> categoryByClass = new HashMap<Class<?>, Object>();
	private Map<String, Field> fieldById = new HashMap<String, Field>();

	public Configuration(AbstractApplication application) {
		this.application = application;
		createCategories();
	}

	@Override
	protected boolean initialize() {
		return true;
	}

	@Override
	protected boolean shutDown() {
		return true;
	}

	protected void createCategories() {
		for (Class<?> categoryClass : getAllCategoryClasses(getClass())) {
			Object category;
			if (isRootCategoryClass(categoryClass)) {
				category = this;
			} else {
				Class<?> parentCategoryClass = getParentCategoryClass(categoryClass);
				if (parentCategoryClass != null) {
					Object parentCategory = categoryByClass.get(parentCategoryClass);
					try {
						Constructor<?> ctor = categoryClass.getConstructor(parentCategoryClass);
						category = ctor.newInstance(parentCategory);
					} catch (Exception e) {
						throw new UnexpectedError(e);
					}
				} else {
					try {
						category = categoryClass.newInstance();
					} catch (Exception e) {
						throw new UnexpectedError(e);
					}
				}
			}
			categoryByClass.put(categoryClass, category);
			for (Field field : getFields(categoryClass)) {
				String fieldId = getCategoryIdentifier(field.getDeclaringClass()) + "." + field.getName();
				if (fieldById.containsKey(fieldId)) {
					throw new UnexpectedError();
				}
				fieldById.put(fieldId, field);
			}
		}
	}

	@SuppressWarnings("unchecked")
	public <C> C getCategory(Class<C> categoryClass) {
		if (!hasInitializationFailed()) {
			ensureIsInitialized();
		}
		return (C) categoryByClass.get(categoryClass);
	}

	private Class<?> getCategoryClass(Field field) {
		return field.getDeclaringClass();
	}

	private List<Field> getFields(Class<?> categoryClass) {
		List<Field> result = new ArrayList<Field>();
		for (Field f : categoryClass.getFields()) {
			if (Modifier.isFinal(f.getModifiers())) {
				continue;
			}
			result.add(f);
		}
		return result;
	}

	private String getCategoryIdentifier(Class<?> categoryClass) {
		List<String> categoryPath = new ArrayList<String>();
		while (true) {
			if (categoryClass == null) {
				return null;
			}
			if (isRootCategoryClass(categoryClass)) {
				break;
			}
			final String localName = categoryClass.getSimpleName();
			if (!localName.endsWith(CLASSES_SUFFIX)) {
				return null;
			}
			String shortLocalName = localName.substring(0, localName.length() - CLASSES_SUFFIX.length());
			categoryPath.add(0, shortLocalName);
			categoryClass = categoryClass.getEnclosingClass();
		}
		return MiscUtils.stringJoin(categoryPath, ".");
	}

	private boolean isRootCategoryClass(Class<?> categoryClass) {
		return categoryClass.isAssignableFrom(getClass());
	}

	private boolean isValidCategoyClass(Class<?> categoryClass) {
		return getCategoryIdentifier(categoryClass) != null;
	}

	private List<Class<?>> getAllCategoryClasses(Class<?> baseCategoryClass) {
		List<Class<?>> result = new ArrayList<Class<?>>();
		if (!isValidCategoyClass(baseCategoryClass)) {
			return Collections.emptyList();
		}
		result.add(baseCategoryClass);
		if (baseCategoryClass.getSuperclass() != null) {
			result.addAll(getAllCategoryClasses(baseCategoryClass.getSuperclass()));
		}
		for (Class<?> subCategoryClasses : getSubCategoryClass(baseCategoryClass)) {
			result.addAll(getAllCategoryClasses(subCategoryClasses));
		}
		return result;
	}

	private Class<?>[] getSubCategoryClass(Class<?> categoryClass) {
		return categoryClass.getDeclaredClasses();
	}

	private Class<?> getParentCategoryClass(Class<?> categoryClass) {
		return categoryClass.getDeclaringClass();
	}

	public String getFilePath() {
		return "app." + FILE_NAME_EXTENSION;
	}

	public static Object deserialize(String stringValue, Field field) throws Exception {
		Class<?> fieldType = field.getType();
		if (fieldType.isPrimitive()) {
			if (fieldType == int.class) {
				fieldType = Integer.class;
			} else if (fieldType == long.class) {
				fieldType = Long.class;
			} else if (fieldType == float.class) {
				fieldType = Float.class;
			} else if (fieldType == double.class) {
				fieldType = Double.class;
			} else if (fieldType == boolean.class) {
				fieldType = Boolean.class;
			} else if (fieldType == char.class) {
				fieldType = Character.class;
			} else {
				throw new Exception("Unhandled type: " + fieldType);
			}
		}
		Object value;
		if (fieldType == String.class) {
			value = stringValue;
		} else if (fieldType == Character.class) {
			stringValue = stringValue.trim();
			if (stringValue.length() != 1) {
				throw new Exception("Invalid value: '" + stringValue + "'. 1 character is expected");
			}
			value = stringValue.charAt(0);
		} else if (fieldType == Color.class) {
			value = ImageUtils.stringToColor(stringValue);
		} else {
			value = fieldType.getMethod("valueOf", new Class[] { String.class }).invoke(null, stringValue);
		}
		return value;
	}

	public AbstractApplication getApplication() {
		return application;
	}

	public void save() {
		save(getFilePath());
	}

	public void load() {
		load(getFilePath());
	}

	public synchronized void load(String filePath) {
		Properties properties = new Properties();
		File configFile = new File(filePath);
		getApplication().getLog().info("Loading the configuration file: '" + configFile.getAbsolutePath() + "'");
		FileInputStream in = null;
		try {
			in = new FileInputStream(configFile);
			properties.load(in);
		} catch (IOException e) {
			throw new UnexpectedError(e);
		} finally {
			if (in != null) {
				try {
					in.close();
				} catch (Exception ignore) {
				}
			}
		}
		load(properties);
	}

	public synchronized void load(Properties savedProperties) {
		for (Field field : getAllFields()) {
			String fieldId = getFieldIdentifier(field);
			try {
				if (!savedProperties.containsKey(fieldId)) {
					continue;
				}
				String propertyValue = savedProperties.getProperty(fieldId);
				Object filedValue = deserialize(propertyValue, field);
				setFieldValue(field, filedValue);
			} catch (Throwable t) {
				throw new UnexpectedError("Configurtion field '" + fieldId + "' loading error", t);
			}
		}
	}

	public void save(String filePath) {
		Properties currentProperties = asProperties();
		File configFile = new File(filePath);
		FileOutputStream out = null;
		try {
			out = new FileOutputStream(configFile);
			currentProperties.store(out, null);
		} catch (Exception e) {
			throw new UnexpectedError(e);
		} finally {
			if (out != null) {
				try {
					out.close();
				} catch (Exception ignore) {
				}
			}
		}
	}

	public Properties asProperties() {
		Properties result = MiscUtils.getSortedProperties();
		for (Field field : getAllFields()) {
			String fieldName = getFieldIdentifier(field);
			try {
				Object fieldVlaue = getFieldValue(field);
				String propertyValue = serialize(fieldVlaue);
				result.setProperty(fieldName, propertyValue);
			} catch (Throwable t) {
				throw new UnexpectedError("Configurtion field '" + fieldName + "' error", t);
			}
		}
		return result;
	}

	public Object getFieldValue(String fieldId) {
		Field field = getFieldFromIdentifier(fieldId);
		return getFieldValue(field);
	}

	public Object getFieldValue(Field field) {
		Class<?> categoryClass = getCategoryClass(field);
		Object category = getCategory(categoryClass);
		try {
			return field.get(category);
		} catch (Throwable t) {
			throw new UnexpectedError("Configurtion field '" + field + "' error", t);
		}
	}

	public void setFieldValue(String fieldId, Object filedValue) {
		Field field = getFieldFromIdentifier(fieldId);
		setFieldValue(field, filedValue);
	}

	public void setFieldValue(Field field, Object filedValue) {
		Class<?> categoryClass = getCategoryClass(field);
		Object category = getCategory(categoryClass);
		try {
			field.set(category, filedValue);
		} catch (Throwable t) {
			throw new UnexpectedError("Configurtion field '" + field + "' loading error", t);
		}
	}

	public Collection<Object> getAllCategories() {
		Collection<Object> result = new ArrayList<Object>();
		for (Class<?> categoryClass : categoryByClass.keySet()) {
			if (getFields(categoryClass).size() == 0) {
				continue;
			}
			result.add(categoryByClass.get(categoryClass));
		}
		return result;
	}

	public Collection<Field> getAllFields() {
		return fieldById.values();
	}

	public Collection<String> getAllFieldIdentifiers() {
		return fieldById.keySet();
	}

	public String getFieldIdentifier(Field configField) {
		return MiscUtils.getFirstKeyFromValue(fieldById, configField);
	}

	public static String serialize(Object value) {
		if (value instanceof Color) {
			return ImageUtils.toString((Color) value);
		}
		return value.toString();
	}

	public Field getFieldFromIdentifier(String fieldId) {
		return fieldById.get(fieldId);
	}

	public String getFieldIdentifier(Class<?> categoryClass, String fieldName) {
		Field field;
		try {
			field = categoryClass.getField(fieldName);
		} catch (Exception e) {
			throw new UnexpectedError(
					"Failed to get field '" + fieldName + "' from class '" + categoryClass.getName() + "'", e);
		}
		return getFieldIdentifier(field);
	}

	public void log() {
		getApplication().getLog().info("Current configuration:\n" + TextUtils.toText(asProperties()));
	}

	public class GlobalConfiguration {
		public String language = getApplication().getTranslationHandler().getMostSuitedLocale().getLanguage();
		public boolean updateNotificationsEnabled = true;
		public boolean usageTrackingEnabled = false;

		public int maximumStoredTrackingCount = 1000;
		public int trackingsDeliveryPeriodSeconds = 3600;
		public int trackingDeliveryTimeoutMilliseconds = 5000;
		public int databaseAccessRetryPeriodSeconds = 30;

		public boolean logFileEnabled = true;
		public int logFileMaxCount = 10;
		public long logFileMaxSizeBytes = 100000;
		public boolean logVerbose = false;

		@Override
		public String toString() {
			return "General";
		}
	}

}