/*******************************************************************************
 * Copyright (C) 2018 OTK Software
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ******************************************************************************/
package com.otk.application;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GraphicsEnvironment;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.Window;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.image.BufferedImage;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.UIDefaults;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;
import javax.swing.plaf.ComponentUI;
import javax.swing.plaf.metal.MetalLookAndFeel;
import javax.swing.plaf.metal.MetalScrollBarUI;

import com.otk.application.error.UnexpectedError;
import com.otk.application.util.TextUtils;

import xy.reflect.ui.CustomizedUI;
import xy.reflect.ui.control.swing.menu.SaveMenuItem;
import xy.reflect.ui.control.swing.renderer.CustomizedSwingRenderer;
import xy.reflect.ui.control.swing.renderer.Form;
import xy.reflect.ui.control.swing.renderer.SwingRenderer;
import xy.reflect.ui.control.swing.util.SwingRendererUtils;
import xy.reflect.ui.control.swing.util.WindowManager;
import xy.reflect.ui.info.field.IFieldInfo;
import xy.reflect.ui.info.method.IMethodInfo;
import xy.reflect.ui.info.method.InvocationData;
import xy.reflect.ui.info.parameter.IParameterInfo;
import xy.reflect.ui.info.type.ITypeInfo;
import xy.reflect.ui.info.type.factory.EncapsulatedObjectFactory;
import xy.reflect.ui.info.type.factory.InfoProxyFactory;
import xy.reflect.ui.util.PrecomputedTypeInstanceWrapper;
import xy.reflect.ui.util.ReflectionUIError;
import xy.reflect.ui.util.ReflectionUIUtils;

/**
 * This class generates the GUI. It uses the ReflectionUI library.
 * 
 * @author olitank
 *
 */
public class IdPhotoUI {

	/*
	 * Look & feel enhancement:
	 */
	static {
		try {
			UIManager.setLookAndFeel(new MetalLookAndFeel() {

				private static final long serialVersionUID = 1L;

				@Override
				protected void initClassDefaults(UIDefaults table) {
					super.initClassDefaults(table);
					table.put("ScrollBarUI", CustomScrollBarUI.class.getName());
				}

			});
		} catch (UnsupportedLookAndFeelException e) {
			throw new UnexpectedError(e);
		}
		Font baseFont = TextUtils.loadFont(IdPhotoUI.class.getResourceAsStream("baseFont.ttf"));
		{
			GraphicsEnvironment.getLocalGraphicsEnvironment().registerFont(baseFont);
			int fontSizeAdjustment = 0;
			for (Map.Entry<Object, Object> entry : javax.swing.UIManager.getDefaults().entrySet()) {
				Object key = entry.getKey();
				Object value = javax.swing.UIManager.get(key);
				if (value != null && value instanceof javax.swing.plaf.FontUIResource) {
					javax.swing.plaf.FontUIResource oldFont = (javax.swing.plaf.FontUIResource) value;
					javax.swing.plaf.FontUIResource newFont = new javax.swing.plaf.FontUIResource(baseFont.getFamily(),
							oldFont.getStyle(), oldFont.getSize() + fontSizeAdjustment);
					javax.swing.UIManager.put(key, newFont);
				}
			}
		}
		System.setProperty("awt.useSystemAAFontSettings", "on");
		System.setProperty("swing.aatext", "true");
	}

	/**
	 * Whether the ReflectionUI design mode is active or not.
	 */
	private static final boolean DESIGN_MODE = Boolean
			.valueOf(System.getProperty(IdPhotoUI.class.getName() + ".designMode", "false"));

	/**
	 * Used to complete (allow to undo) and improve (add usage tracking) some
	 * ReflectionUI type informations.
	 */
	public static final InfoProxyFactory TYPEFACTORY = new InfoProxyFactory() {

		@Override
		protected List<IParameterInfo> getParameters(IMethodInfo method, ITypeInfo containingType) {
			if ((containingType.getName().equals(IdPhotoProject.class.getName()))
					&& method.getName().equals("setPhotograph")) {
				RENDERER.getLastInvocationDataByMethodSignature().remove(method.getSignature());
			}
			return super.getParameters(method, containingType);
		}

		@Override
		protected Runnable getNextInvocationUndoJob(final IMethodInfo method, ITypeInfo containingType, Object object,
				InvocationData invocationData) {
			if ((object instanceof IdPhotoProject) && method.getName().startsWith("setPhotograph")) {
				final IdPhotoProject idPhotoProject = (IdPhotoProject) object;
				final Image oldPhotograph = idPhotoProject.getRawPhotograph();
				return new Runnable() {
					@Override
					public void run() {
						IMethodInfo setPhotographMethod = ReflectionUIUtils.findInfoByName(containingType.getMethods(),
								"setPhotograph");
						invoke(object, new InvocationData(object, setPhotographMethod, oldPhotograph),
								setPhotographMethod, containingType);
					}
				};
			} else if ((object instanceof IdPhotoProject) && method.getName().startsWith("addFaceFrame")
					&& method.getName().endsWith("Ratio")) {
				final double ratioOffset = (double) invocationData.getParameterValue(0);
				return new Runnable() {
					@Override
					public void run() {
						InvocationData oppositeInvocationData = new InvocationData(object, method, -ratioOffset);
						method.invoke(object, oppositeInvocationData);
					}
				};
			} else {
				return super.getNextInvocationUndoJob(method, containingType, object, invocationData);
			}
		}

		@Override
		protected void save(ITypeInfo type, Object object, OutputStream out) {
			if (object instanceof IdPhotoProject) {
				Main.INSTANCE.getApplication().getAnalytics().track("Saving(" + object + ")");
				((IdPhotoProject) object).save(out);
			} else {
				super.save(type, object, out);
			}
		}

		@Override
		protected void load(ITypeInfo type, Object object, InputStream in) {
			if (object instanceof IdPhotoProject) {
				Main.INSTANCE.getApplication().getAnalytics().track("Loading(" + object + ")");
				((IdPhotoProject) object).load(in);
			} else {
				super.load(type, object, in);
			}
		}

		@Override
		protected void setValue(Object object, Object value, IFieldInfo field, ITypeInfo containingType) {
			if (object instanceof IdPhotoProject) {
				Main.INSTANCE.getApplication().getAnalytics().track("Setting(" + object + " - " + field.getName() + ")",
						String.valueOf(value));
			}
			super.setValue(object, value, field, containingType);
		}

		@Override
		protected Object getValue(Object object, IFieldInfo field, ITypeInfo containingType) {
			if (object instanceof IdPhotoProject) {
				if (field.getName().equals("detailedPrintPreview")) {
					Main.INSTANCE.getApplication().getAnalytics()
							.track("Getting(" + object + " - " + field.getName() + ")");
				}
			}
			return super.getValue(object, field, containingType);
		}

		@Override
		protected Object invoke(Object object, InvocationData invocationData, IMethodInfo method,
				ITypeInfo containingType) {
			if (object instanceof IdPhotoProject) {
				Main.INSTANCE.getApplication().getAnalytics().track("Invoking " + method.getName() + "(",
						invocationData.toString() + ") on " + object);
			}
			return super.invoke(object, invocationData, method, containingType);
		}

	};

	/**
	 * Project-specific {@link CustomizedUI} implementation.
	 */
	public static final CustomizedUI REFLECTIONUI = new CustomizedUI() {
		@Override
		protected ITypeInfo getTypeInfoBeforeCustomizations(ITypeInfo type) {
			type = TYPEFACTORY.wrapTypeInfo(type);
			return type;
		}
	};

	/**
	 * Project-specific {@link SwingRenderer} implementation.
	 */
	public static final SwingRenderer RENDERER = DESIGN_MODE ? createDesignModeRenderer() : createClassicRenderer();

	/**
	 * @return A project-specific {@link SwingRenderer} implementation allowing to
	 *         generate and edit the GUI.
	 */
	private static CustomizedSwingRenderer createDesignModeRenderer() {
		String cutomizationsFilePath = "src/main/resources/" + IdPhotoUI.class.getPackage().getName().replace('.', '/')
				+ "/default.icu";
		try {
			return (CustomizedSwingRenderer) Class.forName("xy.reflect.ui.control.swing.customizer.SwingCustomizer")
					.getConstructor(CustomizedUI.class, String.class).newInstance(REFLECTIONUI, cutomizationsFilePath);
		} catch (Exception e) {
			throw new UnexpectedError(e);
		}
	}

	/**
	 * @return A project-specific {@link SwingRenderer} implementation allowing to
	 *         just generate the GUI.
	 */
	private static CustomizedSwingRenderer createClassicRenderer() {
		return new CustomizedSwingRenderer(REFLECTIONUI) {
			{
				try {
					getInfoCustomizations().loadFromStream(IdPhotoUI.class.getResourceAsStream("default.icu"), null);
				} catch (Throwable t) {
					throw new AssertionError(t);
				}
			}

			@Override
			public WindowManager createWindowManager(Window window) {
				return new WindowManager(this, window) {

					@Override
					public void install(Component content, List<Component> toolbarControls, String title,
							Image iconImage) {
						super.install(content, toolbarControls, title, iconImage);
						if (content != null) {
							if (SwingRendererUtils.isForm(content, swingRenderer)) {
								Form rootForm = (Form) content;
								Object object = rootForm.getObject();
								if (object instanceof PrecomputedTypeInstanceWrapper) {
									object = ((PrecomputedTypeInstanceWrapper) object).unwrap();
									if (object instanceof EncapsulatedObjectFactory.Instance) {
										object = ((EncapsulatedObjectFactory.Instance) object).getValue();
										if (object instanceof IdPhotoProject) {
											Form idPhotoProjectForm = SwingRendererUtils
													.findFirstObjectDescendantForm(object, rootForm, RENDERER);
											manageIdPhotoWindowLifeCycle(window, rootForm, idPhotoProjectForm);
										}
									}
								}
							}
						}
					}

					void manageIdPhotoWindowLifeCycle(final Window window, final Form rootForm,
							final Form idPhotoProjectForm) {
						window.addWindowListener(new WindowAdapter() {

							@Override
							public void windowOpened(WindowEvent e) {
								Main.INSTANCE.getApplication().getAnalytics()
										.track("windowOpened(" + idPhotoProjectForm.getObject() + ")");
							}

							@Override
							public void windowClosing(WindowEvent e) {
								Main.INSTANCE.getApplication().getAnalytics()
										.track("windowClosing(" + idPhotoProjectForm.getObject() + ")");
								JMenuBar menuBar = rootForm.getMenuBar();
								SaveMenuItem saveMenuItem = null;
								for (int iMenu = 0; iMenu < menuBar.getMenuCount(); iMenu++) {
									JMenu menu = rootForm.getMenuBar().getMenu(iMenu);
									for (int iItem = 0; iItem < menu.getItemCount(); iItem++) {
										JMenuItem item = menu.getItem(iItem);
										if (item instanceof SaveMenuItem) {
											saveMenuItem = (SaveMenuItem) item;
											break;
										}
									}
								}
								if (saveMenuItem == null) {
									throw new UnexpectedError();
								}
								if (!saveMenuItem.isFileSynchronized()) {
									String objectTitle = RENDERER.getObjectTitle(idPhotoProjectForm.getObject());
									if (RENDERER.openQuestionDialog(form,
											"The '" + objectTitle
													+ "' has been modified.\nSave changes before closing?",
											objectTitle)) {
										saveMenuItem.execute();
									}
								}
							}
						});
					}

				};
			}

			@Override
			public void handleObjectException(Component activatorComponent, Throwable t) {
				Main.INSTANCE.getApplication().getAnalytics().track("Uncaught Exception",
						t.toString() + " " + Arrays.toString(t.getStackTrace()));
				if (new ReflectionUIError(t).toString()
						.contains(new ReflectionUIError(new OutOfMemoryError()).toString())) {
					String errorMessage = "The maximum memory is not enough to perform the requested action.";
					if (System.getProperty(Main.LAUNCHER_PROPERTY_KEY) == null) {
						errorMessage += "\nRestart the program with more memory by using this command (for example):"
								+ "\njava -Xmx1024M phoyo-id.jar";
					}
					t = new ReflectionUIError(errorMessage);
				}
				super.handleObjectException(activatorComponent, t);
			}

		};
	}

	/**
	 * A simple custom scrollbar class.
	 * 
	 * @author olitank
	 *
	 */
	public static class CustomScrollBarUI extends MetalScrollBarUI {

		public static ComponentUI createUI(JComponent c) {
			return new CustomScrollBarUI();
		}

		private Image imageThumb, imageTrack;
		private JButton b = new JButton() {

			private static final long serialVersionUID = 1L;

			@Override
			public Dimension getPreferredSize() {
				return new Dimension(0, 0);
			}

		};

		CustomScrollBarUI() {
			imageThumb = SimpleImage.create(32, 32, Color.darkGray);
			imageTrack = SimpleImage.create(32, 32, Color.lightGray);
		}

		@Override
		protected void paintThumb(Graphics g, JComponent c, Rectangle r) {
			g.setColor(Color.blue);
			int border = 4;
			((Graphics2D) g).drawImage(imageThumb, r.x + border, r.y + border, r.width - 2 * border,
					r.height - 2 * border, null);
		}

		@Override
		protected void paintTrack(Graphics g, JComponent c, Rectangle r) {
			((Graphics2D) g).drawImage(imageTrack, r.x, r.y, r.width, r.height, null);
		}

		@Override
		protected JButton createDecreaseButton(int orientation) {
			return b;
		}

		@Override
		protected JButton createIncreaseButton(int orientation) {
			return b;
		}

		private static class SimpleImage {

			static public Image create(int w, int h, Color c) {
				BufferedImage bi = new BufferedImage(w, h, BufferedImage.TYPE_INT_ARGB);
				Graphics2D g2d = bi.createGraphics();
				g2d.setPaint(c);
				g2d.fillRect(0, 0, w, h);
				g2d.dispose();
				return bi;
			}
		}
	}

}
