package com.otk.application;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStream;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;

import com.otk.application.Configuration.GlobalConfiguration;

import com.otk.application.error.UnexpectedError;
import com.otk.application.util.AbstractLazilyInitialized;
import com.otk.application.util.FileUtils;
import com.otk.application.util.Listener;
import com.otk.application.util.MiscUtils;
import com.otk.application.util.TextUtils;

public class Log extends AbstractLazilyInitialized {

	private AbstractApplication application;
	private List<Listener<LogEvent>> listeners = new CopyOnWriteArrayList<Listener<LogEvent>>();
	private File logFile;

	public Log(AbstractApplication application) {
		super();
		this.application = application;
	}

	public AbstractApplication getApplication() {
		return application;
	}

	@Override
	protected boolean initialize() throws Throwable {
		logFile = new File(getFilePath());
		if (getApplication().getConfiguration().getCategory(GlobalConfiguration.class).logFileEnabled) {
			if (!logFile.exists()) {
				FileUtils.createFile(logFile);
			}
		}
		return true;
	}

	@Override
	protected boolean shutDown() {
		return true;
	}

	protected synchronized void writeLine(Level level, Thread thread, String message) {
		String date = TextUtils.getDefaultDateFormat().format(MiscUtils.now());
		String formattedMessage = String.format("%1$s %3$s - [%2$s] %4$s", date, thread.getName(), level.toString(),
				message);
		System.out.println(formattedMessage);
		if (ensureIsInitialized()) {
			if (getApplication().getConfiguration().getCategory(GlobalConfiguration.class).logFileEnabled) {
				try {
					if (logFile.length() > getApplication().getConfiguration()
							.getCategory(GlobalConfiguration.class).logFileMaxSizeBytes) {
						System.out.println("Rotating log files");
						FileUtils.historizeFile(getFilePath(), getApplication().getConfiguration()
								.getCategory(GlobalConfiguration.class).logFileMaxCount);
					}
					FileWriter logOut = new FileWriter(logFile, true);
					logOut.write(TextUtils.fixNewLines(formattedMessage + "\n"));
					logOut.flush();
					logOut.close();
				} catch (IOException e) {
					throw new UnexpectedError(e);
				}
			}
		}
		for (Listener<LogEvent> l : listeners) {
			l.handle(new LogEvent(level, message));
		}
	}

	public List<Listener<LogEvent>> getListeners() {
		return listeners;
	}

	public void error(Throwable t) {
		error(MiscUtils.getPrintedStackTrace(t));
	}

	public void error(String s) {
		writeLine(Level.ERROR, Thread.currentThread(), s);
	}

	public void info(String s) {
		writeLine(Level.INFO, Thread.currentThread(), s);
	}

	public String getFilePath() {
		return "app.log";
	}

	public OutputStream getStream(final Level level) {
		return new OutputStream() {
			private String line = "";
			private Thread thread = Thread.currentThread();
			private char lastC = 0;
			private char currentC = 0;

			public void write(int b) {
				currentC = (char) (b & 0xff);
				char standardC = TextUtils.standardizeNewLineSequences(lastC, currentC);
				try {
					if (standardC == 0) {
						return;
					}
					if (standardC == '\n') {
						flush();
					} else {
						line = line + standardC;
					}
				} finally {
					lastC = currentC;
				}
			}

			public void flush() {
				writeLine(level, thread, line);
				line = "";
			}

			@Override
			protected void finalize() throws Throwable {
				if (line.length() > 0) {
					flush();
				}
			}

		};
	}

	public class LogEvent {
		private Level level;
		private String message;

		public LogEvent(Level level, String message) {
			super();
			this.level = level;
			this.message = message;
		}

		public Level getLevel() {
			return level;
		}

		public String getMessage() {
			return message;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((level == null) ? 0 : level.hashCode());
			result = prime * result + ((message == null) ? 0 : message.hashCode());
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			LogEvent other = (LogEvent) obj;
			if (level != other.level)
				return false;
			if (message == null) {
				if (other.message != null)
					return false;
			} else if (!message.equals(other.message))
				return false;
			return true;
		}

		@Override
		public String toString() {
			return "LogEvent [level=" + level + ", message=" + message + "]";
		}

	}

	public enum Level {
		INFO, ERROR
	}

}