/*******************************************************************************
 * Copyright (C) 2018 OTK Software
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ******************************************************************************/
package com.otk.application;

import java.awt.Image;
import java.io.IOException;
import java.net.URL;
import javax.swing.JFrame;

import com.otk.application.Configuration.GlobalConfiguration;
import com.otk.application.error.ErrorManager;
import com.otk.application.error.UnexpectedError;
import com.otk.application.util.ImageUtils;

import xy.reflect.ui.control.swing.builder.StandardEditorBuilder;

/**
 * This class serves as the main entry point of the application. It handles all
 * initialization and cleanup tasks.
 * 
 * @author olitank
 *
 */
public class Main {

	public static void main(String[] args) {
		Main.INSTANCE.setInitialApplicationArguments(args);
		Main.INSTANCE.getApplication().launch();
		Runtime.getRuntime().addShutdownHook(new Thread() {
			@Override
			public void run() {
				Main.INSTANCE.getApplication().finalize();
			}
		});
	}

	/**
	 * System property name allowing to identify at runtime the type of packaging
	 * used for the application.
	 */
	public static final String LAUNCHER_PROPERTY_KEY = Main.class.getPackage().getName() + ".launcher";

	/**
	 * The support website.
	 */
	public static final String WEBSITE_URL = "http://www.phoyosystem.com";

	/**
	 * The singleton of this main class.
	 */
	public static final Main INSTANCE = new Main();

	/**
	 * The application instance.
	 */
	protected AbstractApplication application;
	/**
	 * Arguments used to initialize the application (should be retrieved from the
	 * command line).
	 */
	protected String[] initialApplicationArguments;

	/**
	 * This main class constructor made private since a singleton is used.
	 */
	private Main() {
	}

	/**
	 * @return arguments used to initialize the application.
	 */
	public String[] getInitialApplicationArguments() {
		return initialApplicationArguments;
	}

	/**
	 * Sets arguments used to initialize the application (must be set before
	 * creating the application object).
	 * 
	 * @param initialApplicationArguments
	 */
	public void setInitialApplicationArguments(String[] initialApplicationArguments) {
		this.initialApplicationArguments = initialApplicationArguments;
	}

	/**
	 * @return the application instance.
	 */
	public AbstractApplication getApplication() {
		if (application == null) {
			application = createApplication();
		}
		return application;
	}

	/**
	 * @return a new application instance.
	 */
	protected AbstractApplication createApplication() {
		return new AbstractApplication(initialApplicationArguments) {

			private Image iconImage;

			@Override
			public String getName() {
				return "Phoyo-id";
			}

			@Override
			public Image getIconImage() {
				if (iconImage == null) {
					try {
						iconImage = ImageUtils.loadAdaptedImage(Main.class.getResourceAsStream("app.gif"));
					} catch (IOException e) {
						throw new UnexpectedError(e);
					}
				}
				return iconImage;
			}

			@Override
			protected com.otk.application.Configuration createConfiguration() {
				return new com.otk.application.Configuration(this) {
					{
						getCategory(GlobalConfiguration.class).logFileEnabled = false;
						getCategory(GlobalConfiguration.class).usageTrackingEnabled = true;
						getCategory(GlobalConfiguration.class).updateNotificationsEnabled = false;
					}

					@Override
					public String getFilePath() {
						return "app.ini";
					}
				};
			}

			@Override
			protected AbstractAnalytics createAnalytics() {
				return new AbstractAnalytics(this) {

					@Override
					public String getDeliveryUrl() {
						return "http://www.phoyosystem.com/custom/forward-http-to-google-analytics.php";
					}

					@Override
					public void sendAllTrackings() {
						if (!IdPhotoProject.isUsageStatisticsShared()) {
							if (getApplication().getConfiguration().getCategory(GlobalConfiguration.class).logVerbose) {
								getApplication().getLog().info("Skipping trackings delivery");
							}
							trackings.clear();
							return;
						}
						super.sendAllTrackings();
					}
				};
			}

			@Override
			protected AbstractBuildProperties createBuildProperties() {
				return new AbstractBuildProperties() {

					@Override
					protected URL getURL() {
						return null;
					}

					@Override
					public int getBuildNumber() {
						return 77;
					}

					@Override
					public String getVersion() {
						return "1.9.5";
					}
				};
			}

			@Override
			protected ErrorManager createErrorManager() {
				return new ErrorManager(this);
			}

			@Override
			protected JFrame createMainWindow() {
				StandardEditorBuilder windowBuilder = IdPhotoUI.RENDERER.createEditorBuilder(null, new IdPhotoProject(),
						getName(), getIconImage(), false);
				return windowBuilder.createFrame();
			}

		};
	}

}
