package com.otk.application.audio;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.SequenceInputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.WeakHashMap;

import javax.sound.sampled.AudioFileFormat;
import javax.sound.sampled.AudioFileFormat.Type;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Control;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.DataLine.Info;
import javax.sound.sampled.LineListener;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.SourceDataLine;
import javax.sound.sampled.UnsupportedAudioFileException;

import com.otk.application.error.Assertion;
import com.otk.application.error.UnexpectedError;
import com.otk.application.image.camera.Camera;
import com.otk.application.util.CommandExecutor;
import com.otk.application.util.FileUtils;
import com.otk.application.util.MathUtils;
import com.otk.application.util.MiscUtils;
import com.otk.application.util.Pair;
import com.otk.application.util.TextUtils;

/**
 * Audio player class.
 * 
 * @author olitank
 *
 */
public class AudioEngine {

	public static void main(String[] args) throws Exception {
		System.out.println("You should hear simultaneously 5 short sounds and 1 long speech");
		AudioEngine audio = new AudioEngine();
		audio.play(new SoundTrack("test"), Camera.CAMERA_CLICK_SOUND, false);
		Sound shortSound = Camera.CAMERA_FOCUS_SOUND;
		audio.play(SoundTrack.DEFAULT, shortSound, true);
		audio.play(SoundTrack.DEFAULT, shortSound, false);
		MiscUtils.sleepSafely(MathUtils.round(1000 * audio.getDurationSeconds(shortSound)));
		audio.rotate(SoundTrack.DEFAULT, new Sound[] { shortSound }, 3, new ISoundEndListener() {

			@Override
			public void handleInterruption() {
			}

			@Override
			public void handleCompletion(int remainingPlays) {
			}
		});
	}

	private static final Sound SILENCE_1_SECOND = Sound.load(AudioEngine.class, "silence1s.wav");
	private static final ISoundEndListener NO_WAIT_SOUND_LISTENER = new ISoundEndListener() {

		@Override
		public void handleInterruption() {
		}

		@Override
		public void handleCompletion(int remainingPlays) {
		}
	};

	private Map<Pair<SoundTrack, String>, SourceDataLine> soundLineCache = new HashMap<Pair<SoundTrack, String>, SourceDataLine>();
	private Object soundLineCacheMutex = new Object();

	private Map<Sound, Float> soundDurationCache = new WeakHashMap<Sound, Float>();
	private Object soundDurationCacheMutex = new Object();

	private boolean verbose;

	private String customDriverPlayCommandPattern = "";
	private int estimatedPlaybackBeginingSkippedSeconds = 0;
	private int bufferSizeMegaBytes = 16;
	private int estimatedSoundLineInitializationSeconds = 2;
	private float targetFormatSampleRate = 44100.0f;
	private int targetFormatSampleSizeInBits = 16;
	private int targetFormatChannels = 1;
	private boolean targetFormatSigned = true;
	private boolean targetFormatBigEndian = false;

	protected void logInfo(String s) {
		System.out.println(s);
	}

	protected void logError(String s) {
		System.err.println(s);
	}

	public Sound getLastSoundData(SoundTrack track) {
		return track.lastSoundData;
	}

	public void playList(SoundTrack track, Sound[] sounds, boolean wait) {
		rotate(track, sounds, sounds.length, wait);
	}

	public void playList(SoundTrack track, Sound[] sounds, final ISoundEndListener loopListener) {
		rotate(track, sounds, sounds.length, loopListener);
	}

	public void loop(SoundTrack track, Sound sound, final int iterations, boolean wait) {
		rotate(track, new Sound[] { sound }, iterations, wait);
	}

	public void loop(SoundTrack track, Sound sound, final int iterations, final ISoundEndListener loopListener) {
		rotate(track, new Sound[] { sound }, iterations, loopListener);
	}

	public void rotate(SoundTrack track, final Sound[] sounds, final int playCount, boolean wait) {
		rotate(track, sounds, playCount, wait ? null : NO_WAIT_SOUND_LISTENER);
	}

	public void rotate(SoundTrack track, final Sound[] sounds, final int playCount, ISoundEndListener loopListener) {
		Assertion.check(new Assertion.Predicate() {
			@Override
			protected boolean isTrue() {
				return playCount > 0;
			}
		});
		final boolean[] mustWait = new boolean[] { false };
		if (loopListener == null) {
			mustWait[0] = true;
			loopListener = new ISoundEndListener() {
				@Override
				public void handleInterruption() {
					mustWait[0] = false;
				}

				@Override
				public void handleCompletion(int remainingPlays) {
					if (remainingPlays == 0) {
						mustWait[0] = false;
					}
				}
			};
		}
		synchronized (track) {
			track.requestingIntrerruption = true;
			waitForInterruption(track);
			track.requestingIntrerruption = false;
			if (verbose) {
				logInfo("Playing " + getSynchronizationAdjective(loopListener) + " audio sounds (" + playCount
						+ " rotative execution(s) on " + TextUtils.toString(sounds) + ")");
			}
			startRotation(track, sounds, playCount, loopListener);
		}
		while (mustWait[0]) {
			MiscUtils.relieveCPU();
		}
	}

	private String getSynchronizationAdjective(ISoundEndListener listener) {
		if (listener == null) {
			return "synchronous";
		} else {
			return "asynchronous";
		}
	}

	private void startRotation(final SoundTrack track, final Sound[] sounds, final int playCount,
			final ISoundEndListener loopListener) {
		track.rotating = true;
		startClip(track, sounds[0], new ISoundEndListener() {

			@Override
			public void handleInterruption() {
				track.rotating = false;
				if (loopListener != null) {
					new Thread(MiscUtils.formatThreadName(AudioEngine.class, "Rotation Interruption Notifier")) {
						{
							setDaemon(true);
						}

						@Override
						public void run() {
							loopListener.handleInterruption();
						}
					}.start();
				}
			}

			@Override
			public void handleCompletion(final int remainingPlays) {
				Assertion.check(new Assertion.Predicate() {
					@Override
					protected boolean isTrue() {
						return remainingPlays == 0;
					}
				});
				if (playCount > 1) {
					if (verbose) {
						logInfo("Continuing audio sound loop (" + (playCount - 1) + " remaining iteration(s))");
					}
					new Thread(MiscUtils.formatThreadName(AudioEngine.class,
							"Rotation Intermediate Step Completion Notifier")) {
						{
							setDaemon(true);
						}

						@Override
						public void run() {
							loopListener.handleCompletion(playCount - 1);
						}
					}.start();
					startRotation(track, shiftedLeft(sounds), playCount - 1, loopListener);
				} else {
					track.rotating = false;
					if (loopListener != null) {
						new Thread(MiscUtils.formatThreadName(AudioEngine.class,
								"Rotation Last Step Completion Notifier")) {
							{
								setDaemon(true);
							}

							@Override
							public void run() {
								loopListener.handleCompletion(0);
							}
						}.start();
					}
				}
			}

		});
	}

	public void play(SoundTrack track, Sound data, boolean wait) {
		play(track, data, wait ? null : NO_WAIT_SOUND_LISTENER);
	}

	public void play(SoundTrack track, Sound data, final ISoundEndListener listener) {
		rotate(track, new Sound[] { data }, 1, listener);
	}

	private void startClip(final SoundTrack track, Sound data, final ISoundEndListener listener) {
		track.lastSoundData = data;
		if (verbose) {
			logInfo("Starting " + getSynchronizationAdjective(listener) + " audio sound play: " + data);
		}
		startClipAndWhentItEnds(track, data, new Runnable() {

			@Override
			public void run() {
				closeClip(track);
				if (track.requestingIntrerruption) {
					if (listener != null) {
						listener.handleInterruption();
					}
				} else {
					if (listener != null) {
						listener.handleCompletion(0);
					}
				}

			}
		});
	}

	public void stop(SoundTrack track) {
		synchronized (track) {
			track.requestingIntrerruption = true;
			waitForInterruption(track);
			track.requestingIntrerruption = false;
		}
	}

	private void waitForInterruption(SoundTrack track) {
		if (verbose) {
			logInfo("Waiting for currently playing audio sound(s) interruption. Current sound: " + track.lastSoundData);
		}
		while (isClipOpen(track) || track.rotating) {
			MiscUtils.relieveCPU();
		}
	}

	private void startClipAndWhentItEnds(final SoundTrack track, final Sound soundData, final Runnable runnable) {
		final Thread clipRunner = new Thread(MiscUtils.formatThreadName(AudioEngine.class, "Clip Runner")) {
			{
				setDaemon(true);
			}

			@Override
			public void run() {
				runClip(track, soundData);
			}
		};
		track.clipRunner = clipRunner;
		if (!track.requestingIntrerruption) {
			clipRunner.start();
		}
		new Thread(MiscUtils.formatThreadName(AudioEngine.class, "Clip Monitor")) {
			{
				setDaemon(true);
			}

			public void run() {
				while (clipRunner.isAlive()) {
					if (track.requestingIntrerruption) {
						clipRunner.interrupt();
						try {
							clipRunner.join();
						} catch (InterruptedException e) {
							throw new UnexpectedError(e);
						}
						break;
					}
					MiscUtils.relieveCPU();
				}
				runnable.run();
			}
		}.start();
	}

	private void runClip(SoundTrack track, Sound soundData) {
		for (int i = 0; i < estimatedPlaybackBeginingSkippedSeconds; i++) {
			soundData = join(SILENCE_1_SECOND, soundData);
		}
		if (customDriverPlayCommandPattern.length() > 0) {
			runCommandLineClip(soundData);
		} else {
			defaultRunClip(track, soundData);
		}
		if (verbose) {
			logInfo((track.requestingIntrerruption ? "Interrupted" : "Ended") + " audio sound play: "
					+ track.lastSoundData);
		}
	}

	private void defaultRunClip(SoundTrack track, Sound soundData) {
		AudioInputStream audioInputStream = getStream(soundData);
		SourceDataLine soundLine = getSoundLine(audioInputStream, track);
		feedSoundLine(soundLine, audioInputStream, track, getDurationSeconds(soundData));
	}

	private SourceDataLine getSoundLine(AudioInputStream audioInputStream, SoundTrack track) {
		AudioFormat audioFormat = audioInputStream.getFormat();
		return getSoundLine(audioFormat, track);
	}

	private SourceDataLine getSoundLine(AudioFormat audioFormat, SoundTrack track) {
		synchronized (soundLineCacheMutex) {
			Pair<SoundTrack, String> soundLineKey = new Pair<SoundTrack, String>(track, audioFormat.toString());
			SourceDataLine result = soundLineCache.get(soundLineKey);
			if (result == null) {
				if (verbose) {
					logInfo("Opening audio line on track " + track + " for format " + audioFormat);
				}
				DataLine.Info lineInfo = new DataLine.Info(SourceDataLine.class, audioFormat);
				try {
					result = (SourceDataLine) AudioSystem.getLine(lineInfo);
					result.open(audioFormat);
				} catch (Throwable e) {
					logError(MiscUtils.getPrintedStackTrace(e));
					result = getNullSoundDataLine(lineInfo);
				}
				soundLineCache.put(soundLineKey, result);
				MiscUtils.sleepSafely(1000 * estimatedSoundLineInitializationSeconds);
			}
			return result;
		}
	}

	private SourceDataLine getNullSoundDataLine(final Info lineInfo) {
		return new SourceDataLine() {

			@Override
			public void removeLineListener(LineListener listener) {
				throw new UnsupportedOperationException();
			}

			@Override
			public void addLineListener(LineListener listener) {
				throw new UnsupportedOperationException();
			}

			@Override
			public void open() throws LineUnavailableException {
			}

			@Override
			public void open(AudioFormat format, int bufferSize) throws LineUnavailableException {
			}

			@Override
			public void open(AudioFormat format) throws LineUnavailableException {
			}

			@Override
			public boolean isOpen() {
				return true;
			}

			@Override
			public boolean isControlSupported(javax.sound.sampled.Control.Type control) {
				return false;
			}

			@Override
			public javax.sound.sampled.Line.Info getLineInfo() {
				return lineInfo;
			}

			@Override
			public Control[] getControls() {
				return new Control[0];
			}

			@Override
			public Control getControl(javax.sound.sampled.Control.Type control) {
				return null;
			}

			@Override
			public void close() {
			}

			@Override
			public void start() {
			}

			@Override
			public void stop() {
			}

			@Override
			public boolean isRunning() {
				return false;
			}

			@Override
			public boolean isActive() {
				return false;
			}

			@Override
			public long getMicrosecondPosition() {
				return -1;
			}

			@Override
			public long getLongFramePosition() {
				return -1;
			}

			@Override
			public float getLevel() {
				return -1;
			}

			@Override
			public int getFramePosition() {
				return -1;
			}

			@Override
			public AudioFormat getFormat() {
				return lineInfo.getFormats()[0];
			}

			@Override
			public int getBufferSize() {
				return -1;
			}

			@Override
			public void flush() {
			}

			@Override
			public void drain() {
			}

			@Override
			public int available() {
				return 0;
			}

			@Override
			public int write(byte[] b, int off, int len) {
				return -1;
			}

		};
	}

	public AudioInputStream getStream(Sound soundData) {
		AudioInputStream result;
		try {
			result = AudioSystem.getAudioInputStream(new ByteArrayInputStream(soundData.getBytes()));
		} catch (UnsupportedAudioFileException e) {
			throw new UnexpectedError("Audio format not supported", e);
		} catch (IOException e) {
			throw new UnexpectedError(e);
		}
		try {
			result = AudioSystem.getAudioInputStream(getTargetFormat(), result);
		} catch (IllegalArgumentException e) {
			throw new UnexpectedError(
					"Audio format not supported: " + result.getFormat() + ". Expected: " + getTargetFormat(), e);
		}
		return result;
	}

	private AudioFormat getTargetFormat() {
		return new AudioFormat(targetFormatSampleRate, targetFormatSampleSizeInBits, targetFormatChannels,
				targetFormatSigned, targetFormatBigEndian);
	}

	private void feedSoundLine(SourceDataLine soundLine, AudioInputStream audioInputStream, SoundTrack track,
			float durationSeconds) {
		soundLine.start();
		try {
			long startTime = System.currentTimeMillis();
			{
				int nBytesRead = 0;
				int BUFFER_SIZE = bufferSizeMegaBytes * 1024;
				byte[] sampledData = new byte[BUFFER_SIZE];
				while (nBytesRead != -1) {
					if (Thread.currentThread().isInterrupted() || track.requestingIntrerruption) {
						break;
					}
					try {
						nBytesRead = audioInputStream.read(sampledData, 0, sampledData.length);
					} catch (IOException e) {
						throw new UnexpectedError(e);
					}
					if (nBytesRead >= 0) {
						soundLine.write(sampledData, 0, nBytesRead);
					}
				}
			}
			while (!(Thread.currentThread().isInterrupted() || track.requestingIntrerruption)) {
				long elapsedTime = System.currentTimeMillis() - startTime;
				long remainingTime = (long) (durationSeconds * 1000) - elapsedTime;
				if (remainingTime <= 0) {
					break;
				}
				MiscUtils.relieveCPU();
			}
		} finally {
			soundLine.stop();
			soundLine.flush();
		}
	}

	private void runCommandLineClip(Sound soundData) {
		try {
			if (soundData.getFile() == null) {
				File file = FileUtils.createTemporaryFile();
				FileUtils.writeBinary(file, soundData.getBytes(), false);
				soundData.setFile(file);
			}
			final String playCommand = String.format(customDriverPlayCommandPattern, soundData.getFile().getPath());
			CommandExecutor.run(playCommand, true);
		} catch (Exception e) {
			throw new UnexpectedError("Failed to play audio " + soundData + " through command-line '"
					+ customDriverPlayCommandPattern + ": " + e.toString(), e);
		}
	}

	private boolean isClipOpen(SoundTrack track) {
		boolean result = (track.clipRunner != null) && track.clipRunner.isAlive();
		return result;
	}

	private void closeClip(SoundTrack track) {
		if (track.clipRunner != null) {
			track.clipRunner.interrupt();
			try {
				track.clipRunner.join();
			} catch (InterruptedException e) {
				throw new UnexpectedError(e);
			}
		}
	}

	private Sound[] shiftedLeft(Sound[] tab) {
		Sound[] result = new Sound[tab.length];
		for (int i = 0; i < tab.length; i++) {
			result[i] = tab[(i + 1) % tab.length];
		}
		return result;
	}

	public boolean isPlaying(SoundTrack track) {
		return track.rotating || isClipOpen(track);
	}

	public float getDurationSeconds(Sound data) {
		synchronized (soundDurationCacheMutex) {
			Float result = soundDurationCache.get(data);
			if (result == null) {
				try {
					AudioInputStream audioStream = AudioSystem
							.getAudioInputStream(new ByteArrayInputStream(data.getBytes()));
					AudioFormat format = audioStream.getFormat();
					if (!AudioFormat.Encoding.PCM_SIGNED.equals(format.getEncoding())) {
						format = new AudioFormat(format.getSampleRate(), format.getSampleSizeInBits(),
								format.getChannels(), true, format.isBigEndian());
						data = convert(data, format);
					}
					result = (data.getBytes().length / (format.getFrameSize() * format.getFrameRate()));
				} catch (Exception e) {
					throw new UnexpectedError("Failed to compute sound duration: " + e.toString(), e);
				}
				soundDurationCache.put(data, result);
			}
			return result;
		}
	}

	public Sound convert(Sound soundData, AudioFormat newFormat) {
		try {
			AudioInputStream audioStream = AudioSystem
					.getAudioInputStream(new ByteArrayInputStream(soundData.getBytes()));
			AudioInputStream newAudioStream = AudioSystem.getAudioInputStream(newFormat, audioStream);
			File tmpFile = FileUtils.createTemporaryFile();
			try {
				AudioSystem.write(newAudioStream, Type.WAVE, tmpFile);
				return new Sound(soundData.getFile(), FileUtils.readBinary(tmpFile));
			} finally {
				FileUtils.delete(tmpFile);
			}
		} catch (Throwable t) {
			throw new UnexpectedError("Failed to convert sound '" + soundData + "' to format '" + newFormat + "'");
		}
	}

	public Sound join(final Sound... sounds) {
		try {
			long length = 0;
			List<AudioInputStream> inputStreams = new ArrayList<AudioInputStream>();
			for (Sound sound : sounds) {
				inputStreams.add(getStream(sound));
				length += AudioSystem.getAudioInputStream(new ByteArrayInputStream(sound.getBytes())).getFrameLength();
			}
			AudioInputStream appendedStreams = new AudioInputStream(
					new SequenceInputStream(Collections.enumeration(inputStreams)), getTargetFormat(), length);
			ByteArrayOutputStream out = new ByteArrayOutputStream();
			AudioSystem.write(appendedStreams, AudioFileFormat.Type.WAVE, out);
			return new Sound(Sound.getJoinedSoundsVirtualFile(sounds), out.toByteArray());
		} catch (Throwable t) {
			throw new UnexpectedError("Failed to join sounds: " + sounds + ": " + t.toString(), t);
		}
	}
}
