package com.otk.application.audio;

import java.io.File;

import com.otk.application.error.UnexpectedError;
import com.otk.application.util.FileUtils;

/**
 * This class holds the bytes of a sound file in memory for fast convenient
 * access.
 * 
 * @author olitank
 *
 */
public class Sound {

	/**
	 * Will appear in the virtual file path of {@link Sound} instances that were
	 * created by joining other instances.
	 */
	public static final String JOINED_SOUNDS_PATHS_SEPARATOR = "__JOIN__";

	/**
	 * Will appear in the virtual file path of {@link Sound} instances that were
	 * loaded from the class-path.
	 */
	public static final String CLASSPATH_RESOURCE_PREFIX = "__CLASSPATH__";

	/**
	 * A file that will usually (not necessarily) be the path to a sound file
	 * containing the given byte data.
	 */
	private File file;

	/**
	 * The raw data of a sound file.
	 */
	private byte[] bytes;

	/**
	 * Constructs an instance of this class.
	 * 
	 * @param file
	 *            The {@link #file} value.
	 * @param bytes
	 *            The {@link #bytes} value.
	 * 
	 */
	public Sound(File file, byte[] bytes) {
		this.file = file;
		this.bytes = bytes;
	}

	/**
	 * @return {@link #file}
	 */
	public File getFile() {
		return file;
	}

	/**
	 * Update the {@link #file} value.
	 * 
	 * @param file
	 *            The new value.
	 */
	public void setFile(File file) {
		this.file = file;
	}

	/**
	 * @return The {@link #file} value.
	 */
	public byte[] getBytes() {
		return bytes;
	}

	@Override
	public String toString() {
		return "Sound [file=" + file + "]";
	}

	/**
	 * @param file
	 *            The sound file to load.
	 * @return A {@link Sound} instance containing the given sound file data;
	 */
	public static Sound load(File file) {
		return new Sound(file, FileUtils.requireBytes(file));
	}

	/**
	 * @param resourceClass
	 *            A class in the sound resource package.
	 * @param resourceName
	 *            The name of the sound resource.
	 * @return A {@link Sound} instance containing the data of a sound file stored
	 *         as a class-path resource.
	 */
	public static Sound load(Class<?> resourceClass, String resourceName) {
		File file = new File(CLASSPATH_RESOURCE_PREFIX + "/" + resourceClass.getPackage().getName().replace('.', '/')
				+ "/" + resourceName);
		byte[] bytes;
		try {
			bytes = FileUtils.readBinary(resourceClass.getResourceAsStream(resourceName));
		} catch (Exception e) {
			throw new UnexpectedError(e);
		}
		return new Sound(file, bytes);
	}

	/**
	 * @param sounds
	 *            The sounds to join.
	 * @return A virtual file identifying a sound created by joining the given
	 *         sounds.
	 */
	public static File getJoinedSoundsVirtualFile(Sound[] sounds) {
		StringBuilder virtualFilePathBuilder = new StringBuilder();
		for (int i = 0; i < sounds.length; i++) {
			if (i > 0) {
				virtualFilePathBuilder.append(JOINED_SOUNDS_PATHS_SEPARATOR);
			}
			virtualFilePathBuilder.append(sounds[i].getFile().getPath());
		}
		return new File(virtualFilePathBuilder.toString());
	}
}