package com.otk.application.error;

import com.otk.application.util.MiscUtils;

/**
 * This class allows to add debug assertions in the code that can be
 * activated/deactivated at runtime by setting a system property.
 * 
 * @author olitank
 *
 */
public class Assertion {

	/**
	 * Key of the system property used to activate/deactivate all the assertions.
	 */
	public static final String ACTIVATION_PROPERTY_KEY = Assertion.class.getName() + ".active";

	/**
	 * Checks the given predicate: an exception is thrown only and only if the
	 * predicate is false.
	 * 
	 * @param predicate
	 *            the predicate to evaluate.
	 */
	public static void check(Predicate predicate) {
		if (!isActive()) {
			return;
		}
		if (!predicate.isTrue()) {
			throw new UnexpectedError("Assertion failure at:\n" + MiscUtils.getPrintedStackTrace(new Exception()));
		}
		try {
			predicate.doesNotFail();
		} catch (Throwable t) {
			throw new UnexpectedError("Assertion failure at:\n" + MiscUtils.getPrintedStackTrace(t));
		}
	}

	/**
	 * @return Whether the assertions are active (can throw exceptions) or not.
	 */
	protected static boolean isActive() {
		return System.getProperty(ACTIVATION_PROPERTY_KEY, "false").equals("true");
	}

	/**
	 * This represent an assertion predicate.
	 * 
	 * @author olitank
	 *
	 */
	public static abstract class Predicate {
		/**
		 * @return true in order to validate the assertion.
		 */
		protected boolean isTrue() {
			return true;
		}

		/**
		 * Throws an exception to indicate that the assertion is not valid.
		 * 
		 * @throws Throwable
		 *             In order to invalidate the assertion.
		 */
		protected void doesNotFail() throws Throwable {
		}
	}

}
