package com.otk.application.error;

import java.awt.Component;
import java.awt.Image;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;

import com.otk.application.AbstractApplication;
import com.otk.application.translation.T;
import com.otk.application.util.Listener;
import com.otk.application.util.MiscUtils;
import com.otk.application.window.UnexpectedErrorDialog;

public class ErrorManager {

	private Lock exceptionHandlerLock = new ReentrantLock();
	private List<Listener<Throwable>> errorlisteners = new ArrayList<Listener<Throwable>>();
	private AbstractApplication application;
	private boolean displayingError = false;

	public ErrorManager(AbstractApplication application) {
		this.application = application;
		initialize();
	}

	public AbstractApplication getApplication() {
		return application;
	}

	protected void initialize() {
		Thread.setDefaultUncaughtExceptionHandler(new TheUncaughtExceptionHandler());
		TheUncaughtExceptionHandler.DELEGATE = this;
		System.setProperty("sun.awt.exception.handler", TheUncaughtExceptionHandler.class.getName());
	}

	public List<Listener<Throwable>> getErrorlisteners() {
		return errorlisteners;
	}

	protected boolean isBenign(Throwable t) {
		return false;
	}

	public void handleUncaughtException(final Throwable t) {
		if (isBenign(t)) {
			try {
				getApplication().getLog().info("Benign exception skipped: " + MiscUtils.getPrintedStackTrace(t));
				application.getAnalytics().track("Benign Exception",
						t.toString() + "\n" + Arrays.toString(t.getStackTrace()));
			} catch (Throwable ignore) {
			}
			return;
		}
		notifyListeners(t);
		if (!exceptionHandlerLock.tryLock()) {
			getApplication().getLog().error(t);
			return;
		}
		try {
			try {
				getApplication().getLog().error(t);
				application.getAnalytics().track("Uncaught Exception", MiscUtils.getPrintedStackTrace(t));
			} catch (Throwable ignore) {
			}
			SwingUtilities.invokeLater(new Runnable() {
				@Override
				public void run() {
					displayError(t);
				}
			});

		} finally {
			exceptionHandlerLock.unlock();
		}
	}

	public void displayError(Throwable t) {
		if (displayingError) {
			return;
		}
		displayingError = true;
		try {
			if (t instanceof StandardError) {
				showStandardErrorDialog(t);
			} else {
				showUnexpectedErrorDialog(t);
			}
		} finally {
			displayingError = false;
		}
	}

	protected void notifyListeners(Throwable t) {
		for (final Listener<Throwable> l : errorlisteners) {
			new Thread(MiscUtils.formatThreadName(getClass(), "Uncaught Exception Notifier")) {
				{
					setDaemon(true);
				}

				@Override
				public void run() {
					l.handle(t);
				}
			}.start();
		}
	}

	public void showStandardErrorDialog(Throwable t) {
		String text = formatError(t);
		int messageType = JOptionPane.ERROR_MESSAGE;
		Component parent = MiscUtils.getCurrentWindow();
		String title = MessageFormat.format(T.t("Error - {0}"), getApplication().getName());
		Image iconImage = getApplication().getIconImage();
		MiscUtils.showLargeMessageDialog(parent, title, iconImage, text, messageType);
	}

	public void showUnexpectedErrorDialog(Throwable t) {
		UnexpectedErrorDialog.show(application, MiscUtils.getCurrentWindow(), t);
	}

	public String formatError(Throwable t) {
		if (t instanceof OutOfMemoryError) {
			return T.t(
					"The memory is not enough to perform the requested action(s). You may need to restart the software.");
		}
		if (t instanceof NullPointerException) {
			return T.t("An internal error occurred. The system may become unstable.");
		}
		String result = t.toString();
		if ((result == null) || (result.length() == 0)) {
			result = t.getClass().getName();
		}
		String className = t.getClass().getName();
		if (result.contains(className)) {
			String classCaption = MiscUtils.identifierToCaption(t.getClass().getSimpleName());
			if (!classCaption.contains("Error") && classCaption.contains("Exception")) {
				classCaption = classCaption.replace("Exception", "Error");
			}
			result = result.replace(className, classCaption);
		}
		if (t.getCause() != null) {
			result = result.replace(t.getCause().toString(), formatError(t.getCause()));
		}
		result = result.replace("javax", "system");
		result = result.replace("java", "system");
		return result;
	}

	public static void rethrow(Throwable t) {
		throw getToRethrow(t);
	}

	public static RuntimeException getToRethrow(Throwable t) {
		return getToRethrow(null, t, true);
	}

	public static void rethrow(String errorContextDescription, Throwable t) {
		throw getToRethrow(errorContextDescription, t);
	}

	public static RuntimeException getToRethrow(String errorContextDescription, Throwable t) {
		return getToRethrow(errorContextDescription, t, true);
	}

	public static void rethrow(String errorContextDescription, Throwable t, boolean unknownErrorIsUnexpected) {
		throw getToRethrow(errorContextDescription, t, unknownErrorIsUnexpected);
	}

	private static RuntimeException getToRethrow(String errorContextDescription, Throwable t,
			boolean unknownErrorIsUnexpected) {
		if ((t instanceof UnexpectedError) || (!(t instanceof StandardError) && unknownErrorIsUnexpected)) {
			if (errorContextDescription != null) {
				return new UnexpectedError(errorContextDescription + ": " + t.toString(), t);
			} else {
				return new UnexpectedError(t.toString(), t);
			}
		} else {
			if (errorContextDescription != null) {
				return new StandardError(errorContextDescription + ": " + t.toString(), t);
			} else {
				return new StandardError(t.toString(), t);
			}
		}
	}

	public static class TheUncaughtExceptionHandler implements Thread.UncaughtExceptionHandler {

		public static ErrorManager DELEGATE;

		public void handle(Throwable thrown) {
			uncaughtException(null, thrown);
		}

		@Override
		public void uncaughtException(Thread thread, Throwable t) {
			if (DELEGATE != null) {
				DELEGATE.handleUncaughtException(t);
			}
		}

	}

}