package com.otk.application.image;

import java.awt.Point;
import java.awt.image.BufferedImage;
import java.awt.image.DataBufferInt;
import java.awt.image.WritableRaster;
import java.util.ArrayList;
import java.util.List;

import com.otk.application.error.UnexpectedError;

public class FastestPixels {

	public static boolean isRGBConversionRequired(BufferedImage image) {
		return image.getType() != BufferedImage.TYPE_INT_ARGB;
	}

	public static int readPixel(BufferedImage image, int x, int y) {
		if (isRGBConversionRequired(image)) {
			return image.getRGB(x, y);
		} else {
			int[] rgbArray = accessPixels(image);
			int pixelIndex = x + (image.getWidth() * y);
			return rgbArray[pixelIndex];
		}
	}

	public static void writePixel(BufferedImage image, int x, int y, int rgb) {
		if (isRGBConversionRequired(image)) {
			image.setRGB(x, y, rgb);
		} else {
			int[] rgbArray = accessPixels(image);
			int pixelIndex = x + (image.getWidth() * y);
			rgbArray[pixelIndex] = rgb;
		}
	}

	public static int[] readPixels(BufferedImage image) {
		int width = image.getWidth();
		int height = image.getHeight();
		boolean rgbConversionRequired = isRGBConversionRequired(image);
		if (rgbConversionRequired) {
			return image.getRGB(0, 0, width, height, null, 0, width);
		} else {
			int[] rgbArray = accessPixels(image);
			int[] result = new int[rgbArray.length];
			System.arraycopy(rgbArray, 0, result, 0, rgbArray.length);
			return result;
		}

	}

	public static void writePixels(int[] pixels, BufferedImage image) {
		int width = image.getWidth();
		int height = image.getHeight();
		boolean dstRGBConversionRequired = isRGBConversionRequired(image);
		if (dstRGBConversionRequired) {
			image.setRGB(0, 0, width, height, pixels, 0, width);
		} else {
			int[] rgbArray = accessPixels(image);
			System.arraycopy(pixels, 0, rgbArray, 0, pixels.length);
		}
	}

	public static int[] accessPixels(BufferedImage image) {
		if (isRGBConversionRequired(image)) {
			throw new UnexpectedError();
		}
		WritableRaster maskRaster = image.getRaster();
		return ((DataBufferInt) maskRaster.getDataBuffer()).getData();
	}

	public static void scan(BufferedImage srcImage, ScanFilter filter) {
		int width = srcImage.getWidth();
		int height = srcImage.getHeight();
		List<Point> scan = new ArrayList<Point>();
		for (int x = 0; x < width; x++) {
			for (int y = 0; y < height; y++) {
				scan.add(new Point(x, y));
			}
		}
		scan(srcImage, filter, scan);
	}

	public static void scan(BufferedImage srcImage, ScanFilter filter, Iterable<Point> scan) {
		int width = srcImage.getWidth();
		int[] srcPixels;
		if (isRGBConversionRequired(srcImage)) {
			srcPixels = readPixels(srcImage);
		} else {
			srcPixels = accessPixels(srcImage);
		}
		for (Point p : scan) {
			int pixelIndex = p.y * width + p.x;
			filter.handle(p.x, p.y, srcPixels[pixelIndex], srcPixels);
		}
	}

	public static void transfer(BufferedImage srcImage, BufferedImage dstImage, TransferFilter filter) {
		int width = dstImage.getWidth();
		int height = dstImage.getHeight();
		List<Point> scan = new ArrayList<Point>();
		for (int x = 0; x < width; x++) {
			for (int y = 0; y < height; y++) {
				scan.add(new Point(x, y));
			}
		}
		transfer(srcImage, dstImage, filter, scan);
	}

	public static void transfer(BufferedImage srcImage, BufferedImage dstImage, TransferFilter filter,
			Iterable<Point> scan) {
		int width = dstImage.getWidth();
		int height = dstImage.getHeight();
		int[] dstPixels = new int[width * height];
		int[] srcPixels;
		if (isRGBConversionRequired(srcImage)) {
			srcPixels = readPixels(srcImage);
		} else {
			srcPixels = accessPixels(srcImage);
		}
		for (Point p : scan) {
			int pixelIndex = p.y * width + p.x;
			dstPixels[pixelIndex] = filter.get(p.x, p.y, srcPixels[pixelIndex], srcPixels);
		}
		writePixels(dstPixels, dstImage);
	}

	public static void compose(BufferedImage srcImage1, BufferedImage srcImage2, BufferedImage dstImage,
			ComposeFilter filter) {
		compose(new BufferedImage[] { srcImage1, srcImage2 }, dstImage, toComposeNFilter(filter));
	}

	public static void compose(BufferedImage srcImage1, BufferedImage srcImage2, BufferedImage dstImage,
			ComposeFilter filter, Iterable<Point> scan) {
		compose(new BufferedImage[] { srcImage1, srcImage2 }, dstImage, toComposeNFilter(filter), scan);
	}

	public static void compose(BufferedImage[] srcImages, BufferedImage dstImage, ComposeNFilter filter) {
		int width = dstImage.getWidth();
		int height = dstImage.getHeight();
		List<Point> scan = new ArrayList<Point>();
		for (int x = 0; x < width; x++) {
			for (int y = 0; y < height; y++) {
				scan.add(new Point(x, y));
			}
		}
		compose(srcImages, dstImage, filter, scan);
	}

	public static void compose(BufferedImage[] srcImages, BufferedImage dstImage, ComposeNFilter filter,
			Iterable<Point> scan) {
		int width = dstImage.getWidth();
		int height = dstImage.getHeight();
		int[][] allSrcPixels = new int[srcImages.length][];
		for (int i = 0; i < srcImages.length; i++) {
			if (isRGBConversionRequired(srcImages[i])) {
				allSrcPixels[i] = readPixels(srcImages[i]);
			} else {
				allSrcPixels[i] = accessPixels(srcImages[i]);
			}
		}
		int[] dstPixels = new int[width * height];
		for (Point p : scan) {
			int pixelIndex = p.y * width + p.x;
			int[] currentSrcPixels = new int[srcImages.length];
			for (int i = 0; i < srcImages.length; i++) {
				currentSrcPixels[i] = allSrcPixels[i][pixelIndex];
			}
			dstPixels[pixelIndex] = filter.get(p.x, p.y, currentSrcPixels, allSrcPixels);
		}
		writePixels(dstPixels, dstImage);
	}

	public static ComposeNFilter toComposeNFilter(ComposeFilter filter) {
		return new ComposeNFilter() {
			@Override
			public int get(int x, int y, int[] currentSrcPixels, int[][] allSrcPixels) {
				return filter.get(x, y, currentSrcPixels[0], currentSrcPixels[1], allSrcPixels[0], allSrcPixels[1]);
			}
		};
	}

	public interface ScanFilter {
		void handle(int x, int y, int currentSrcPixel, int[] srcPixels);
	}

	public interface TransferFilter {
		int get(int x, int y, int currentSrcPixel, int[] srcPixels);
	}

	public interface ComposeFilter {
		int get(int x, int y, int currentSrcPixel1, int currentSrcPixel2, int[] srcPixels1, int[] srcPixels2);
	}

	public interface ComposeNFilter {
		int get(int x, int y, int[] currentSrcPixels, int[][] allSrcPixels);
	}

}