package com.otk.application.image.camera;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.image.BufferedImage;
import java.text.DecimalFormat;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.ThreadFactory;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.BorderFactory;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.OverlayLayout;

import com.otk.application.audio.AudioEngine;
import com.otk.application.audio.Sound;
import com.otk.application.audio.SoundTrack;
import com.otk.application.error.Assertion;
import com.otk.application.error.ErrorManager;
import com.otk.application.error.StandardError;
import com.otk.application.image.filter.AbstractFilter;
import com.otk.application.image.filter.AbstractHeadlessFilter;
import com.otk.application.image.filter.FilteringContext;
import com.otk.application.service.AbstractService;
import com.otk.application.translation.T;
import com.otk.application.util.Accessor;
import com.otk.application.util.ImageUtils;
import com.otk.application.util.Listener;
import com.otk.application.util.MathUtils;
import com.otk.application.util.MiscUtils;
import com.otk.application.util.TextUtils;
import com.otk.application.util.draw.DropShadow;
import com.otk.application.window.component.ImagePanel;
import com.otk.application.window.component.TransparentPanel;

public class Camera extends AbstractService {

	public static void main(String[] args) throws Throwable {
		Camera camera = new Camera() {
			public List<AbstractCameraDriver> createDrivers() {
				List<AbstractCameraDriver> result = new ArrayList<AbstractCameraDriver>();
				result.add(new NOCamDriver(this));
				result.add(new OpenIMAJDriver(this));
				return result;
			}
		};
		List<String> deviceNames = camera.listAvailableDevices(null);
		camera.setDeviceFullName(deviceNames.get(deviceNames.size() - 1));
		camera.start();
		camera.getVideoComponent().setPreferredSize(new Dimension(800, 600));
		JOptionPane.showMessageDialog(null, camera.getVideoComponent());
		camera.stop();
	}

	public static final int IDEAL_FPS = 24;

	private static final String DRIVER_AND_DEVICE_NAME_PATTERN = "\\[([^\\]]+)\\] (.+)";
	private static final String DRIVER_AND_DEVICE_NAME_FORMAT = "[{0}] {1}";

	public static final Sound CAMERA_FOCUS_SOUND = Sound.load(Camera.class, "cameraFocus.wav");
	public static final Sound CAMERA_CLICK_SOUND = Sound.load(Camera.class, "cameraClick.wav");

	protected static final String EXCLUDED_DEVICES_PATTERN_PROPERTY_KEY = Camera.class.getName() + ".deviceExclusionPattern";

	private AbstractCameraDriver activeDriver;
	private FrameFormat lastVideoFrameFormat;
	private String lastVideoFrameFormatId;
	private FrameFormat lastPhotoFrameFormat;
	private String lastPhotoFrameFormatId;
	private BufferedImage lastImageFromDevice;
	private BufferedImage currentImage;
	private BufferedImage displayedImage;
	private Object currentlyDisplayedImageMutex = new Object();
	private TransparentPanel videoComponent;
	private ImagePanel canvas;
	private AbstractFilter displayFilter;
	private Map<String, List<FrameFormat>> videoFormatsByDevice = new HashMap<String, List<FrameFormat>>();
	private Map<String, List<FrameFormat>> photoFormatsByDevice = new HashMap<String, List<FrameFormat>>();
	private boolean videoComponentImageFrozen = false;
	private JLabel fpsLabel;
	private float fps = 0;
	private String fpsText = "";
	private boolean fpsLabelEnabled = false;
	private boolean fpsUpdateEnabled = true;
	private long lastImageDisplayTime;
	private int frameCountSinceStarted = 0;
	private FrameProcessor frameProcessor;
	private boolean asynchronouslyExecuting = false;
	private ExecutorService asynchronousExecutor = Executors.newSingleThreadExecutor(new ThreadFactory() {
		@Override
		public Thread newThread(final Runnable r) {
			Thread result = new Thread(new Runnable() {
				@Override
				public void run() {
					asynchronouslyExecuting = true;
					try {
						r.run();
					} finally {
						asynchronouslyExecuting = false;
					}

				}
			}, MiscUtils.formatThreadName(Camera.this.getClass(), "asynchronousExecutor"));
			result.setDaemon(true);
			return result;
		}
	});
	private List<AbstractCameraDriver> drivers;

	private AudioEngine audioEngine = new AudioEngine();
	private boolean dropShadowBorderAddedToVideoFrame = true;
	private boolean videoImageSmoothnessImproved = false;
	private String deviceFullName = "";
	private List<AbstractFilter> videoInputCorrectionFilters = Collections.emptyList();
	private List<AbstractFilter> photoSpecificInputCorrectionFilters = Collections.emptyList();
	private boolean photoAutoMatchingVideoAspectRatio = false;
	private boolean concurrentFrameProcessing = false;
	private FrameFormat videoFormat = new FrameFormat(320, 240);
	private FrameFormat photoSpecificFormat = new FrameFormat(640, 480);
	private String threadPriority = "NORM_PRIORITY";
	private int freezeTimeoutMilliSeconds = 30000;
	private long webcamAfterShootingPauseDurationMilliseconds = 1000;
	private int webcamAutoWhiteBalanceStabilizationDurationMilliseconds = 2000;
	private int brightnessCompensation = 0;
	private int contrastCompensation = 0;

	public Camera() {
		drivers = createDrivers();
	}

	@Override
	protected String getServiceName() {
		return "Camera";
	}

	public int getBrightnessCompensation() {
		return brightnessCompensation;
	}

	public void setBrightnessCompensation(int brightnessCompensation) {
		this.brightnessCompensation = brightnessCompensation;
		driverSettingsUpdated();
	}

	protected void driverSettingsUpdated() {
		if (activeDriver != null) {
			activeDriver.settingsUpdated();
		}
	}

	public int getContrastCompensation() {
		return contrastCompensation;
	}

	public void setContrastCompensation(int contrastCompensation) {
		this.contrastCompensation = contrastCompensation;
		driverSettingsUpdated();
	}

	public long getWebcamAfterShootingPauseDurationMilliseconds() {
		return webcamAfterShootingPauseDurationMilliseconds;
	}

	public void setWebcamAfterShootingPauseDurationMilliseconds(long webcamAfterShootingPauseDurationMilliseconds) {
		this.webcamAfterShootingPauseDurationMilliseconds = webcamAfterShootingPauseDurationMilliseconds;
		driverSettingsUpdated();
	}

	public int getWebcamAutoWhiteBalanceStabilizationDurationMilliseconds() {
		return webcamAutoWhiteBalanceStabilizationDurationMilliseconds;
	}

	public void setWebcamAutoWhiteBalanceStabilizationDurationMilliseconds(
			int webcamAutoWhiteBalanceStabilizationDurationMilliseconds) {
		this.webcamAutoWhiteBalanceStabilizationDurationMilliseconds = webcamAutoWhiteBalanceStabilizationDurationMilliseconds;
		driverSettingsUpdated();
	}

	public int getFreezeTimeoutMilliSeconds() {
		return freezeTimeoutMilliSeconds;
	}

	public void setFreezeTimeoutMilliSeconds(int freezeTimeoutMilliSeconds) {
		this.freezeTimeoutMilliSeconds = freezeTimeoutMilliSeconds;
		driverSettingsUpdated();
	}

	public AudioEngine getAudioEngine() {
		return audioEngine;
	}

	public void setAudioEngine(AudioEngine audioEngine) {
		this.audioEngine = audioEngine;
	}

	public boolean isDropShadowBorderAddedToVideoFrame() {
		return this.dropShadowBorderAddedToVideoFrame;
	}

	public void setDropShadowBorderAddedToVideoFrame(boolean dropShadowBorderAddedToVideoFrame) {
		this.dropShadowBorderAddedToVideoFrame = dropShadowBorderAddedToVideoFrame;
	}

	public boolean isVideoImageSmoothnessImproved() {
		return this.videoImageSmoothnessImproved;
	}

	public void setVideoImageSmoothnessImproved(boolean videoImageSmoothnessImproved) {
		this.videoImageSmoothnessImproved = videoImageSmoothnessImproved;
	}

	public String getDeviceFullName() {
		return this.deviceFullName;
	}

	public void setDeviceFullName(String deviceFullName) {
		this.deviceFullName = deviceFullName;
		driverSettingsUpdated();
	}

	public List<AbstractFilter> getVideoInputCorrectionFilters() {
		return this.videoInputCorrectionFilters;
	}

	public void setVideoInputCorrectionFilters(List<AbstractFilter> videoInputCorrectionFilters) {
		this.videoInputCorrectionFilters = videoInputCorrectionFilters;
	}

	public List<AbstractFilter> getPhotoSpecificInputCorrectionFilters() {
		return this.photoSpecificInputCorrectionFilters;
	}

	public void setPhotoSpecificInputCorrectionFilters(List<AbstractFilter> photoSpecificInputCorrectionFilters) {
		this.photoSpecificInputCorrectionFilters = photoSpecificInputCorrectionFilters;
	}

	public boolean isPhotoAutoMatchingVideoAspectRatio() {
		return this.photoAutoMatchingVideoAspectRatio;
	}

	public void setPhotoAutoMatchingVideoAspectRatio(boolean photoAutoMatchingVideoAspectRatio) {
		this.photoAutoMatchingVideoAspectRatio = photoAutoMatchingVideoAspectRatio;
	}

	public boolean isConcurrentFrameProcessing() {
		return this.concurrentFrameProcessing;
	}

	public void setConcurrentFrameProcessing(boolean concurrentFrameProcessing) {
		this.concurrentFrameProcessing = concurrentFrameProcessing;
	}

	public FrameFormat getVideoFormat() {
		return this.videoFormat;
	}

	public void setVideoFormat(FrameFormat videoFormat) {
		this.videoFormat = videoFormat;
		driverSettingsUpdated();
	}

	public FrameFormat getPhotoSpecificFormat() {
		return this.photoSpecificFormat;
	}

	public void setPhotoSpecificFormat(FrameFormat photoSpecificFormat) {
		this.photoSpecificFormat = photoSpecificFormat;
		driverSettingsUpdated();
	}

	public FrameFormat getPhotoFormat() {
		if (getPhotoSpecificFormat() != null) {
			return getPhotoSpecificFormat();
		} else {
			if (activeDriver != null) {
				return activeDriver.getDefaultPhotoFormat();
			} else {
				return getVideoFormat();
			}
		}
	}

	public String getThreadPriority() {
		return this.threadPriority;
	}

	public void setThreadPriority(String threadPriority) {
		this.threadPriority = threadPriority;
		driverSettingsUpdated();
	}

	public float getFps() {
		return fps;
	}

	protected ImagePanel createCanvas() {
		return new ImagePanel() {

			private static final long serialVersionUID = 1L;

			@Override
			protected void paintImage(Graphics g, Image imageToPaint) {
				if (isDropShadowBorderAddedToVideoFrame()) {
					Rectangle dstBounds = super.getImageDestinationBounds(imageToPaint);
					g.drawImage(DropShadow.FRAME, dstBounds.x, dstBounds.y, dstBounds.width, dstBounds.height, null);
					super.paintImage(g, imageToPaint);
				} else {
					super.paintImage(g, imageToPaint);
				}
			}

			@Override
			protected Rectangle getImageDestinationBounds(Image imageToPaint) {
				if (isDropShadowBorderAddedToVideoFrame()) {
					Rectangle result = super.getImageDestinationBounds(imageToPaint);
					result = DropShadow.getInnerArea(result);
					return result;
				} else {
					return super.getImageDestinationBounds(imageToPaint);
				}
			}

		};
	}

	public void logInfo(String s) {
		System.out.println(s);
	}

	public void logError(String s) {
		System.err.println(s);
	}

	public int getFrameCountSinceStarted() {
		return frameCountSinceStarted;
	}

	public boolean isFpsUpdateEnabled() {
		return fpsUpdateEnabled;
	}

	public void setFpsUpdateEnabled(boolean fpsUpdateEnabled) {
		this.fpsUpdateEnabled = fpsUpdateEnabled;
		fps = 0.0f;
	}

	public boolean isFpsLabelEnabled() {
		return fpsLabelEnabled;
	}

	public void setFpsLabelEnabled(boolean fpsLabelEnabled) {
		this.fpsLabelEnabled = fpsLabelEnabled;
	}

	public AbstractFilter getFpsDisplayFilter() {
		return new AbstractHeadlessFilter() {

			@Override
			public FilteringContext doExecute(FilteringContext context) {
				if (fpsText.length() > 0) {
					Graphics2D g2d = context.getImage().createGraphics();
					Image fpsTextImage = ImageUtils.getTextImage(fpsText, 50, Color.GREEN);
					fpsTextImage = ImageUtils.scalePreservingRatio(fpsTextImage, context.getImage().getWidth(),
							context.getImage().getHeight() / 4, false, false);
					g2d.drawImage(fpsTextImage, 0, 0, null);
					g2d.dispose();
				}
				return context;
			}
		};
	}

	protected void updateFps() {
		if (!fpsUpdateEnabled) {
			fps = 0.0f;
			fpsText = "";
		} else {
			long currentTime = System.currentTimeMillis();
			long sinceLastImage = currentTime - lastImageDisplayTime;
			lastImageDisplayTime = currentTime;
			float instantFps = 1.0f / (Math.max(1, sinceLastImage) / 1000.0f);
			if (fps == 0.0) {
				fps = instantFps;
			} else {
				float INSTANT_FPS_IMPACT_RATIO = 0.1f;
				fps = ((1 - INSTANT_FPS_IMPACT_RATIO) * fps) + (INSTANT_FPS_IMPACT_RATIO * instantFps);
			}
			if (fps >= 1000.0f) {
				fps = 1000.0f;
				fpsText = "FPS>999";
			} else {
				fpsText = "FPS=" + new DecimalFormat("000").format(fps);
			}
		}
	}

	public BufferedImage getCurrentImage() {
		return currentImage;
	}

	public void setCurrentImage(BufferedImage image) {
		synchronized (currentlyDisplayedImageMutex) {
			currentImage = image;
			if (image != null) {
				image = correctVideo(image);
				if (displayFilter != null) {
					image = displayFilter.execute(new FilteringContext().withImage(image)).getImage();
				}
			}
			displayedImage = image;
			if (videoComponent != null) {
				if (isVideoImageSmoothnessImproved() != canvas.isScalingQualitHigh()) {
					canvas.setScalingQualitHigh(isVideoImageSmoothnessImproved());
				}
				canvas.setImage(image);
				if (fpsLabel.isDisplayable()) {
					if (fpsLabelEnabled) {
						boolean fpsValueAvailable = fpsText.length() > 0;
						if (fpsLabel.isVisible() || fpsValueAvailable) {
							fpsLabel.setVisible(fpsValueAvailable);
							fpsLabel.setText(fpsText);
							videoComponent.validate();
						}
					} else {
						if (fpsLabel.isVisible()) {
							fpsLabel.setVisible(false);
						}
					}
				}
			}
		}
	}

	public BufferedImage getDisplayedImage() {
		return displayedImage;
	}

	public float getVideoComponentOpacity() {
		if (videoComponent == null) {
			return -1;
		}
		return videoComponent.getOpacity();
	}

	public void setVideoComponentOpacity(float opacity) {
		if (videoComponent == null) {
			return;
		}
		if (videoComponentImageFrozen) {
			return;
		}
		videoComponent.setOpacity(opacity);
	}

	public boolean isVideoComponentImageFrozen() {
		return videoComponentImageFrozen;
	}

	public void setVideoComponentImageFrozen(boolean videoComponentImageFrozen) {
		this.videoComponentImageFrozen = videoComponentImageFrozen;
	}

	@Override
	protected void handleInconsistentState() {
		super.handleInconsistentState();
	}

	public BufferedImage getCurrentRawImage() {
		return lastImageFromDevice;
	}

	public Component getVideoComponent() {
		if (videoComponent == null) {
			videoComponent = new TransparentPanel();
			videoComponent.setLayout(new OverlayLayout(videoComponent));
			fpsLabel = new JLabel();
			{
				videoComponent.add(fpsLabel);
				fpsLabel.setOpaque(true);
				fpsLabel.setBackground(Color.BLACK);
				fpsLabel.setForeground(Color.GREEN);
				fpsLabel.setBorder(BorderFactory.createLineBorder(Color.GREEN));
			}
			canvas = createCanvas();
			{
				videoComponent.add(canvas);
				canvas.setPreservingRatio(true);
				canvas.setDoubleBuffered(true);
			}
		}
		return videoComponent;
	}

	public void setDisplayFilter(AbstractFilter filter) {
		this.displayFilter = filter;
	}

	public AbstractFilter getDisplayFilter() {
		return displayFilter;
	}

	protected List<FrameFormat> getFormats(Accessor<List<FrameFormat>> activeDriverFormatsAccessor,
			Map<String, List<FrameFormat>> formatCache) {
		String deviceName = isSetup() ? formatDeviceFullName(activeDriver, extractLocalDeviceName(getDeviceFullName()))
				: getDeviceFullName();
		List<FrameFormat> result = formatCache.get(deviceName);
		if (result != null) {
			return result;
		}
		result = new ArrayList<FrameFormat>();
		synchronized (getLifeCycleMutex()) {
			if (isSetup()) {
				result.addAll(activeDriverFormatsAccessor.get());
			} else {
				try {
					setUp();
				} catch (Exception e) {
					return Collections.emptyList();
				}
				result.addAll(activeDriverFormatsAccessor.get());
				try {
					cleanUp();
				} catch (Exception ignore) {
					handleInconsistentState();
				}
			}
		}
		formatCache.put(deviceName, result);
		return result;
	}

	public List<FrameFormat> getVideoFormats() {
		Accessor<List<FrameFormat>> activeDriverFormatsAccessor = new Accessor<List<FrameFormat>>() {

			@Override
			public List<FrameFormat> get() {
				logInfo("Enumerating Video Formats");
				return activeDriver.getVideoFormatsWhileCameraInitializedAndNotActive();
			}
		};
		return getFormats(activeDriverFormatsAccessor, videoFormatsByDevice);
	}

	public List<FrameFormat> getPhotoFormats() {
		Accessor<List<FrameFormat>> activeDriverFormatsAccessor = new Accessor<List<FrameFormat>>() {

			@Override
			public List<FrameFormat> get() {
				logInfo("Enumerating Photo Formats");
				return activeDriver.getPhotoFormatsWhileCameraInitializedAndNotActive();
			}
		};
		return getFormats(activeDriverFormatsAccessor, photoFormatsByDevice);
	}

	public List<String> listAvailableDevices(Listener<Throwable> errorListener) {
		logInfo("Enumerating camera devices");
		List<String> result = new ArrayList<String>();
		synchronized (getLifeCycleMutex()) {
			withTemporaryInterruptionIfActive(new Runnable() {
				@Override
				public void run() {
					String excludedDevicePattern = System.getProperty(EXCLUDED_DEVICES_PATTERN_PROPERTY_KEY);
					for (AbstractCameraDriver driver : drivers) {
						logInfo("Enumerating  '" + driver.getName() + "' driver devices");
						try {
							for (String deviceLocalName : driver.listDeviceLocalNamesWhileCameraNotInitialized()) {
								String deviceFullName = formatDeviceFullName(driver, deviceLocalName);
								if (excludedDevicePattern != null) {
									if (deviceFullName.matches(excludedDevicePattern)) {
										continue;
									}
								}
								result.add(deviceFullName);
							}
						} catch (Throwable t) {
							if (errorListener != null) {
								errorListener.handle(ErrorManager.getToRethrow(MessageFormat
										.format(T.t("{0} device list enumeration failure"), driver.getName()), t));
							}
						}
					}
				}
			});
		}
		return result;
	}

	public boolean isAsynchronouslyExecuting() {
		return asynchronouslyExecuting;
	}

	public Future<?> startAsynchronously() {
		Runnable task = new Runnable() {
			@Override
			public void run() {
				start();
			}
		};
		Future<?> future = asynchronousExecutor.submit(task);
		return future;
	}

	public Future<?> stopAsynchronously() {
		Runnable task = new Runnable() {
			@Override
			public void run() {
				stop();
			}
		};
		Future<?> future = asynchronousExecutor.submit(task);
		return future;
	}

	public Future<BufferedImage> shootAsynchronously() {
		Callable<BufferedImage> task = new Callable<BufferedImage>() {
			@Override
			public BufferedImage call() throws Exception {
				return shoot();
			}
		};
		Future<BufferedImage> future = asynchronousExecutor.submit(task);
		return future;
	}

	public BufferedImage shoot() {
		synchronized (getLifeCycleMutex()) {
			logInfo("Capturing camera single image");
			BufferedImage result;
			boolean wasInitialized = initialized;
			boolean wasActive = isActive();
			if (wasActive && activeDriver.canShootWhileCameraActive()) {
				checkFormat(getPhotoFormat(), getPhotoFormats());
				audioEngine.play(SoundTrack.DEFAULT, CAMERA_FOCUS_SOUND, true);
				result = activeDriver.doShootWhileCameraActive();
			} else {
				setupShootingWhileCameraInitializedAndNotActive(wasInitialized, wasActive);
				checkFormat(getPhotoFormat(), getPhotoFormats());
				try {
					audioEngine.play(SoundTrack.DEFAULT, CAMERA_FOCUS_SOUND, false);
					result = activeDriver.doShootWhileCameraInitializedAndNotActive();
					cleanupShootingWhileCameraInitializedAndNotActive(wasInitialized, wasActive);
				} catch (Exception e) {
					try {
						cleanupShootingWhileCameraInitializedAndNotActive(wasInitialized, wasActive);
					} catch (Exception ignore) {
					}
					throw ErrorManager.getToRethrow("Shooting error", e);
				}
			}
			result = correctPhoto(result);
			return result;
		}
	}

	protected void checkFormat(FrameFormat format, List<FrameFormat> formats) {
		if (!formats.contains(format)) {
			throw new StandardError("Format not supported: " + format + ".\n"
					+ "Restart the software if there was a hardware modification.");
		}
	}

	protected void setupShootingWhileCameraInitializedAndNotActive(boolean wasInitialized, boolean wasActive) {
		Assertion.check(new Assertion.Predicate() {
			@Override
			protected boolean isTrue() {
				return (!wasActive) || wasInitialized;
			}
		});
		try {
			if (!wasInitialized) {
				setUp();
				initialized = true;
			}
			if (wasActive) {
				triggerInterruption();
				waitForCompleteInterruption();
				active = false;
			}
		} catch (Exception e) {
			try {
				triggerInterruption();
				waitForCompleteInterruption();
				active = false;
			} catch (Exception ignore) {
			}
			try {
				cleanUp();
				initialized = false;
			} catch (Exception ignore) {
			}
			handleInconsistentState();
			throw new StandardError(getServiceName() + " shooting initialization error: " + e.toString(), e);
		}
	}

	protected void cleanupShootingWhileCameraInitializedAndNotActive(boolean wasInitialized, boolean wasActive) {
		try {
			if (!wasInitialized) {
				cleanUp();
				initialized = false;
			}
			if (wasActive) {
				triggerActivation();
				active = true;
			}
		} catch (Exception e) {
			try {
				triggerInterruption();
				waitForCompleteInterruption();
				active = false;
			} catch (Exception ignore) {
			}
			try {
				cleanUp();
				initialized = false;
			} catch (Exception ignore) {
			}
			handleInconsistentState();
			throw new StandardError(getServiceName() + " shooting finalization error: " + e.toString(), e);
		}
	}

	public BufferedImage correctVideo(BufferedImage bufferedImage) {
		FrameFormat videoFormat = getVideoFormat();
		if ((bufferedImage.getWidth() != videoFormat.width) || (bufferedImage.getHeight() != videoFormat.height)) {
			bufferedImage = ImageUtils.getBufferedImage(
					bufferedImage.getScaledInstance(videoFormat.width, videoFormat.height, Image.SCALE_FAST));
		}
		BufferedImage result = ImageUtils
				.filter(new FilteringContext().withImage(bufferedImage), getVideoInputCorrectionFilters()).getImage();
		return result;
	}

	public BufferedImage correctPhoto(BufferedImage bufferedImage) {
		FrameFormat photoFormat = getPhotoFormat();
		if ((bufferedImage.getWidth() != photoFormat.width) || (bufferedImage.getHeight() != photoFormat.height)) {
			bufferedImage = ImageUtils.getBufferedImage(
					bufferedImage.getScaledInstance(photoFormat.width, photoFormat.height, Image.SCALE_SMOOTH));
		}
		BufferedImage result = ImageUtils
				.filter(new FilteringContext().withImage(bufferedImage), getPhotoInputCorrectionFilters()).getImage();
		if (isPhotoAutoMatchingVideoAspectRatio()) {
			FrameFormat videoFormat = getCorrectedVideoFormat();
			double videoFrameRatio = (double) videoFormat.getWidth() / (double) videoFormat.getHeight();
			double currentResultRatio = (double) result.getWidth() / (double) result.getHeight();
			if (videoFrameRatio != currentResultRatio) {
				int newResultHeight = MathUtils.round(result.getWidth() / videoFrameRatio);
				result = ImageUtils.getBufferedImage(
						result.getScaledInstance(result.getWidth(), newResultHeight, Image.SCALE_SMOOTH));
			}
		}
		return result;
	}

	public List<AbstractFilter> getPhotoInputCorrectionFilters() {
		if (getPhotoSpecificInputCorrectionFilters() != null) {
			return getPhotoSpecificInputCorrectionFilters();
		} else {
			return getVideoInputCorrectionFilters();
		}
	}

	public void handleNewImageFromDevice(final BufferedImage newImage) {
		lastImageFromDevice = newImage;
		frameProcessor.handleNewImageFromDevice();
	}

	@Override
	protected void triggerActivation() throws Exception {
		checkFormat(getVideoFormat(), getVideoFormats());
		activeDriver.startCapture();
	}

	@Override
	protected void triggerInterruption() throws Exception {
		activeDriver.stopCapture();
	}

	@Override
	protected void waitForCompleteInterruption() {
		activeDriver.doWaitForCompleteInterruption();
	}

	@Override
	protected void setUp() throws Exception {
		logInfo("Connecting to camera device '" + getDeviceFullName() + "'");
		activeDriver = getDriver();
		activeDriver.settingsUpdated();
		activeDriver.doSetUp();
		frameCountSinceStarted = 0;
		setVideoComponentOpacity(0.0f);
		if (isConcurrentFrameProcessing()) {
			frameProcessor = new ConcurrentFrameProcessor();
		} else {
			frameProcessor = new SynchronousFrameProcessor();
		}
		frameProcessor.begin();
	}

	@Override
	protected void cleanUp() throws Exception {
		logInfo("Disonnecting from camera device '" + getDeviceFullName() + "'");
		frameProcessor.end();
		if (!videoComponentImageFrozen) {
			setCurrentImage(null);
		}
		setVideoComponentOpacity(0.0f);
		activeDriver.doCleanUp();
		activeDriver = null;
		lastImageFromDevice = null;
		frameProcessor = null;
	}

	@Override
	public void handleRunErrorAndStop(Throwable t) {
		super.handleRunErrorAndStop(t);
	}

	public void withoutInputCorrection(Runnable runnable) {
		List<AbstractFilter> photoCorrectionFiltersToRestore = getPhotoSpecificInputCorrectionFilters();
		List<AbstractFilter> videoCorrectionFiltersToRestore = getVideoInputCorrectionFilters();
		boolean autoMatchPhotoToVideoFormatRatioToRestore = isPhotoAutoMatchingVideoAspectRatio();
		setPhotoSpecificInputCorrectionFilters(Collections.emptyList());
		setVideoInputCorrectionFilters(Collections.emptyList());
		setPhotoAutoMatchingVideoAspectRatio(false);
		try {
			runnable.run();
		} finally {
			setPhotoSpecificInputCorrectionFilters(photoCorrectionFiltersToRestore);
			setVideoInputCorrectionFilters(videoCorrectionFiltersToRestore);
			setPhotoAutoMatchingVideoAspectRatio(autoMatchPhotoToVideoFormatRatioToRestore);
		}
	}

	public void withThisTemporaryVideoFormat(int width, int height, Runnable runnable) {
		FrameFormat videoFormatToRestore = getVideoFormat();
		setVideoFormat(new FrameFormat(width, height));
		try {
			runnable.run();
		} finally {
			setVideoFormat(videoFormatToRestore);
		}
	}

	public FrameFormat getCorrectedVideoFormat() {
		String videoFrameFormatId = getVideoFormat() + "\ncorrected by:\n"
				+ TextUtils.stringJoin(getVideoInputCorrectionFilters(), "; ");
		if (videoFrameFormatId.equals(lastVideoFrameFormatId)) {
			return lastVideoFrameFormat;
		}
		FrameFormat result;
		BufferedImage sample = new BufferedImage(getVideoFormat().getWidth(), getVideoFormat().getHeight(),
				ImageUtils.getAdaptedBufferedImageType());
		BufferedImage sampleCorrected = correctVideo(sample);
		result = new FrameFormat(sampleCorrected.getWidth(), sampleCorrected.getHeight());
		lastVideoFrameFormat = result;
		lastVideoFrameFormatId = videoFrameFormatId;
		return result;
	}

	public FrameFormat getCorrectedPhotoFormat() {
		String photoFrameSizeId = getPhotoFormat() + "\ncorrected by:\n"
				+ TextUtils.stringJoin(getPhotoInputCorrectionFilters(), "; ") + "\nautoMatchVideoFormatRatio:"
				+ isPhotoAutoMatchingVideoAspectRatio();
		if (photoFrameSizeId.equals(lastPhotoFrameFormatId)) {
			return lastPhotoFrameFormat;
		}
		BufferedImage sample = new BufferedImage(getPhotoFormat().getWidth(), getPhotoFormat().getHeight(),
				ImageUtils.getAdaptedBufferedImageType());
		BufferedImage sampleCorrected = correctPhoto(sample);
		FrameFormat result = lastPhotoFrameFormat = new FrameFormat(sampleCorrected.getWidth(),
				sampleCorrected.getHeight());
		lastPhotoFrameFormatId = photoFrameSizeId;
		return result;
	}

	public void scheduleFadeInEffect() {
		setVideoComponentOpacity(0.0f);
	}

	public List<AbstractCameraDriver> createDrivers() {
		List<AbstractCameraDriver> result = new ArrayList<AbstractCameraDriver>();
		result.add(new NOCamDriver(this));
		return result;
	}

	public String getDriverName() {
		return extractDriverName(getDeviceFullName());
	}

	public String getLocalDeviceName() {
		return extractLocalDeviceName(getDeviceFullName());
	}

	public List<AbstractCameraDriver> getDrivers() {
		return drivers;
	}

	public AbstractCameraDriver getDriver() {
		if ("".equals(getDeviceFullName())) {
			throw new StandardError("Camera device not selected!");
		}
		String driverName = getDriverName();
		if (driverName == null) {
			throw new StandardError("Invalid camera device name: '" + getDeviceFullName() + "'");
		}
		for (AbstractCameraDriver driver : drivers) {
			if (driver.getName().equals(driverName)) {
				return driver;
			}
		}
		throw new StandardError("Invalid camera driver name: '" + driverName + "'");
	}

	public <D extends AbstractCameraDriver> D findDriver(Class<D> driverClass) {
		for (AbstractCameraDriver driver : drivers) {
			if (driverClass.isInstance(driver)) {
				return driverClass.cast(driver);
			}
		}
		return null;
	}

	public static String extractDriverName(String deviceFullName) {
		Matcher matcher = Pattern.compile(DRIVER_AND_DEVICE_NAME_PATTERN).matcher(deviceFullName);
		if (matcher.matches()) {
			String result = matcher.group(1);
			if (result.length() > 0) {
				return result;
			}
		}
		return null;
	}

	public static String formatDeviceFullName(AbstractCameraDriver driver, String deviceLocalName) {
		return MessageFormat.format(DRIVER_AND_DEVICE_NAME_FORMAT, driver.getName(), deviceLocalName);
	}

	protected static String extractLocalDeviceName(String deviceFullName) {
		Matcher matcher = Pattern.compile(DRIVER_AND_DEVICE_NAME_PATTERN).matcher(deviceFullName);
		if (!matcher.matches()) {
			return "";
		}
		String result = matcher.group(2);
		return result;
	}

	protected interface FrameProcessor {

		void begin();

		void handleNewImageFromDevice();

		void end();

	}

	protected class SynchronousFrameProcessor implements FrameProcessor {

		@Override
		public void begin() {
		}

		@Override
		public void handleNewImageFromDevice() {
			BufferedImage imageFromDevice = lastImageFromDevice;
			if (imageFromDevice == null) {
				return;
			}
			displayNewImageFromDevice(imageFromDevice);
		}

		private void displayNewImageFromDevice(BufferedImage imageFromDevice) {
			updateFps();
			if (!videoComponentImageFrozen) {
				setCurrentImage(imageFromDevice);
				frameCountSinceStarted++;
			}
			if (getVideoComponentOpacity() < 1.0) {
				setVideoComponentOpacity(Math.min(1.0f, getVideoComponentOpacity() + 0.34f));
			}
		}

		@Override
		public void end() {
		}

	}

	protected class ConcurrentFrameProcessor extends Thread implements FrameProcessor {
		private BufferedImage lastDisplayedImage;
		private Object newImageFromDeviceMonitor = new Object();
		private SynchronousFrameProcessor imageDisplayUtil = new SynchronousFrameProcessor();

		public ConcurrentFrameProcessor() {
			super(MiscUtils.formatThreadName(Camera.this.getClass(), "Frame Processor"));
			setDaemon(true);
			setPriority((Integer) MiscUtils.getFieldValue(Thread.class, Camera.this.getThreadPriority(), null));
		}

		@Override
		public void handleNewImageFromDevice() {
			notifyNewImageFromDevice();
		}

		@Override
		public void begin() {
			start();
		}

		@Override
		public void end() {
			while (isAlive()) {
				interrupt();
				MiscUtils.relieveCPU();
			}
		}

		public void run() {
			while (true) {
				if (isInterrupted()) {
					break;
				}
				BufferedImage imageFromDevice = lastImageFromDevice;
				if (imageFromDevice == null) {
					waitForNewImageFromDevice();
					continue;
				}

				if (imageFromDevice == lastDisplayedImage) {
					waitForNewImageFromDevice();
					continue;
				}
				lastDisplayedImage = imageFromDevice;
				imageDisplayUtil.displayNewImageFromDevice(imageFromDevice);
			}
		}

		public void waitForNewImageFromDevice() {
			synchronized (newImageFromDeviceMonitor) {
				try {
					newImageFromDeviceMonitor.wait();
				} catch (InterruptedException ignore) {
					interrupt();
				}
			}
		}

		public void notifyNewImageFromDevice() {
			synchronized (newImageFromDeviceMonitor) {
				newImageFromDeviceMonitor.notify();
			}
		}

	}

}
