package com.otk.application.image.camera;

import java.awt.Image;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.JOptionPane;
import javax.swing.JScrollPane;

import com.otk.application.audio.SoundTrack;
import com.otk.application.error.UnexpectedError;
import com.otk.application.image.video.MJPEGVideoSequence;
import com.otk.application.util.Accessor;
import com.otk.application.util.CommandExecutor;
import com.otk.application.util.CommandLineInterface;
import com.otk.application.util.ImageUtils;
import com.otk.application.util.Listener;
import com.otk.application.util.MiscUtils;
import com.otk.application.window.component.ImagePanel;

/*
 * TYPICAL COMMANDS
 * ----------------
 * 
  * - List devices:
 * v4l2-ctl --list-devices
 * 
 * - List encoding/resolutions:
 * ffmpeg  -f video4linux2 -list_formats all -i /dev/video0
 * 
* - Streaming:
 * ffmpeg -f video4linux2 -s 320x240 -i /dev/video0  -qscale:v 2 -f mjpeg - | ffplay -
 * 
 */
public class FFMpegDriver extends AbstractCommandLineBasedCameraDriver {

	public static void main(String[] args) {
		Camera camera = new Camera() {
			@Override
			public List<AbstractCameraDriver> createDrivers() {
				return Collections.singletonList(new FFMpegDriver(this));
			}
		};
		camera.setFpsUpdateEnabled(true);
		camera.setFpsLabelEnabled(true);
		List<String> devices = camera.listAvailableDevices(null);
		camera.setDeviceFullName(devices.get(0));
		List<FrameFormat> formats = camera.getVideoFormats();
		System.out.println("formats: " + formats);
		camera.setVideoFormat(formats.get(0));
		camera.start();
		camera.getVideoComponent().setPreferredSize(MiscUtils.getWindowAverageBounds().getSize());
		JOptionPane.showMessageDialog(null, camera.getVideoComponent());
		JOptionPane.showMessageDialog(null, new JScrollPane(new ImagePanel(camera.shoot())) {
			private static final long serialVersionUID = 1L;

			{
				setPreferredSize(MiscUtils.getWindowAverageBounds().getSize());
			}
		});
		camera.stop();
		camera.setPhotoSpecificFormat(formats.get(formats.size() - 1));
		JOptionPane.showMessageDialog(null, new JScrollPane(new ImagePanel(camera.shoot())) {
			private static final long serialVersionUID = 1L;

			{
				setPreferredSize(MiscUtils.getWindowAverageBounds().getSize());
			}
		});
	}

	private BufferedImage lastImage;

	public FFMpegDriver(Camera camera) {
		super(camera);
	}

	@Override
	public String getName() {
		return "Video4Linux";
	}

	@Override
	protected String cleanLogs(String outputLogs) {
		return outputLogs;
	}

	@Override
	protected CameraControlInterface<Void> getCompatibilityCheckingCommandInterface() {
		return new CameraControlInterface<Void>(new File("ffmpeg")) {
			{
				setCommandOptions("-version");
			}

			@Override
			protected Void doCommunicate(Process process) {
				waitSafelyFor(process, getSafeReadingTimeoutMilliseconds());
				if (process.exitValue() != 0) {
					throw new UnexpectedError("Invalid status code");
				}
				return null;
			}
		};
	}

	@Override
	protected CameraControlInterface<List<String>> getDeviceListingCommandInterface() {
		return new CameraControlInterface<List<String>>(new File("v4l2-ctl")) {
			{
				setCommandOptions("--list-devices");
			}

			@Override
			protected List<String> doCommunicate(Process process) {
				acquireSafelyAvailableOutputData(process);
				waitSafelyFor(process, getSafeReadingTimeoutMilliseconds());
				List<String> result = new ArrayList<String>();
				String output = new String(getOutputBuffer().toByteArray());
				output = output.trim();
				if (output.length() > 0) {
					Pattern pattern = Pattern.compile("(.*):\\s+(.*)");
					Matcher macther = pattern.matcher(output);
					while (macther.find()) {
						if (macther.groupCount() != 2) {
							throw new UnexpectedError("Invalid output");
						}
						String deviceId = macther.group(2);
						result.add(deviceId);
					}
				}
				return result;
			}
		};
	}

	@Override
	protected CommandLineInterface<List<FrameFormat>> getPhotoFormatListingCommandInterface(String localDeviceName) {
		return getVideoFormatListingCommandInterface(localDeviceName);
	}

	@Override
	protected CommandLineInterface<List<FrameFormat>> getVideoFormatListingCommandInterface(String localDeviceName) {
		return new CameraControlInterface<List<FrameFormat>>(new File("ffmpeg")) {
			{
				setCommandOptions(
						"-f video4linux2 -list_formats all -i " + CommandExecutor.quoteArgument(localDeviceName));
			}

			@Override
			protected List<FrameFormat> doCommunicate(Process process) {

				acquireSafelyAvailableErrorData(process);
				waitSafelyFor(process, getSafeReadingTimeoutMilliseconds());

				String output = new String(getErrorBuffer().toByteArray());
				if (!output.contains("Immediate exit requested")) {
					throw new UnexpectedError("Invalid output");
				}

				SortedSet<FrameFormat> result = new TreeSet<FrameFormat>();
				output = output.trim();
				if (output.length() > 0) {
					Pattern pattern = Pattern.compile("(?<=\\s)([0-9]+)x([0-9]+\\s)");
					Matcher macther = pattern.matcher(output);
					while (macther.find()) {
						if (macther.groupCount() != 2) {
							throw new UnexpectedError("Invalid output");
						}
						int width = Integer.valueOf(macther.group(1).trim());
						int height = Integer.valueOf(macther.group(2).trim());
						result.add(new FrameFormat(width, height));
					}
				}
				return new ArrayList<FrameFormat>(result);
			}
		};
	}

	@Override
	protected CameraControlInterface<Void> getLiveViewCommandInterface(String deviceLocalName, FrameFormat format,
			final Listener<BufferedImage> deviceImageListener, final Accessor<Boolean> interruptionRequested) {
		return new CameraControlInterface<Void>(new File("ffmpeg")) {
			{
				setCommandOptions("-f video4linux2 -s " + format.width + "x" + format.height + " -i "
						+ CommandExecutor.quoteArgument(deviceLocalName) + " -qscale:v 2 -f mjpeg -");
			}

			@Override
			protected Void doCommunicate(Process process) {
				if (-1 == acquireSafelyErrorDataUntil(process, "Press [q] to stop, [?] for help\n".getBytes())) {
					camera.handleRunErrorAndStop(new UnexpectedError("Invalid output"));
					return null;
				}
				try {
					MJPEGVideoSequence videoSequence = new MJPEGVideoSequence(
							new MJPEGVideoSequence.IMJPEGStreamProvider() {
								@Override
								public InputStream acquireInputStream() {
									return process.getInputStream();
								}

								@Override
								public void releaseInputStream(InputStream inputStream) {
								}

							}, 0f);
					Iterator<Image> frameIterator = videoSequence.createFrameIterator();
					while (true) {
						if (interruptionRequested.get()) {
							process.getOutputStream().write("q".getBytes());
							process.getOutputStream().flush();
							acquireSafelyAvailableOutputData(process);
							waitSafelyFor(process, getSafeReadingTimeoutMilliseconds());
							break;
						}
						if (!MiscUtils.runWithTimeout(new Runnable() {
							@Override
							public void run() {
								if (!frameIterator.hasNext()) {
									throw new UnexpectedError("Next frame unavailable");
								}
								lastImage = ImageUtils.adapt(frameIterator.next());
							}
						}, getSafeReadingTimeoutMilliseconds(), getSafeReadingExecutor())) {
							throw new UnexpectedError("Failed to get the next frame");
						}
						deviceImageListener.handle(lastImage);
						if (process.getErrorStream().available() > 0) {
							process.getErrorStream().read(new byte[process.getErrorStream().available()]);
						}
					}

				} catch (Throwable t) {
					camera.handleRunErrorAndStop(t);
				} finally {
					lastImage = null;
				}
				return null;
			}
		};
	}

	@Override
	protected CameraControlInterface<BufferedImage> getShootingCommandInterface(String deviceLocalName,
			FrameFormat format) {
		return new CameraControlInterface<BufferedImage>(new File("echo")) {

			{
				setCommandOptions("Getting first frames...");
			}

			@Override
			protected boolean isProcessSynchronized() {
				return false;
			}

			@Override
			protected BufferedImage doCommunicate(Process process) {
				waitSafelyFor(process, getSafeReadingTimeoutMilliseconds());
				final BufferedImage[] result = new BufferedImage[1];
				Listener<BufferedImage> deviceImageListener = new Listener<BufferedImage>() {
					long startTime = 0;

					@Override
					public void handle(BufferedImage newImage) {
						if (startTime == 0) {
							startTime = System.currentTimeMillis();
							return;
						}
						long sinceStart = System.currentTimeMillis() - startTime;
						if (sinceStart < camera.getWebcamAutoWhiteBalanceStabilizationDurationMilliseconds()) {
							return;
						}
						camera.getAudioEngine().play(SoundTrack.DEFAULT, Camera.CAMERA_CLICK_SOUND, false);
						result[0] = newImage;
					}
				};
				Accessor<Boolean> interruptionRequested = new Accessor<Boolean>() {

					@Override
					public Boolean get() {
						return result[0] != null;
					}
				};
				CameraControlInterface<Void> liveView = getLiveViewCommandInterface(deviceLocalName, format,
						deviceImageListener, interruptionRequested);
				liveView.execute();
				return result[0];
			}
		};
	}

	@Override
	public FrameFormat getDefaultPhotoFormat() {
		return camera.getVideoFormat();
	}

	@Override
	public boolean canShootWhileCameraActive() {
		return camera.getPhotoFormat().equals(camera.getVideoFormat());
	}

	@Override
	public BufferedImage doShootWhileCameraActive() {
		camera.getAudioEngine().play(SoundTrack.DEFAULT, Camera.CAMERA_CLICK_SOUND, false);
		BufferedImage result;
		while ((result = lastImage) == null) {
			if (!camera.isActive()) {
				throw new UnexpectedError("Failed to get the next frame");
			}
			MiscUtils.relieveCPU();
		}
		return correctDeviceImage(result, camera.getPhotoFormat());
	}

	@Override
	protected boolean isErrorDiagnosticEnabled() {
		return true;
	}

	@Override
	public boolean isActive() {
		return true;
	}

	@Override
	public void setActive(boolean b) {
		throw new UnsupportedOperationException();
	}

}
