package com.otk.application.image.camera;

import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.util.Iterator;
import java.util.List;

import com.otk.application.audio.SoundTrack;
import com.otk.application.error.AbstractApplicationError;
import com.otk.application.error.UnexpectedError;
import com.otk.application.util.ImageUtils;
import com.otk.application.util.MathUtils;
import com.otk.application.util.MiscUtils;

public abstract class WebcamDriver extends AbstractCameraDriver {

	private long shootingSoundStartTime;

	protected abstract void startCapture(FrameFormat format);

	protected WebcamDriver(Camera camera) {
		super(camera);
	}

	@Override
	public String getName() {
		return "Webcam";
	}

	@Override
	public FrameFormat getDefaultPhotoFormat() {
		return camera.getVideoFormat();
	}

	@Override
	public boolean canShootWhileCameraActive() {
		return camera.getPhotoFormat().equals(camera.getVideoFormat());
	}

	@Override
	public BufferedImage doShootWhileCameraActive() {
		if (camera.isVerbose()) {
			camera.logInfo("Getting " + WebcamDriver.this + " single image in active mode");
		}
		startShootingSound();
		final BufferedImage result = getNextImage();
		waitForShootingSoundEnd();
		MiscUtils.sleepSafely(camera.getWebcamAfterShootingPauseDurationMilliseconds());
		return result;
	}

	@Override
	public BufferedImage doShootWhileCameraInitializedAndNotActive() {
		if (camera.isVerbose()) {
			camera.logInfo("Getting " + WebcamDriver.this + " single image in inactive mode");
		}
		startCapture(camera.getPhotoFormat());
		MiscUtils.sleepSafely(camera.getWebcamAutoWhiteBalanceStabilizationDurationMilliseconds());
		startShootingSound();
		BufferedImage result = getNextImage();
		stopCapture();
		waitForShootingSoundEnd();
		MiscUtils.sleepSafely(camera.getWebcamAfterShootingPauseDurationMilliseconds());
		return result;
	}

	@Override
	public void startCapture() {
		startCapture(camera.getVideoFormat());
	}

	private BufferedImage getNextImage() {
		if (camera.isVerbose()) {
			camera.logInfo("Waiting for " + WebcamDriver.this + " next image");
		}
		camera.handleNewImageFromDevice(null);
		final BufferedImage[] result = new BufferedImage[1];
		if (!MiscUtils.runWithTimeout(new Runnable() {
			@Override
			public void run() {
				while ((result[0] = camera.getCurrentRawImage()) == null) {
					MiscUtils.relieveCPU();
					if (Thread.currentThread().isInterrupted()) {
						result[0] = null;
					}
				}
			}
		}, 5000, withoutFreezingExecutor)) {
			throw new UnexpectedError("Time out error. The software execution may become unstable");
		}
		return result[0];
	}

	protected void preventSuccessiveActivationCrashWhile(Runnable runnable) {
		try {
			doCleanUp();
			doSetUp();
		} catch (Exception e) {
			AbstractApplicationError.rethrow(e);
		}
		try {
			runnable.run();
		} finally {
			try {
				doCleanUp();
				doSetUp();
			} catch (Exception e) {
				AbstractApplicationError.rethrow(e);
			}
		}
	}

	private void startShootingSound() {
		shootingSoundStartTime = System.currentTimeMillis();
		camera.getAudioEngine().play(SoundTrack.DEFAULT, Camera.CAMERA_CLICK_SOUND, false);
	}

	private void waitForShootingSoundEnd() {
		long soundDurationMillis = MathUtils
				.round(1000 * camera.getAudioEngine().getDurationSeconds(Camera.CAMERA_CLICK_SOUND));
		long remainingSoundTime = shootingSoundStartTime + soundDurationMillis - System.currentTimeMillis();
		if (remainingSoundTime > 0) {
			MiscUtils.sleepSafely(remainingSoundTime);
		}
		while ((camera.getAudioEngine().getLastSoundData(SoundTrack.DEFAULT) == Camera.CAMERA_CLICK_SOUND)
				&& camera.getAudioEngine().isPlaying(SoundTrack.DEFAULT)) {
			MiscUtils.relieveCPU();
		}
	}

	protected static void arrangeFormatList(List<FrameFormat> list) {
		for (Iterator<FrameFormat> it = list.iterator(); it.hasNext();) {
			FrameFormat format = it.next();
			if (format.width == 0) {
				it.remove();
			}
			if (format.height == 0) {
				it.remove();
			}
		}
	}

	public BufferedImage correctDeviceImage(BufferedImage inputImage) {
		int width = inputImage.getWidth(null);
		int height = inputImage.getHeight(null);
		BufferedImage result = new BufferedImage(width, height, ImageUtils.getAdaptedBufferedImageType());
		Graphics2D g2d = result.createGraphics();
		g2d.translate(width, 0);
		g2d.scale(-1, 1);
		g2d.drawImage(inputImage, getDeviceImageCorrectionOp(inputImage), 0, 0);
		g2d.dispose();
		return result;
	}

	@Override
	public List<FrameFormat> getPhotoFormatsWhileCameraInitializedAndNotActive() {
		return getVideoFormatsWhileCameraInitializedAndNotActive();
	}

}
