package com.otk.application.image.filter;

import java.awt.Color;
import java.awt.Point;
import java.awt.image.BufferedImage;

import com.otk.application.error.StandardError;
import com.otk.application.image.FastestPixels;
import com.otk.application.util.ImageUtils;

public abstract class AbstractCompositionFunction {

	private OptimizationCache optimizationCache = new OptimizationCache();

	protected abstract int getNewPixelColorComponentValue(BufferedImage oldImage, BufferedImage newImage,
			char componentName, Point pixelPosition, Color oldPixelColor, Color newPixelColor);

	public BufferedImage execute(BufferedImage oldImage, BufferedImage newImage) {
		if (!ImageUtils.getSize(oldImage).equals(ImageUtils.getSize(newImage))) {
			throw new StandardError(
					"Image composition error: Cannot apply an image composition with a resizing filter. Use a '"
							+ ResetComposition.class.getSimpleName() + "' filter before the resizing filter. ");
		}
		int w = newImage.getWidth();
		int h = newImage.getHeight();
		BufferedImage dstImage = new BufferedImage(w, h, ImageUtils.getAdaptedBufferedImageType());
		FastestPixels.compose(oldImage, newImage, dstImage, new FastestPixels.ComposeFilter() {
			@Override
			public int get(int x, int y, int rgb1, int rgb2, int[] srcPixels1, int[] srcPixels2) {
				Point pixelPosition = new Point(x, y);
				Color oldPixelColor = new Color(rgb1, true);
				Color newPixelColor = new Color(rgb2, true);
				Object optimizationKey = getOptimizationCacheKey(oldImage, newImage, pixelPosition, oldPixelColor,
						newPixelColor);
				Color destColor;
				if (optimizationKey != null) {
					destColor = optimizationCache.get(optimizationKey);
					if (destColor == null) {
						destColor = getNewPixelColor(oldImage, newImage, pixelPosition, oldPixelColor, newPixelColor);
						optimizationCache.put(optimizationKey, destColor);
					}
				} else {
					destColor = getNewPixelColor(oldImage, newImage, pixelPosition, oldPixelColor, newPixelColor);
				}
				return destColor.getRGB();
			}
		});
		return dstImage;
	}

	protected Color getNewPixelColor(BufferedImage oldImage, BufferedImage newImage, Point pixelPosition,
			Color oldPixelColor, Color newPixelColor) {
		int destR = getNewPixelColorComponentValue(oldImage, newImage, 'r', pixelPosition, oldPixelColor,
				newPixelColor);
		int destG = getNewPixelColorComponentValue(oldImage, newImage, 'g', pixelPosition, oldPixelColor,
				newPixelColor);
		int destB = getNewPixelColorComponentValue(oldImage, newImage, 'b', pixelPosition, oldPixelColor,
				newPixelColor);
		int destA = getNewPixelColorComponentValue(oldImage, newImage, 'a', pixelPosition, oldPixelColor,
				newPixelColor);
		destR = Math.min(255, Math.max(0, destR));
		destG = Math.min(255, Math.max(0, destG));
		destB = Math.min(255, Math.max(0, destB));
		destA = Math.min(255, Math.max(0, destA));
		return new Color(destR, destG, destB, destA);
	}

	protected Object getOptimizationCacheKey(BufferedImage oldImage, BufferedImage newImage, Point pixelPosition,
			Color oldPixelColor, Color newPixelColor) {
		return null;
	}

}