package com.otk.application.image.filter;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.Stroke;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.image.BufferedImage;
import java.util.Arrays;
import java.util.List;

import javax.swing.JCheckBox;
import javax.swing.JPanel;

import com.otk.application.error.UnexpectedError;
import com.otk.application.translation.T;
import com.otk.application.util.ImageUtils;
import com.otk.application.util.MiscUtils;
import com.otk.application.window.component.crud.AbstractCrudItem;
import com.otk.application.window.component.crud.ICrudItemControlListener;

public abstract class AbstractFilter extends AbstractCrudItem {

	public static final String DISABLED_PREFIX = "DISABLED: ";

	public static final AbstractFilter NULL_FILTER = new AbstractFilter() {

		@Override
		protected void createFieldControls(JPanel panel, ICrudItemControlListener listener) {
		}

		@Override
		public String getCaption() {
			return "";
		}

		@Override
		protected FilteringContext doExecute(FilteringContext context) {
			return context;
		}
	};

	protected IFilterEditor editor;
	private boolean enabled = true;
	private JCheckBox enabledStatusField;
	private Object executionMutex = createExecutionMutex();

	protected abstract FilteringContext doExecute(FilteringContext context);

	public abstract String getCaption();

	protected Object createExecutionMutex() {
		return new Object();
	}

	public BufferedImage getPreview(FilteringContext context) {
		if (getCategory() == Category.IMAGE_MODIFIER) {
			BufferedImage image = context.getImage();
			image = optimizeForPreview(image);
			try {
				return doExecute(context.withImage(image)).getImage();
			} catch (Exception e) {
				context.logError(MiscUtils.getPrintedStackTrace(e));
				return ImageUtils.getBufferedImage(ImageUtils.NOT_FOUND_IMAGE);
			}
		} else {
			return context.getImage();
		}
	}

	protected BufferedImage optimizeForPreview(BufferedImage image) {
		int maxPreviewSize = getMaximumPreviewSize();
		if ((image.getWidth() > maxPreviewSize) || (image.getHeight() > maxPreviewSize)) {
			image = ImageUtils.getBufferedImage(
					ImageUtils.scalePreservingRatio(image, maxPreviewSize, maxPreviewSize, true, true));
		}
		return image;
	}

	protected int getMaximumPreviewSize() {
		return 100;
	}

	public Category getCategory() {
		return Category.IMAGE_MODIFIER;
	}

	public boolean isEnabled() {
		return enabled;
	}

	@Override
	protected void onFieldControlStateChange(String fieldName, Object value) {
		super.onFieldControlStateChange(fieldName, value);
		if (fieldListenerEnabled) {
			editor.onFieldControlStateChange();
		}
	}

	public final FilteringContext execute(final FilteringContext context) {
		if (!isEnabled()) {
			return context;
		}
		if (executionMutex == null) {
			return doExecute(context);
		} else {
			synchronized (executionMutex) {
				return doExecute(context);
			}
		}
	}

	@Override
	protected void onApplyRequested(ICrudItemControlListener listener) {
		setEnabled(enabledStatusField.isSelected());
		super.onApplyRequested(editor);
	}

	@Override
	public Component createComponent(ICrudItemControlListener listener) {
		this.editor = (IFilterEditor) listener;
		Component component = super.createComponent(editor);
		enabledStatusField = new JCheckBox(T.t("Enable"));
		enabledStatusField.setSelected(enabled);
		enabledStatusField.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				onFieldControlStateChange("enabled", enabledStatusField.isSelected());
			}
		});
		int applyButtonPosition = Arrays.asList(fieldsPanel.getComponents()).indexOf(applyButton);
		fieldsPanel.add(enabledStatusField, applyButtonPosition + 1);
		return component;
	}

	public void setEnabled(boolean enabled) {
		this.enabled = enabled;
	}

	public static AbstractFilter getMultiHeadlessFilter(final List<AbstractFilter> filters) {
		return new MultiFilter(filters);
	}

	public boolean isEnabledStatusFieldSelected() {
		return enabledStatusField.isSelected();
	}

	protected Stroke getPreviewStroke(BufferedImage image) {
		return new BasicStroke(Math.max(1f, (image.getWidth() + image.getHeight()) / 200f));
	}

	protected static int getColorComponentValue(Color color, char component) {
		if (component == 'r') {
			return color.getRed();
		} else if (component == 'g') {
			return color.getGreen();
		} else if (component == 'b') {
			return color.getBlue();
		} else if (component == 'a') {
			return color.getAlpha();
		} else {
			throw new UnexpectedError();
		}
	}

	public enum Category {
		IMAGE_MODIFIER, FILTER_COMPOSER, DECORATION
	}

}
