package com.otk.application.image.filter;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.image.BufferedImage;
import java.util.Collections;
import java.util.List;

import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.SpinnerNumberModel;

import com.otk.application.error.UnexpectedError;
import com.otk.application.image.camera.Camera;
import com.otk.application.translation.T;
import com.otk.application.util.MathUtils;
import com.otk.application.window.component.crud.ICrudItemControlListener;

public class Lighting extends AbstractPixelFilter {

	public static void main(String[] args) {
		Camera camera = new Camera();
		camera.setFpsUpdateEnabled(true);
		Lighting filter = new Lighting();
		filter.radius = 1.0;
		filter.saturation = 255;
		camera.setVideoInputCorrectionFilters(Collections.singletonList(filter));
		List<String> deviceNames = camera.listAvailableDevices(null);
		camera.setDeviceFullName(deviceNames.get(deviceNames.size() - 1));
		camera.start();
		camera.getVideoComponent().setPreferredSize(new Dimension(400, 300));
		JOptionPane.showMessageDialog(null, camera.getVideoComponent());
		camera.stop();
	}

	public double x = 0.5;
	public double y = 0.5;
	public double radius = 0.25;
	public int amount = 50;
	public int hue = 0;
	public int saturation = 0;
	public int concentration = 50;

	private int offsetRed;
	private int offsetGreen;
	private int offsetBlue;

	@Override
	public String getCaption() {
		return T.t("Lighting");
	}

	@Override
	public void createFieldControls(JPanel panel, final ICrudItemControlListener listener) {
		createSlinnerFieldControl(panel, "x", new SpinnerNumberModel(-1.0, -1.0, 2.0, 0.01),
				T.t("horizontal position"));
		createSlinnerFieldControl(panel, "y", new SpinnerNumberModel(-1.0, -1.0, 2.0, 0.01), T.t("vertical position"));
		createSlinnerFieldControl(panel, "radius", new SpinnerNumberModel(0.0, 0.0, 2.0, 0.01), T.t("radius"));
		createSlinnerFieldControl(panel, "amount", -100, 100, T.t("amount"));
		createSlinnerFieldControl(panel, "hue", 0, 255, T.t("hue"));
		createSlinnerFieldControl(panel, "saturation", 0, 255, T.t("saturation"));
		createSlinnerFieldControl(panel, "concentration", 0, 100, T.t("concentration"));
	}

	@Override
	public BufferedImage getPreview(FilteringContext context) {
		BufferedImage image = super.getPreview(context);
		Graphics2D g2d = image.createGraphics();
		int w = image.getWidth();
		int h = image.getHeight();

		int lightCenterX = MathUtils.round((x * w));
		int lightCenterY = MathUtils.round((y * h));
		g2d.setStroke(getPreviewStroke(image));
		g2d.setColor(Color.RED);
		int LIGHT_SOURCE_DISC_RADIUS = 4;
		g2d.fillOval(lightCenterX - LIGHT_SOURCE_DISC_RADIUS, lightCenterY - LIGHT_SOURCE_DISC_RADIUS,
				LIGHT_SOURCE_DISC_RADIUS * 2, LIGHT_SOURCE_DISC_RADIUS * 2);

		int lightW = MathUtils.round(radius * w * 2);
		int lightH = MathUtils.round(radius * h * 2);
		int lightX = MathUtils.round(lightCenterX - (lightW / 2.0));
		int lightY = MathUtils.round(lightCenterY - (lightH / 2.0));
		g2d.drawOval(lightX, lightY, lightW, lightH);
		g2d.dispose();
		return image;
	}

	@Override
	protected void onImageProcessingStart(FilteringContext inputContext) {
		int offsetRGB = Color.HSBtoRGB(hue / 255.0f, saturation / 255.0f, Math.abs(amount / 255.0f));
		Color offsetColor = new Color(offsetRGB);
		int offsetSign = Math.round(Math.signum(amount));
		offsetRed = offsetColor.getRed() * offsetSign;
		offsetGreen = offsetColor.getGreen() * offsetSign;
		offsetBlue = offsetColor.getBlue() * offsetSign;
	}

	@Override
	protected int getNewPixelColorComponentValue(char c, BufferedImage im, Point p, Color pixelColor) {
		double lightSourceDistance = Math.sqrt(
				Math.pow(((double) p.x) / im.getWidth() - x, 2) + Math.pow(((double) p.y) / im.getHeight() - y, 2));
		double lightSourceProximity = Math.max(0.0, (radius - lightSourceDistance)) / radius;
		double lightSourceFactor;
		if (lightSourceProximity == 0.0) {
			lightSourceFactor = 0.0;
		} else {
			lightSourceFactor = MathUtils.convexRatio(lightSourceProximity, concentration);
		}
		int componentValue;
		int componentValueOffset;
		if (c == 'r') {
			componentValue = pixelColor.getRed();
			componentValueOffset = offsetRed;
		} else if (c == 'g') {
			componentValue = pixelColor.getGreen();
			componentValueOffset = offsetGreen;
		} else if (c == 'b') {
			componentValue = pixelColor.getBlue();
			componentValueOffset = offsetBlue;
		} else if (c == 'a') {
			return pixelColor.getAlpha();
		} else {
			throw new UnexpectedError();
		}
		return MathUtils.round(componentValue + (componentValueOffset * lightSourceFactor));
	}

}
