package com.otk.application.image.printer;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.geom.Rectangle2D;
import java.awt.geom.Rectangle2D.Double;
import java.awt.image.BufferedImage;
import java.awt.print.Book;
import java.awt.print.PageFormat;
import java.awt.print.Paper;
import java.awt.print.Printable;
import java.awt.print.PrinterException;
import java.io.File;
import java.io.IOException;
import java.text.MessageFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.concurrent.ExecutorService;
import java.util.stream.Collectors;

import javax.print.Doc;
import javax.print.DocFlavor;
import javax.print.DocPrintJob;
import javax.print.PrintService;
import javax.print.PrintServiceLookup;
import javax.print.SimpleDoc;
import javax.print.attribute.HashPrintRequestAttributeSet;
import javax.print.attribute.PrintRequestAttributeSet;
import javax.print.attribute.standard.JobName;
import javax.print.attribute.standard.Media;
import javax.print.attribute.standard.MediaPrintableArea;
import javax.print.attribute.standard.MediaSize;
import javax.print.attribute.standard.MediaSizeName;
import javax.print.attribute.standard.PrintQuality;
import javax.print.event.PrintJobEvent;
import javax.print.event.PrintJobListener;

import org.apache.pdfbox.pdmodel.PDDocument;
import org.apache.pdfbox.pdmodel.PDPage;
import org.apache.pdfbox.pdmodel.PDPageContentStream;
import org.apache.pdfbox.pdmodel.common.PDRectangle;
import org.apache.pdfbox.pdmodel.graphics.image.LosslessFactory;
import org.apache.pdfbox.pdmodel.graphics.image.PDImageXObject;

import com.otk.application.error.ErrorManager;
import com.otk.application.error.StandardError;
import com.otk.application.error.UnexpectedError;
import com.otk.application.translation.T;
import com.otk.application.util.ImageUtils;
import com.otk.application.util.MathUtils;
import com.otk.application.util.MiscUtils;

public class PrintingManager {

	public static final int BAD_QUALITY_MEMORY_PRINT_DPI = 72;
	public static final int MEDIUM_QUALITY_MEMORY_PRINT_DPI = 150;
	public static final int GOOD_QUALITY_MEMORY_PRINT_DPI = 300;
	public static final int PRINT_DPI = 72;

	private String printerName = "";
	private int leftMarginMillimiters = 2;
	private int rightMarginMillimiters = 2;
	private int topMarginMillimiters = 2;
	private int bottomMarginMillimiters = 2;
	private double paperWidthMillimeters = 105;
	private double paperHeightMillimeters = 148;
	private int copyCount = 1;
	private int freezeTimeoutMilliSeconds = 300000;
	private int deviceDpi = 300;
	private boolean antialiasing = true;
	private boolean verbose = false;
	private String jobName = new SimpleDateFormat("yyyy-MM-dd_HH-mm-ss").format(new Date());

	private boolean retrying;
	private boolean printMethodExecuting = false;
	private boolean printingTaskExecuting = false;
	private PrintingStrategy printingStrategy = new DefaultPrintingStrategy(this);
	private ExecutorService asynchronousPrintingExecutor = MiscUtils
			.newExecutor(PrintingManager.class.getName() + ".executor", 0, true);

	public String getJobName() {
		return jobName;
	}

	public void setJobName(String jobName) {
		this.jobName = jobName;
	}

	public PrintingStrategy getPrintingStrategy() {
		return printingStrategy;
	}

	public void setPrintingStrategy(PrintingStrategy printingStrategy) {
		this.printingStrategy = printingStrategy;
	}

	public String getPrinterName() {
		return printerName;
	}

	public void setPrinterName(String printerName) {
		this.printerName = printerName;
	}

	public int getLeftMarginMillimiters() {
		return leftMarginMillimiters;
	}

	public void setLeftMarginMillimiters(int leftMarginMillimiters) {
		this.leftMarginMillimiters = leftMarginMillimiters;
	}

	public int getRightMarginMillimiters() {
		return rightMarginMillimiters;
	}

	public void setRightMarginMillimiters(int rightMarginMillimiters) {
		this.rightMarginMillimiters = rightMarginMillimiters;
	}

	public int getTopMarginMillimiters() {
		return topMarginMillimiters;
	}

	public void setTopMarginMillimiters(int topMarginMillimiters) {
		this.topMarginMillimiters = topMarginMillimiters;
	}

	public int getBottomMarginMillimiters() {
		return bottomMarginMillimiters;
	}

	public void setBottomMarginMillimiters(int bottomMarginMillimiters) {
		this.bottomMarginMillimiters = bottomMarginMillimiters;
	}

	public double getPaperWidthMillimeters() {
		return paperWidthMillimeters;
	}

	public void setPaperWidthMillimeters(double paperWidthMillimeters) {
		this.paperWidthMillimeters = paperWidthMillimeters;
	}

	public double getPaperHeightMillimeters() {
		return paperHeightMillimeters;
	}

	public void setPaperHeightMillimeters(double paperHeightMillimeters) {
		this.paperHeightMillimeters = paperHeightMillimeters;
	}

	public PageOrientation getPageOrientation() {
		return PageOrientation.get(getPrintableAreaDots(GOOD_QUALITY_MEMORY_PRINT_DPI).getSize());
	}

	public int getCopyCount() {
		return copyCount;
	}

	public void setCopyCount(int copyCount) {
		this.copyCount = copyCount;
	}

	public int getFreezeTimeoutMilliSeconds() {
		return freezeTimeoutMilliSeconds;
	}

	public void setFreezeTimeoutMilliSeconds(int freezeTimeoutMilliSeconds) {
		this.freezeTimeoutMilliSeconds = freezeTimeoutMilliSeconds;
	}

	public int getDeviceDpi() {
		return deviceDpi;
	}

	public void setDeviceDpi(int deviceDpi) {
		this.deviceDpi = deviceDpi;
	}

	public boolean isAntialiasing() {
		return antialiasing;
	}

	public void setAntialiasing(boolean antialiasing) {
		this.antialiasing = antialiasing;
	}

	public boolean isVerbose() {
		return verbose;
	}

	public void setVerbose(boolean verbose) {
		this.verbose = verbose;
	}

	protected void logInfo(String s) {
		System.out.println(s);
	}

	protected void logError(String s) {
		System.err.println(s);
	}

	public boolean isPrinting() {
		return printMethodExecuting || printingTaskExecuting;
	}

	protected boolean isRetrying() {
		return retrying;
	}

	protected void setRetrying(boolean retrying) {
		this.retrying = retrying;
	}

	public List<PaperSpecification> getSupportedPaperSpecifications() {
		List<PaperSpecification> result = new ArrayList<PaperSpecification>();
		PrintService printService = getPrintService();
		Media[] medias = (Media[]) printService.getSupportedAttributeValues(Media.class, null, null);
		for (Media media : medias) {
			if (media instanceof MediaSizeName) {
				MediaSizeName msn = (MediaSizeName) media;
				MediaSize ms = MediaSize.getMediaSizeForName(msn);
				if (ms == null) {
					continue;
				}
				PrintRequestAttributeSet aset = new HashPrintRequestAttributeSet();
				aset.add(msn);
				MediaPrintableArea[] pas = (MediaPrintableArea[]) printService
						.getSupportedAttributeValues(MediaPrintableArea.class, null, aset);
				for (MediaPrintableArea pa : pas) {
					float paperWidth = ms.getX(MediaSize.MM);
					float paperHeight = ms.getY(MediaSize.MM);
					float printableAreaX = pa.getX(MediaSize.MM);
					float printableAreaY = pa.getY(MediaSize.MM);
					float printableAreaWidth = pa.getWidth(MediaSize.MM);
					float printableAreaHeight = pa.getHeight(MediaSize.MM);
					PaperSpecification ps = new PaperSpecification();
					ps.setWidthMillimeters((int) Math.round(paperWidth));
					ps.setHeightMillimeters((int) Math.round(paperHeight));
					ps.setLeftMarginMillimiters(((int) Math.round(printableAreaX)));
					ps.setTopMarginMillimiters(((int) Math.round(printableAreaY)));
					ps.setRightMarginMillimiters(((int) Math.round(paperWidth - printableAreaX - printableAreaWidth)));
					ps.setBottomMarginMillimiters(
							((int) Math.round(paperHeight - printableAreaY - printableAreaHeight)));
					if (!ps.isValid()) {
						continue;
					}
					result.add(ps);
				}
			}
		}
		result = result.stream().distinct().sorted().collect(Collectors.toList());
		return result;
	}

	public BufferedImage printToMemory(Image image, int dpi) {
		PageFormat pageFormat = createPageFormat(dpi);
		Paper paper = pageFormat.getPaper();
		BufferedImage result = new BufferedImage(MathUtils.round(paper.getWidth()), MathUtils.round(paper.getHeight()),
				ImageUtils.getAdaptedBufferedImageType());
		Graphics2D g2d = result.createGraphics();
		g2d.setColor(Color.WHITE);
		g2d.fillRect(0, 0, result.getWidth(), result.getHeight());
		try {
			createPrintable(image).print(g2d, pageFormat, 0);
		} catch (PrinterException e) {
			throw new UnexpectedError(e);
		}
		g2d.dispose();
		return result;
	}

	public void printToPDF(Image image, File file) throws IOException {
		PageFormat pageFormat = createPageFormat(PRINT_DPI);
		Paper paper = pageFormat.getPaper();
		PDDocument doc = new PDDocument();
		try {
			PDPage page = new PDPage(new PDRectangle((float) paper.getWidth(), (float) paper.getHeight()));
			doc.addPage(page);
			PDPageContentStream contents = new PDPageContentStream(doc, page);
			createPrintable(image).print(contents, pageFormat, doc);
			contents.close();
			doc.save(file);
		} finally {
			doc.close();
		}
	}

	public void printWithoutMargins(BufferedImage image) {
		int topMarginMillimitersToRestore = topMarginMillimiters;
		int bottomMarginMillimitersToRestore = bottomMarginMillimiters;
		int leftMarginMillimitersToRestore = leftMarginMillimiters;
		int rightMarginMillimitersToRestore = rightMarginMillimiters;
		topMarginMillimiters = 0;
		bottomMarginMillimiters = 0;
		leftMarginMillimiters = 0;
		rightMarginMillimiters = 0;
		try {
			print(image);
		} finally {
			topMarginMillimiters = topMarginMillimitersToRestore;
			bottomMarginMillimiters = bottomMarginMillimitersToRestore;
			leftMarginMillimiters = leftMarginMillimitersToRestore;
			rightMarginMillimiters = rightMarginMillimitersToRestore;
		}
	}

	public synchronized void print(final Image image) {
		print(image, true);
	}

	public synchronized void print(final Image image, final boolean waitForPrintingJobEnd) {
		printMethodExecuting = true;
		try {
			withoutFreezing(new Runnable() {

				@Override
				public void run() {
					getPrintingStrategy().handlePrintingInitializationEvent();
					logInfo("Printing...");
					final PageFormat pageFormat = createPageFormat(PRINT_DPI);
					final Book book = new Book();
					for (int i = 0; i < copyCount; i++) {
						book.append(createPrintable(fixPrintPoorQualityOnSomeDevices(image)), pageFormat);
					}
					getPrintingStrategy().resetReport();
					Runnable printingTask = new Runnable() {
						public void run() {
							if (printingTaskExecuting) {
								throw new UnexpectedError("Cannot execute multiple printing tasks concurrently!");
							}
							printingTaskExecuting = true;
							try {
								do {
									retrying = false;
									try {
										DocPrintJob printJob = getPrintService().createPrintJob();
										Doc doc = new SimpleDoc(book, DocFlavor.SERVICE_FORMATTED.PAGEABLE, null);
										ThePrintJobListener listener = createPrintJobListener();
										printJob.addPrintJobListener(listener);
										getPrintingStrategy().handlePrintjobStartEvent();
										PrintRequestAttributeSet printAttributes = new HashPrintRequestAttributeSet();
										printAttributes.add(PrintQuality.HIGH);
										printAttributes.add(new JobName(jobName, Locale.getDefault()));
										printJob.print(doc, printAttributes);
										while (!listener.isDead()) {
											MiscUtils.sleepSafely(1000);
										}
									} catch (Throwable t) {
										getPrintingStrategy().setLastError(t);
										getPrintingStrategy().handlePrintJobErrorEvent(t);
									} finally {
										getPrintingStrategy().handlePrintJobStopEvent();
									}
									if (retrying) {
										logInfo("Retrying to print");
									}
								} while (retrying);
								getPrintingStrategy().handlePrintingTerminationEvent();
							} finally {
								printingTaskExecuting = false;
							}
						}
					};
					if (waitForPrintingJobEnd) {
						printingTask.run();
						Throwable error = getPrintingStrategy().getLastError();
						if (error != null) {
							ErrorManager.rethrow(error);
						}
					} else {
						asynchronousPrintingExecutor.submit(printingTask);
					}
				}

				private Image fixPrintPoorQualityOnSomeDevices(Image image) {
					Double scaledBounds = getImageScaledBoundsInsidePrintableArea(image.getWidth(null),
							image.getHeight(null), deviceDpi);
					Dimension scaledSize = MathUtils.toDimension(MathUtils.getSize(scaledBounds));
					if (antialiasing) {
						return ImageUtils.scalePreservingRatio(image, scaledSize.width, scaledSize.height, true, true);
					} else {
						return image.getScaledInstance(scaledSize.width, scaledSize.height, Image.SCALE_SMOOTH);
					}
				}
			});
		} finally {
			printMethodExecuting = false;
		}
	}

	protected ThePrintable createPrintable(Image image) {
		return new ThePrintable(image);
	}

	protected void withoutFreezing(final Runnable runnable) {
		if (!MiscUtils.runWithTimeout(new Runnable() {
			@Override
			public void run() {
				runnable.run();
			}
		}, freezeTimeoutMilliSeconds, null)) {
			throw new UnexpectedError("Time out error. The software execution may become unstable");
		}
	}

	protected ThePrintJobListener createPrintJobListener() {
		return new ThePrintJobListener();
	}

	protected PrintService getPrintService() {
		logInfo("Initializing the printing service");
		if (printerName.length() == 0) {
			throw new StandardError(T.t("The printing device to use was not specified in the configuration"));
		}
		List<PrintService> availablePrinters = getAvailablePrintServices();
		PrintService printService = null;
		for (PrintService candidatePrinter : availablePrinters) {
			if (verbose) {
				logInfo("Found printer service: '" + candidatePrinter.getName() + "'");
			}
			if (printerName.equals(candidatePrinter.getName())) {
				printService = candidatePrinter;
			}
		}
		logInfo(availablePrinters.size() + " printer service(s) found");
		if (printService == null) {
			throw new StandardError("'" + printerName + "' printer service not found");
		}
		logInfo("Selected printer service: '" + printerName + "'");
		return printService;
	}

	public List<String> getAvailablePrinterNames() {
		List<String> result = new ArrayList<String>();
		for (PrintService serv : getAvailablePrintServices()) {
			result.add(serv.getName());
		}
		return result;
	}

	private List<PrintService> getAvailablePrintServices() {
		List<PrintService> result = new ArrayList<PrintService>();
		result.addAll(Arrays.asList(PrintServiceLookup.lookupPrintServices(DocFlavor.INPUT_STREAM.GIF, null)));
		PrintService printerStrategyTestService = getPrintingStrategy().getTestFakePrintService();
		if (printerStrategyTestService != null) {
			result.add(printerStrategyTestService);
		}
		return result;
	}

	protected PageFormat createPageFormat(int dpi) {
		PageFormat pageFormat = new PageFormat();
		Paper paper = pageFormat.getPaper();
		double widthDots = MathUtils.millimetersToDots(paperWidthMillimeters, dpi);
		double heightDots = MathUtils.millimetersToDots(paperHeightMillimeters, dpi);
		paper.setSize(widthDots, heightDots);
		double leftMarginDots = MathUtils.millimetersToDots(leftMarginMillimiters, dpi);
		double rightMarginDots = MathUtils.millimetersToDots(rightMarginMillimiters, dpi);
		double topMarginDots = MathUtils.millimetersToDots(topMarginMillimiters, dpi);
		double bottomMarginDots = MathUtils.millimetersToDots(bottomMarginMillimiters, dpi);
		double insideMarginsWidth = paper.getWidth() - (rightMarginDots + leftMarginDots);
		double insideMarginsHeight = paper.getHeight() - (bottomMarginDots + topMarginDots);
		paper.setImageableArea(leftMarginDots, topMarginDots, insideMarginsWidth, insideMarginsHeight);
		pageFormat.setPaper(paper);
		return pageFormat;
	}

	public Double getImageScaledBoundsInsidePrintableArea(int imageWidth, int imageHeight, int dpi) {
		return MathUtils.getImageScaledBoundsInsidePrintableArea(imageWidth, imageHeight, createPageFormat(dpi));
	}

	public double getPrintableAreaWidthOverHeightRatio() {
		PageFormat pageFormat = createPageFormat(MEDIUM_QUALITY_MEMORY_PRINT_DPI);
		return pageFormat.getImageableWidth() / pageFormat.getImageableHeight();
	}

	public Rectangle getPrintableAreaDots(int dpi) {
		PageFormat pageFormat = createPageFormat(dpi);
		int x = MathUtils.round(pageFormat.getImageableX());
		int y = MathUtils.round(pageFormat.getImageableY());
		int w = MathUtils.round(pageFormat.getImageableWidth());
		int h = MathUtils.round(pageFormat.getImageableHeight());
		return new Rectangle(x, y, w, h);
	}

	public Dimension getPaperSizeDots(int dpi) {
		Paper paper = createPageFormat(dpi).getPaper();
		return new Dimension(MathUtils.round(paper.getWidth()), MathUtils.round(paper.getHeight()));
	}

	protected class ThePrintJobListener implements PrintJobListener {

		private boolean dead = false;

		public ThePrintJobListener() {
			getPrintingStrategy().setLastError(null);
		}

		public boolean isDead() {
			return dead;
		}

		protected void handleError(String generalErrorMsg) {
			if (generalErrorMsg == null) {
				getPrintingStrategy().setLastError(null);
			} else {
				if (getPrintingStrategy().getLastError() != null) {
					getPrintingStrategy().setLastError(
							new StandardError(generalErrorMsg + ": " + getPrintingStrategy().getLastError().toString(),
									getPrintingStrategy().getLastError()));
				} else {
					getPrintingStrategy().setLastError(new StandardError(generalErrorMsg));
				}
			}
		}

		@Override
		public void printJobFailed(PrintJobEvent paramPrintJobEvent) {
			dead = true;
			handleError("Print job failed !");
		}

		@Override
		public void printJobCanceled(PrintJobEvent paramPrintJobEvent) {
			dead = true;
			handleError("Cancelled print job !");
		}

		@Override
		public void printJobNoMoreEvents(PrintJobEvent paramPrintJobEvent) {
			dead = true;
			getPrintingStrategy().handlePrintJobNoMoreEvents(this, paramPrintJobEvent);
		}

		@Override
		public void printJobRequiresAttention(PrintJobEvent paramPrintJobEvent) {
			handleError("Blocked print job !");
		}

		@Override
		public void printDataTransferCompleted(PrintJobEvent pje) {
		}

		@Override
		public void printJobCompleted(PrintJobEvent pje) {
			dead = true;
			handleError(null);
			logInfo("Print job completed");
		}
	}

	protected class ThePrintable implements Printable {
		private Image image;

		public ThePrintable(Image image) {
			this.image = image;
		}

		@Override
		public int print(Graphics graphics, PageFormat pageFormat, int pageIndex) throws PrinterException {
			Graphics2D graphics2d = (Graphics2D) graphics;
			final Rectangle2D.Double bounds = MathUtils.getImageScaledBoundsInsidePrintableArea(image.getWidth(null),
					image.getHeight(null), pageFormat);
			ImageUtils.setSmoothScalingRenderingHints(graphics2d);
			graphics2d.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);
			graphics2d.drawImage(image, MathUtils.round(bounds.x), MathUtils.round(bounds.y),
					MathUtils.round(bounds.width), MathUtils.round(bounds.height), null);
			return PAGE_EXISTS;
		}

		public void print(PDPageContentStream contents, PageFormat pageFormat, PDDocument doc) throws IOException {
			PDImageXObject pdImage = LosslessFactory.createFromImage(doc, ImageUtils.getBufferedImage(image));
			final Rectangle2D.Double drawBounds = MathUtils
					.getImageScaledBoundsInsidePrintableArea(image.getWidth(null), image.getHeight(null), pageFormat);
			drawBounds.x += pageFormat.getImageableX();
			drawBounds.y += pageFormat.getImageableY();
			contents.drawImage(pdImage, (float) drawBounds.x, (float) drawBounds.y, (float) drawBounds.width,
					(float) drawBounds.height);
		}
	}

	public static class PaperSpecification implements Comparable<PaperSpecification> {

		private int widthMillimeters;
		private int heightMillimeters;
		private int leftMarginMillimiters;
		private int rightMarginMillimiters;
		private int topMarginMillimiters;
		private int bottomMarginMillimiters;

		public int getWidthMillimeters() {
			return widthMillimeters;
		}

		public void setWidthMillimeters(int widthMillimeters) {
			this.widthMillimeters = widthMillimeters;
		}

		public int getHeightMillimeters() {
			return heightMillimeters;
		}

		public void setHeightMillimeters(int heightMillimeters) {
			this.heightMillimeters = heightMillimeters;
		}

		public int getLeftMarginMillimiters() {
			return leftMarginMillimiters;
		}

		public void setLeftMarginMillimiters(int leftMarginMillimiters) {
			this.leftMarginMillimiters = leftMarginMillimiters;
		}

		public int getRightMarginMillimiters() {
			return rightMarginMillimiters;
		}

		public void setRightMarginMillimiters(int rightMarginMillimiters) {
			this.rightMarginMillimiters = rightMarginMillimiters;
		}

		public int getTopMarginMillimiters() {
			return topMarginMillimiters;
		}

		public void setTopMarginMillimiters(int topMarginMillimiters) {
			this.topMarginMillimiters = topMarginMillimiters;
		}

		public int getBottomMarginMillimiters() {
			return bottomMarginMillimiters;
		}

		public void setBottomMarginMillimiters(int bottomMarginMillimiters) {
			this.bottomMarginMillimiters = bottomMarginMillimiters;
		}

		public boolean isValid() {
			if (leftMarginMillimiters < 0) {
				return false;
			}
			if (rightMarginMillimiters < 0) {
				return false;
			}
			if (topMarginMillimiters < 0) {
				return false;
			}
			if (bottomMarginMillimiters < 0) {
				return false;
			}
			if (leftMarginMillimiters >= (widthMillimeters - rightMarginMillimiters)) {
				return false;
			}
			if (topMarginMillimiters >= (heightMillimeters - bottomMarginMillimiters)) {
				return false;
			}
			return true;
		}

		@Override
		public int compareTo(PaperSpecification o) {
			int result = 0;
			result = new Integer(widthMillimeters).compareTo(new Integer(o.widthMillimeters));
			if (result != 0) {
				return result;
			}
			result = new Integer(heightMillimeters).compareTo(new Integer(o.heightMillimeters));
			if (result != 0) {
				return result;
			}
			return 0;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + bottomMarginMillimiters;
			result = prime * result + heightMillimeters;
			result = prime * result + leftMarginMillimiters;
			result = prime * result + rightMarginMillimiters;
			result = prime * result + topMarginMillimiters;
			result = prime * result + widthMillimeters;
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			PaperSpecification other = (PaperSpecification) obj;
			if (bottomMarginMillimiters != other.bottomMarginMillimiters)
				return false;
			if (heightMillimeters != other.heightMillimeters)
				return false;
			if (leftMarginMillimiters != other.leftMarginMillimiters)
				return false;
			if (rightMarginMillimiters != other.rightMarginMillimiters)
				return false;
			if (topMarginMillimiters != other.topMarginMillimiters)
				return false;
			if (widthMillimeters != other.widthMillimeters)
				return false;
			return true;
		}

		@Override
		public String toString() {
			return MessageFormat.format("(-{0}) {1} (-{2}) mm  x  (-{3}) {4} (-{5}) mm", leftMarginMillimiters,
					widthMillimeters, rightMarginMillimiters, topMarginMillimiters, heightMillimeters,
					bottomMarginMillimiters);
		}

	}

}
