package com.otk.application.image.video;

import java.awt.Image;
import java.awt.image.BufferedImage;
import java.io.BufferedInputStream;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.util.Iterator;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.TimeUnit;

import javax.imageio.ImageIO;
import javax.imageio.ImageReader;
import javax.imageio.stream.ImageInputStream;

import com.otk.application.error.UnexpectedError;
import com.otk.application.util.MiscUtils;

public class MJPEGVideoSequence implements IVideoSequence {

	private IMJPEGStreamProvider streamProvider;
	private float fps;

	public MJPEGVideoSequence(IMJPEGStreamProvider source, float fps) {
		this.streamProvider = source;
		this.fps = fps;
	}

	public float getFps() {
		return fps;
	}

	@Override
	public Iterator<Image> createFrameIterator() {
		return new MJPEGFrameIterator(this);
	}

	@Override
	public void disposeFrameIterator(Iterator<Image> frameIterator) {
		((MJPEGFrameIterator) frameIterator).dispose();
	}

	public static interface IMJPEGStreamProvider {

		InputStream acquireInputStream();

		void releaseInputStream(InputStream inputStream);

	}

	protected static class MJPEGFrameIterator implements Iterator<Image> {

		private static final int STREAM_READ_TIMEOUT = 5000;
		private static final byte[] JPEG_END = new byte[] { (byte) 0xFF, (byte) 0xD9 };
		private static final int MAX_IMAGE_SIZE = 10 * 1000 * 1000;

		protected MJPEGVideoSequence sequence;
		protected InputStream input;
		protected int skippedImageReaderErrors = 0;
		protected ExecutorService safeReadingExecutor = Executors.newSingleThreadExecutor(new ThreadFactory() {
			@Override
			public Thread newThread(Runnable r) {
				Thread result = new Thread(r, MiscUtils.formatThreadName(MJPEGVideoSequence.class, "StreamReader"));
				result.setDaemon(true);
				return result;
			}
		});;
		protected BufferedImage nextImage;
		protected BufferedInputStream bufferedStream;
		protected boolean firsFrameRead = false;

		public MJPEGFrameIterator(MJPEGVideoSequence sequence) {
			this.sequence = sequence;
			this.input = sequence.streamProvider.acquireInputStream();
			bufferedStream = new BufferedInputStream(input);
		}

		protected void prepareNext() {
			try {
				bufferedStream.mark(MAX_IMAGE_SIZE);
				if (bufferedStream.read() == -1) {
					nextImage = null;
					return;
				}
				bufferedStream.reset();
				ImageInputStream imageStream = ImageIO.createImageInputStream(bufferedStream);
				try {
					Iterator<ImageReader> readerIterator = ImageIO.getImageReaders(imageStream);
					if (!readerIterator.hasNext()) {
						throw new UnexpectedError("Invalid image format");
					}
					ImageReader reader = readerIterator.next();
					reader.setInput(imageStream);
					nextImage = reader.read(0);
					skippedImageReaderErrors = 0;
				} catch (Throwable t) {
					skippedImageReaderErrors++;
					if (skippedImageReaderErrors == 10) {
						throw new UnexpectedError("Persistent image reader error detected: " + t, t);
					}
					if (-1 == MiscUtils.readUntil(safeReadingExecutor, bufferedStream, new ByteArrayOutputStream(),
							STREAM_READ_TIMEOUT, null, JPEG_END)) {
						nextImage = null;
						return;
					}
					prepareNext();
					return;
				}
				if (nextImage == null) {
					throw new UnexpectedError();
				}
				long lastImageByteCount = imageStream.getStreamPosition();
				bufferedStream.reset();
				bufferedStream.skip(lastImageByteCount);
			} catch (Throwable t) {
				nextImage = null;
				throw new UnexpectedError(t);
			} finally {
				if (nextImage == null) {
					dispose();
				}
			}
		}
		
		protected void ensureFirstFrameRead() {
			if(firsFrameRead ) {
				return;
			}
			prepareNext();
			firsFrameRead = true;
		}

		@Override
		public Image next() {
			ensureFirstFrameRead();
			Image result = nextImage;
			prepareNext();
			return result;
		}

		
		@Override
		public boolean hasNext() {
			ensureFirstFrameRead();
			return nextImage != null;
		}

		public void dispose() {
			safeReadingExecutor.shutdown();
			try {
				safeReadingExecutor.awaitTermination(STREAM_READ_TIMEOUT * 2, TimeUnit.MILLISECONDS);
			} catch (InterruptedException e) {
				throw new UnexpectedError(e);
			}
			sequence.streamProvider.releaseInputStream(input);
		}
	}

}
