package com.otk.application.util;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.charset.Charset;
import java.util.Arrays;
import java.util.Date;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.TimeUnit;

import com.otk.application.error.ErrorManager;
import com.otk.application.error.StandardError;
import com.otk.application.error.UnexpectedError;

public abstract class CommandLineInterface<T> {

	private String commandOptions;
	private File executableFile;
	private File workingDirectory;
	private Map<String, String> env;
	private int safeReadingTimeoutMilliseconds = 5000;
	private boolean diagnosticDataCollectionEnabled = true;

	private final Object STARTUP_MUTEX = new Object();
	private CommandExecutor commandExecutor;
	private Thread orchestrationThread;
	private boolean starting = false;
	private ByteArrayOutputStream outputBuffer = createOutputBuffer();
	private ByteArrayOutputStream errorBuffer = createErrorBuffer();
	private ExecutorService safeReadingExecutor;
	private T executionResult;
	private Throwable executionError;

	protected abstract T communicate(Process process);

	public CommandLineInterface(File workingDirectory, File executableFile, String commandOptions) {
		this.commandOptions = commandOptions;
		this.executableFile = executableFile;
		this.workingDirectory = workingDirectory;
	}

	public String getCommandOptions() {
		return commandOptions;
	}

	public void setCommandOptions(String commandOptions) {
		this.commandOptions = commandOptions;
	}

	public File getExecutableFile() {
		return executableFile;
	}

	public void setExecutableFile(File executableFile) {
		this.executableFile = executableFile;
	}

	public File getWorkingDirectory() {
		return workingDirectory;
	}

	public void setWorkingDirectory(File workingDirectory) {
		this.workingDirectory = workingDirectory;
	}

	public Map<String, String> getEnv() {
		return env;
	}

	public void setEnv(Map<String, String> env) {
		this.env = env;
	}

	public int getSafeReadingTimeoutMilliseconds() {
		return safeReadingTimeoutMilliseconds;
	}

	public void setSafeReadingTimeoutMilliseconds(int safeReadingTimeoutMilliseconds) {
		this.safeReadingTimeoutMilliseconds = safeReadingTimeoutMilliseconds;
	}

	protected ByteArrayOutputStream createOutputBuffer() {
		return new ByteArrayOutputStream();
	}

	protected ByteArrayOutputStream createErrorBuffer() {
		return new ByteArrayOutputStream();
	}

	protected ByteArrayOutputStream getOutputBuffer() {
		return outputBuffer;
	}

	protected ByteArrayOutputStream getErrorBuffer() {
		return errorBuffer;
	}

	public CommandExecutor getCommandExecutor() {
		return commandExecutor;
	}

	public ExecutorService getSafeReadingExecutor() {
		return safeReadingExecutor;
	}

	protected boolean isDiagnosticDataCollectionEnabled() {
		return diagnosticDataCollectionEnabled;
	}

	protected void setDiagnosticDataCollectionEnabled(boolean diagnosticDataCollectionEnabled) {
		this.diagnosticDataCollectionEnabled = diagnosticDataCollectionEnabled;
	}

	protected void handleOrchestrationError(final Throwable t) {
		executionError = t;
	}

	public boolean isExecuting() {
		if (starting) {
			return true;
		}
		if (orchestrationThread != null) {
			if (orchestrationThread.isAlive()) {
				return true;
			}
		}
		if (commandExecutor != null) {
			if (commandExecutor.isProcessAlive()) {
				return true;
			}
		}
		return false;
	}

	public T execute() {
		executeAsynchronously();
		waitForExecutionEnd();
		Throwable error = getExecutionError();
		if (error != null) {
			if (isDiagnosticDataCollectionEnabled()) {
				throw new DiagnosticData(error, collectAvailableDiagnosticData());
			} else {
				ErrorManager.rethrow(error);
			}
		}
		return executionResult;
	}

	public void executeAsynchronously() {
		synchronized (STARTUP_MUTEX) {
			if (isExecuting()) {
				throw new UnexpectedError("The command already executing");
			}
			starting = true;
			try {
				orchestrationThread = createOrchestrationThread();
				orchestrationThread.start();
			} finally {
				starting = false;
			}
		}
	}

	public void waitForExecutionEnd() {
		while (isExecuting()) {
			MiscUtils.relieveCPU();
		}
	}

	public void requestExecutionInterruption() {
		if (orchestrationThread != null) {
			orchestrationThread.interrupt();
		}
	}

	public T getExecutionResult() {
		return executionResult;
	}

	public Throwable getExecutionError() {
		return executionError;
	}

	protected CommandExecutor createCommandExecutor() {
		CommandExecutor result = new CommandExecutor();
		result.setCommandLine(CommandExecutor.quoteArgument(getExecutableFile().getPath()) + " " + getCommandOptions());
		result.setWorkingDir(getWorkingDirectory());
		result.setEnv(getEnv());
		return result;
	}

	protected Thread createOrchestrationThread() {
		return new Thread(
				MiscUtils.formatThreadName(CommandLineInterface.class, "Executor of " + CommandLineInterface.this)) {
			{
				setDaemon(true);
			}

			@Override
			public void run() {
				try {
					orchestrate();
				} catch (Throwable t) {
					handleOrchestrationError(t);
				}
			}
		};
	}

	protected void orchestrate() {
		commandExecutor = createCommandExecutor();
		commandExecutor.startProcess();
		safeReadingExecutor = createSafeReadingExecutor();
		try {
			executionResult = communicate(commandExecutor.getLaunchedProcess());
		} finally {
			safeReadingExecutor.shutdownNow();
			commandExecutor.disconnectProcess();
			if (commandExecutor.isProcessAlive()) {
				commandExecutor.killProcess();
			}
		}
	}

	protected ExecutorService createSafeReadingExecutor() {
		return Executors.newSingleThreadExecutor(new ThreadFactory() {
			@Override
			public Thread newThread(Runnable runnable) {
				return createSafeReadingThread(runnable);
			}
		});
	}

	protected Thread createSafeReadingThread(Runnable runnable) {
		Thread result = new Thread(runnable, MiscUtils.formatThreadName(CommandLineInterface.class, "Safe Reader"));
		result.setDaemon(true);
		return result;
	}

	protected int acquireSafelyOutputDataUntil(Process process, byte[]... endings) {
		try {
			@SuppressWarnings("rawtypes")
			Listener<Future> onTimeout = null;
			InputStream inputStream = process.getInputStream();
			OutputStream outputStream = outputBuffer;
			return MiscUtils.readUntil(safeReadingExecutor, inputStream, outputStream, safeReadingTimeoutMilliseconds,
					onTimeout, endings);
		} catch (IOException e) {
			throw new StandardError("Output stream reading error: " + e.toString(), e);
		}
	}

	protected int acquireSafelyErrorDataUntil(Process process, byte[]... endings) {
		try {
			@SuppressWarnings("rawtypes")
			Listener<Future> onTimeout = null;
			InputStream inputStream = process.getErrorStream();
			OutputStream outputStream = errorBuffer;
			return MiscUtils.readUntil(safeReadingExecutor, inputStream, outputStream, safeReadingTimeoutMilliseconds,
					onTimeout, endings);
		} catch (IOException e) {
			throw new StandardError("Error stream reading error: " + e.toString(), e);
		}
	}

	protected void acquireSafelyAvailableOutputData(Process process) {
		int initialTimeout = getSafeReadingTimeoutMilliseconds();
		setSafeReadingTimeoutMilliseconds(1000);
		try {
			acquireSafelyOutputDataUntil(process, new byte[0][0]);
		} catch (Throwable ignore) {
		} finally {
			setSafeReadingTimeoutMilliseconds(initialTimeout);
		}
	}

	protected void acquireSafelyAvailableErrorData(Process process) {
		int initialTimeout = getSafeReadingTimeoutMilliseconds();
		setSafeReadingTimeoutMilliseconds(1000);
		try {
			acquireSafelyErrorDataUntil(process, new byte[0][0]);
		} catch (Throwable ignore) {
		} finally {
			setSafeReadingTimeoutMilliseconds(initialTimeout);
		}
	}

	protected int acquireSafelyOutputDataUntil(Process process, String... endings) {
		byte[][] endingsAsBytes = new byte[endings.length][];
		for (int i = 0; i < endings.length; i++) {
			endingsAsBytes[i] = endings[i].getBytes(getCharset());
		}
		return acquireSafelyOutputDataUntil(process, endingsAsBytes);
	}

	protected int acquireSafelyErrorDataUntil(Process process, String... endings) {
		byte[][] endingsAsBytes = new byte[endings.length][];
		for (int i = 0; i < endings.length; i++) {
			endingsAsBytes[i] = endings[i].getBytes(getCharset());
		}
		return acquireSafelyErrorDataUntil(process, endingsAsBytes);
	}

	protected void provideSafelyInput(Process process, byte[] bytes) {
		try {
			process.getOutputStream().write(bytes);
			process.getOutputStream().flush();
		} catch (IOException e) {
			throw new StandardError("Input stream writing error: " + e.toString(), e);
		}
	}

	protected void provideSafelyInput(Process process, String s) {
		provideSafelyInput(process, s.getBytes(getCharset()));
	}

	protected void waitSafelyFor(final Process process, int timeoutMilliSeconds) {
		try {
			if (!process.waitFor(timeoutMilliSeconds, TimeUnit.MILLISECONDS)) {
				throw new StandardError("Process interruption timeout");
			}
		} catch (InterruptedException e) {
			throw new UnexpectedError(e);
		}
	}

	protected String collectAvailableDiagnosticData() {
		try {
			String result = "";
			Process process = commandExecutor.getLaunchedProcess();
			result += "- Options: " + Arrays.toString(CommandExecutor.splitArguments(commandOptions));
			if (process == null) {
				result += "\n- Process status: failed to start";
			} else {
				try {
					result += "\n- Process status: terminated with exit value: "
							+ commandExecutor.getLaunchedProcess().exitValue();
				} catch (IllegalThreadStateException e) {
					result += "\n- Process status: not terminated";
				}
				acquireSafelyAvailableOutputData(process);
				acquireSafelyAvailableErrorData(process);
				String outputLogs = new String(getOutputBuffer().toByteArray(), getCharset());
				String errorLogs = new String(getErrorBuffer().toByteArray(), getCharset());
				outputLogs = cleanLogs(outputLogs);
				result += "\n- Diagnostic output logs: " + ((outputLogs.length() > 0) ? outputLogs : "<NONE>");
				errorLogs = cleanLogs(errorLogs);
				result += "\n- Diagnostic error logs: " + ((errorLogs.length() > 0) ? errorLogs : "<NONE>");
			}
			return result;
		} catch (Throwable t) {
			return "Diagnostic data collection failed (" + t.toString() + ")";
		}
	}

	protected String cleanLogs(String outputLogs) {
		return outputLogs;
	}

	protected Charset getCharset() {
		return Charset.defaultCharset();
	}

	@Override
	public String toString() {
		return "CommandLineInterface [commandOptions=" + commandOptions + "]";
	}

	public static class DiagnosticData extends RuntimeException {

		private static final long serialVersionUID = 1L;

		public DiagnosticData(Throwable cause, String diagnosticDataString) {
			super(cause.toString() + "\nContext: " + diagnosticDataString, cause);
		}

	}

	public static void main(String[] args) {
		CommandLineInterface<Date> dateInterface = new CommandLineInterface<Date>(null, new File("cmd"), "/C datee") {

			@Override
			protected CommandExecutor createCommandExecutor() {
				CommandExecutor result = super.createCommandExecutor();
				return result;
			}

			@SuppressWarnings("deprecation")
			@Override
			protected Date communicate(Process process) {
				acquireSafelyOutputDataUntil(process, "The current date is: ".getBytes());

				getOutputBuffer().reset();
				acquireSafelyOutputDataUntil(process, "/".getBytes());
				Integer date = Integer.valueOf(getOutputBuffer().toString().replace("/", "").trim());

				getOutputBuffer().reset();
				acquireSafelyOutputDataUntil(process, "/".getBytes());
				Integer month = Integer.valueOf(getOutputBuffer().toString().replace("/", "").trim());

				getOutputBuffer().reset();
				acquireSafelyOutputDataUntil(process, "\n".getBytes());
				Integer year = Integer.valueOf(getOutputBuffer().toString().trim());

				return new Date(year, month, date);
			}
		};
		Date currentDate = dateInterface.execute();
		System.out.println("date according to the command line: " + currentDate);
	}

}
