package com.otk.application.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.nio.channels.FileChannel;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import javax.swing.JPanel;

import com.otk.application.error.StandardError;
import com.otk.application.error.UnexpectedError;
import com.otk.application.window.component.crud.AbstractPojoWrapperCrudItem;
import com.otk.application.window.component.crud.ICrudItemControlListener;
import com.otk.application.window.component.crud.SingleTypeCrudPanel.SingleTypeCrudModel;

public class FilePacker {

	public static void main(String[] args) throws Exception {
		String usageText = "Expected arguments: (--pack|--unpack) <directory> <packFile> <catalogFile>";
		if (args.length != 4) {
			throw new StandardError("Invalid arguments: " + Arrays.toString(args) + ".\n" + usageText);
		}
		String action = args[0];
		File directory = new File(args[1]);
		File packFile = new File(args[2]);
		File catalogFile = new File(args[3]);
		if ("--pack".equals(action)) {
			packFiles(packFile, catalogFile, directory, directory.list());
		} else if ("--unpack".equals(action)) {
			unpackFiles(packFile, catalogFile, directory);
		} else {
			throw new StandardError("Invalid arguments: " + Arrays.toString(args) + ".\n" + usageText);
		}
	}

	public static void packFiles(File targetPackFile, File targetCatalogFile, File baseDirectory, String... filePaths) {
		Catalog catalog;
		FileOutputStream packFileOutputStream = null;
		try {
			packFileOutputStream = new FileOutputStream(targetPackFile);
			catalog = createPackFile(packFileOutputStream, baseDirectory, filePaths);
		} catch (Throwable t) {
			throw new UnexpectedError(t);
		} finally {
			try {
				packFileOutputStream.close();
			} catch (Throwable ignore) {
			}
		}
		saveCatalog(catalog, targetCatalogFile);
	}

	private static String[] expandFilePaths(File baseDirectory, String[] filePaths) {
		List<String> result = new ArrayList<String>();
		for (String filePath : filePaths) {
			File file = new File(baseDirectory, filePath);
			if (file.isFile()) {
				result.add(filePath);
			} else if (file.isDirectory()) {
				for (String childRelativeFilePath : expandFilePaths(file, file.list()))
					result.add(filePath + "/" + childRelativeFilePath);
			} else {
				throw new UnexpectedError("This path does not denote a file or a directory");
			}
		}
		return result.toArray(new String[result.size()]);
	}

	private static void saveCatalog(Catalog catalog, File targetCatalogFile) {
		try {
			FileUtils.write(targetCatalogFile, catalog.save(), false);
		} catch (Exception e) {
			throw new UnexpectedError(e);
		}
	}

	public static Catalog createPackFile(FileOutputStream packOutputStream, File baseDirectory,
			String[] filePaths) {
		filePaths = expandFilePaths(baseDirectory, filePaths);
		Catalog catalog = new Catalog();
		FileChannel outputChannel = null;
		try {
			outputChannel = packOutputStream.getChannel();
			long currentPosition = 0;
			for (String filePath : filePaths) {
				File file = new File(baseDirectory, filePath);
				FileInputStream inputStream = null;
				FileChannel inputChannel = null;
				try {
					inputStream = new FileInputStream(file);
					inputChannel = inputStream.getChannel();
					TrackCrudItem trackCrudItem = new TrackCrudItem();
					catalog.add(trackCrudItem);
					Track track = trackCrudItem.getWrapped();
					track.filePath = filePath;
					track.position = currentPosition;
					track.size = inputChannel.size();
					currentPosition += track.size;
					outputChannel.transferFrom(inputChannel, track.position, track.size);
				} finally {
					try {
						inputChannel.close();
					} catch (Throwable ignore) {
					}
					try {
						inputStream.close();
					} catch (Throwable ignore) {
					}
				}
			}
		} catch (Throwable t) {
			throw new UnexpectedError(t);
		} finally {
			try {
				outputChannel.close();
			} catch (Throwable ignore) {
			}
		}
		return catalog;
	}

	public static void unpackFiles(File sourcePackFile, File sourceCatalogFile, File baseDirectory) {
		Catalog catalog = loadCatalog(sourceCatalogFile);
		FileInputStream packFileInputStream = null;
		try {
			packFileInputStream = new FileInputStream(sourcePackFile);
			recreateCatalogFiles(catalog, packFileInputStream, baseDirectory);
		} catch (Throwable t) {
			throw new UnexpectedError(t);
		} finally {
			try {
				packFileInputStream.close();
			} catch (Throwable ignore) {
			}
		}
	}

	public static void recreateCatalogFiles(Catalog catalog, FileInputStream packInputStream, File baseDirectory) {
		FileChannel inputChannel = null;
		try {
			inputChannel = packInputStream.getChannel();
			for (Track track : catalog.getUnwrappedList()) {
				File file = new File(baseDirectory, track.filePath);
				FileUtils.createAncestors(file);
				FileOutputStream outputStream = null;
				FileChannel outputChannel = null;
				try {
					outputStream = new FileOutputStream(file);
					outputChannel = outputStream.getChannel();
					inputChannel.transferTo(track.position, track.size, outputChannel);
				} finally {
					try {
						outputChannel.close();
					} catch (Throwable ignore) {
					}
					try {
						outputStream.close();
					} catch (Throwable ignore) {
					}
				}
			}
		} catch (Throwable t) {
			throw new UnexpectedError(t);
		} finally {
			try {
				inputChannel.close();
			} catch (Throwable ignore) {
			}
		}
	}

	private static Catalog loadCatalog(File sourceCatalogFile) {
		Catalog catalog = new Catalog();
		try {
			catalog.load(FileUtils.read(sourceCatalogFile));
		} catch (Exception e) {
			throw new UnexpectedError(e);
		}
		return catalog;
	}

	public static class Track {
		public String filePath;
		public long position;
		public long size;
	}

	public static class TrackCrudItem extends AbstractPojoWrapperCrudItem<Track> {

		@Override
		protected Track createWrapped() {
			return new Track();
		}

		@Override
		protected void createFieldControls(JPanel panel, ICrudItemControlListener listener) {
			throw new UnsupportedOperationException();
		}

	}

	public static class Catalog extends SingleTypeCrudModel<Track> {

		public Catalog() {
			super(TrackCrudItem.class);
		}

	}
}
