package com.otk.application.util;

import java.awt.Component;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.nio.channels.FileChannel;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.JFileChooser;
import javax.swing.filechooser.FileFilter;

import com.otk.application.error.AbstractApplicationError;
import com.otk.application.error.UnexpectedError;

import com.otk.application.translation.T;
import com.otk.application.util.FilePacker.Catalog;

public class FileUtils {

	private static File lastBrowsedDirectory;

	public static String read(File file) throws Exception {
		return new String(readBinary(file));
	}

	public static String read(InputStream in) throws Exception {
		return new String(readBinary(in));
	}

	public static byte[] readBinary(InputStream in) throws Exception {
		try {
			ByteArrayOutputStream buffer = new ByteArrayOutputStream();
			int nRead;
			byte[] data = new byte[16384];
			while ((nRead = in.read(data, 0, data.length)) != -1) {
				buffer.write(data, 0, nRead);
			}
			buffer.flush();
			return buffer.toByteArray();
		} catch (IOException e) {
			throw new Exception("Error while reading input stream: " + e.getMessage(), e);
		}
	}

	public static byte[] readBinary(File file) throws Exception {
		FileInputStream in = null;
		try {
			in = new FileInputStream(file);
			return readBinary(in);
		} catch (IOException e) {
			throw new Exception("Unable to read file : '" + file.getAbsolutePath() + "': " + e.getMessage(), e);
		} finally {
			if (in != null) {
				try {
					in.close();
				} catch (IOException e) {
				}
			}
		}
	}

	public static void write(File file, String text, boolean append) throws Exception {
		writeBinary(file, text.getBytes(), append);
	}

	public static void writeBinary(File file, byte[] bytes, boolean append) throws Exception {
		FileOutputStream out = null;
		try {
			out = new FileOutputStream(file, append);
			out.write(bytes);
			out.flush();
			out.close();
		} catch (IOException e) {
			throw new Exception("Unable to write file : '" + file.getAbsolutePath() + "': " + e.getMessage(), e);
		} finally {
			if (out != null) {
				try {
					out.close();
				} catch (IOException e) {
				}
			}
		}

	}

	public static void historizeFile(String filePath, int historySize) {
		for (int i = (historySize - 1); i >= 0; i--) {
			File file;
			if (i == 0) {
				file = new File(filePath);
			} else {
				file = new File(getHistorizedFilePath(filePath, i));
			}
			if (file.exists()) {
				try {
					if (i == (historySize - 1)) {
						delete(file);
					} else {
						File nextRotatedFile = new File(getHistorizedFilePath(filePath, (i + 1)));
						try {
							rename(file, nextRotatedFile.getName());
						} catch (Exception e) {
							copy(file, nextRotatedFile);
							delete(file);
						}
					}
				} catch (Exception e) {
					throw new UnexpectedError(e);
				}
			}
		}
	}

	public static String getHistorizedFilePath(String filePath, int number) {
		String fileNameExtension = getFileNameExtension(new File(filePath).getName());
		if ((fileNameExtension != null) && (fileNameExtension.length() > 0)) {
			String filePathWithoutExtension = removeFileNameExtension(filePath);
			return filePathWithoutExtension + "-" + number + "." + fileNameExtension;
		} else {
			return filePath + "-" + number;
		}
	}

	public static boolean isHistorizedFilePath(String baseFilePath, String historizedFilePath) {
		String fileNameExtension = getFileNameExtension(new File(baseFilePath).getName());
		if ((fileNameExtension != null) && (fileNameExtension.length() > 0)) {
			String filePathWithoutExtension = removeFileNameExtension(baseFilePath);
			return historizedFilePath.matches(filePathWithoutExtension + "-.*\\." + fileNameExtension);
		} else {
			return historizedFilePath.matches(baseFilePath + "-.*");
		}
	}

	public static String removeFileNameExtension(String fileName) {
		String extension = getFileNameExtension(fileName);
		if (extension.length() > 0) {
			return fileName.substring(0, fileName.length() - ("." + extension).length());
		} else {
			return fileName;
		}
	}

	public static String getFileNameExtension(String fileName) {
		int lastDotIndex = fileName.lastIndexOf(".");
		if (lastDotIndex == -1) {
			return "";
		} else if (lastDotIndex == 0) {
			return "";
		} else {
			return fileName.substring(lastDotIndex + 1);
		}
	}

	public static String askFilePath(Component parent, String actionTitle, final String[] extensions,
			File defaultFile) {
		FileFilter fileFilter = new FileFilter() {
			@Override
			public String getDescription() {
				StringBuilder result = new StringBuilder(T.t("Files") + " (");
				if (extensions == null) {
					result.append("*.*");
				} else {
					boolean first = true;
					for (String ext : extensions) {
						if (!first) {
							result.append(", ");
						}
						result.append("*." + ext);
						first = false;
					}
				}
				result.append(")");
				return result.toString();
			}

			@Override
			public boolean accept(File f) {
				if (f.isDirectory()) {
					return true;
				}
				if (extensions == null) {
					return true;
				}
				if (FileUtils.hasFileNameExtension(f.getName(), extensions)) {
					return true;
				}
				return false;
			}
		};
		return askPath(parent, actionTitle, defaultFile, false, fileFilter);
	}

	public static String askDirectoryPath(Component parent, String actionTitle, File defaultFile) {
		return askPath(parent, actionTitle, defaultFile, true, null);

	}

	public static String askPath(Component parent, String actionTitle, File defaultFile, boolean directory,
			FileFilter fileFilter) {
		final JFileChooser fileChooser = new JFileChooser();
		if (directory) {
			fileChooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
		}
		if (fileFilter != null) {
			fileChooser.setFileFilter(fileFilter);
		}
		if (defaultFile != null) {
			fileChooser.setSelectedFile(defaultFile.getAbsoluteFile());
		} else {
			if (lastBrowsedDirectory != null) {
				fileChooser.setCurrentDirectory(lastBrowsedDirectory);
			}
		}
		int returnVal = fileChooser.showDialog(parent, actionTitle);
		if (returnVal != JFileChooser.APPROVE_OPTION) {
			return null;
		}
		File file = fileChooser.getSelectedFile();
		String result = file.getPath();
		String relativepath = getRelativePath(file, new File("."));
		if (relativepath != null) {
			result = relativepath;
		}
		result = getOSAgnosticFilePath(result);
		lastBrowsedDirectory = fileChooser.getCurrentDirectory();
		return result;
	}

	public static String getOSAgnosticFilePath(String path) {
		return path.replace(File.separator, "/");
	}

	public static byte[] requireBytes(File file) {
		try {
			return FileUtils.readBinary(file);
		} catch (Exception e) {
			throw new UnexpectedError(e);
		}
	}

	public static void copy(File src, File dst) throws Exception {
		copy(src, dst, true);
	}

	public static void copy(File src, File dst, boolean recusrsively) throws Exception {
		copy(src, dst, recusrsively, null, null);
	}

	public static void copy(File src, File dst, boolean recusrsively, FilenameFilter srcFilesFilter,
			Listener<Pair<File, Exception>> errorHandler) throws Exception {
		try {
			if ((srcFilesFilter != null) && !srcFilesFilter.accept(src.getParentFile(), src.getName())) {
				return;
			}
			if (src.isDirectory()) {
				try {
					createDirectory(dst);
				} catch (Exception e) {
					if (errorHandler != null) {
						errorHandler.handle(new Pair<File, Exception>(src, e));
					} else {
						throw e;
					}
				}
				if (recusrsively) {
					for (File srcChild : src.listFiles()) {
						copy(srcChild, new File(dst, srcChild.getName()), recusrsively, srcFilesFilter, errorHandler);
					}
				}
			} else if (src.isFile()) {
				try {
					copyFileUsingNIOChannels(src, dst);
				} catch (Exception e) {
					if (errorHandler != null) {
						errorHandler.handle(new Pair<File, Exception>(src, e));
					} else {
						throw e;
					}
				}
			} else {
				throw new Exception("File not found: '" + src + "'", null);
			}
		} catch (Exception e) {
			throw new Exception("Unable to copy resource: '" + src.getAbsolutePath() + "' > '" + dst.getAbsolutePath()
					+ "': " + e.getMessage(), e);
		}
	}

	public static void copyFileUsingIOStreams(File source, File dest) throws Exception {
		InputStream is = null;
		OutputStream os = null;
		try {
			is = new FileInputStream(source);
			os = new FileOutputStream(dest);
			byte[] buffer = new byte[1024];
			int length;
			while ((length = is.read(buffer)) > 0) {
				os.write(buffer, 0, length);
			}
		} finally {
			try {
				is.close();
			} catch (Throwable ignore) {
			}
			try {
				os.close();
			} catch (Throwable ignore) {
			}

		}
	}

	public static void copyFileUsingNIOChannels(File source, File dest) throws Exception {
		FileInputStream is = null;
		FileOutputStream os = null;
		FileChannel sourceChannel = null;
		FileChannel destChannel = null;
		try {
			is = new FileInputStream(source);
			os = new FileOutputStream(dest);
			sourceChannel = is.getChannel();
			destChannel = os.getChannel();
			destChannel.transferFrom(sourceChannel, 0, sourceChannel.size());
		} finally {
			try {
				sourceChannel.close();
			} catch (Throwable ignore) {
			}
			try {
				destChannel.close();
			} catch (Throwable ignore) {
			}
			try {
				is.close();
			} catch (Throwable ignore) {
			}
			try {
				os.close();
			} catch (Throwable ignore) {
			}
		}
	}

	public static void copyFileUsingJava7FilesAPI(File source, File dest) throws Exception {
		Files.copy(source.toPath(), dest.toPath());
	}

	public static File relativizeFile(File ancestor, File file) {
		try {
			ancestor = ancestor.getCanonicalFile();
			file = file.getCanonicalFile();
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
		if (!isAncestor(ancestor, file)) {
			return null;
		}
		String relativePath = file.getPath().substring(ancestor.getPath().length(), file.getPath().length());
		if (relativePath.startsWith("/") || relativePath.startsWith("\\")) {
			relativePath = relativePath.substring(1);
		}
		return new File(relativePath);
	}

	public static String getRelativePath(File child, File ancestor) {
		File relativeFile = relativizeFile(ancestor, child);
		if (relativeFile == null) {
			return null;
		}
		return relativeFile.getPath();
	}

	public static boolean canonicallyEquals(File file1, File file2) {
		try {
			return file1.getCanonicalFile().equals(file2.getCanonicalFile());
		} catch (IOException e) {
			throw new RuntimeException(e.toString(), e);
		}
	}

	public static File getCanonicalParent(File file) {
		try {
			return file.getCanonicalFile().getParentFile();
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	public static void delete(File file) throws Exception {
		delete(file, null, null);
	}

	public static void delete(File file, FilenameFilter filter, Listener<Pair<File, Exception>> errorHandler)
			throws Exception {
		if (file.isDirectory()) {
			for (File childFile : file.listFiles(filter)) {
				delete(childFile, filter, errorHandler);
			}
			if (file.listFiles().length > 0) {
				return;
			}
		}
		boolean success;
		try {
			success = file.delete();
			if (!success) {
				throw new Exception("System error");
			}
		} catch (Exception e) {
			e = new Exception("Failed to delete resource: '" + file.getAbsolutePath() + "': " + e.getMessage(), e);
			if (errorHandler != null) {
				errorHandler.handle(new Pair<File, Exception>(file, e));
			} else {
				throw e;
			}
		}
	}

	public static void rename(File file, String destFileName) throws Exception {
		try {
			if (new File(destFileName).getParent() != null) {
				throw new Exception("Destination file name is not is not a local name: '" + destFileName + "'");
			}
			File destFile = new File(file.getParent(), destFileName);
			boolean success = file.renameTo(destFile);
			if (!success) {
				throw new Exception("System error");
			}
		} catch (Exception e) {
			throw new Exception("Failed to rename resource: '" + file.getAbsolutePath() + "' to '" + destFileName
					+ "': " + e.getMessage(), e);
		}
	}

	public static boolean hasFileNameExtension(String fileName, String[] extensions) {
		for (String ext : extensions) {
			if (ext.toLowerCase().equals(getFileNameExtension(fileName).toLowerCase())) {
				return true;
			}
		}
		return false;
	}

	public static boolean isAncestor(File ancestor, File file) {
		File mayBeAncestor = getCanonicalParent(file);
		while (true) {
			if (mayBeAncestor == null) {
				return false;
			}
			if (canonicallyEquals(mayBeAncestor, ancestor)) {
				return true;
			}
			mayBeAncestor = getCanonicalParent(mayBeAncestor);
		}
	}

	public static void createFile(File file) throws Exception {
		if (file.isFile()) {
			delete(file);
		}
		try {
			if (!file.createNewFile()) {
				throw new UnexpectedError("System error");
			}
		} catch (IOException e) {
			throw new Exception("Failed to create the file '" + file + "': " + e.toString(), e);
		}

	}

	public static void createDirectory(File dir) throws Exception {
		if (dir.isDirectory()) {
			return;
		}
		try {
			if (!dir.mkdir()) {
				throw new UnexpectedError("System error");
			}
		} catch (Exception e) {
			throw new Exception("Failed to create directory: '" + dir + "': " + e.toString(), e);
		}

	}

	public static File createTemporaryFile() throws Exception {
		return createTemporaryFile("tmp");
	}

	public static File createTemporaryFile(String extension) throws Exception {
		return File.createTempFile("file-", "." + extension);
	}

	public static File createTemporaryDirectory() throws Exception {
		File result = File.createTempFile("directory-", ".tmp");
		delete(result);
		createDirectory(result);
		return result;
	}

	public static void withFileRestoredInCaseOfError(File file, Runnable runnable) {
		File backupFile;
		try {
			backupFile = FileUtils.createTemporaryFile();
			FileUtils.copy(file, backupFile);
		} catch (Exception e) {
			throw new UnexpectedError(e);
		}
		try {
			runnable.run();
		} catch (Exception e) {
			try {
				FileUtils.copy(backupFile, file);
			} catch (Exception e1) {
				throw new UnexpectedError(e1);
			}
			AbstractApplicationError.rethrow(e);
		} finally {
			try {
				FileUtils.delete(backupFile);
			} catch (Exception e) {
				throw new UnexpectedError(e);
			}
		}
	}

	public static List<File> grep(Pattern regex, File fileOrDir, boolean recursive,
			FilenameFilter directoryContentFilter) throws Exception {
		List<File> result = new ArrayList<File>();
		if (fileOrDir.isDirectory()) {
			for (File child : fileOrDir.listFiles(directoryContentFilter)) {
				result.addAll(grep(regex, child, recursive, directoryContentFilter));
			}
		} else if (fileOrDir.isFile()) {
			String fileContent = read(fileOrDir);
			Matcher matcher = regex.matcher(fileContent);
			if (matcher.find()) {
				result.add(fileOrDir);
			}
		}
		return result;
	}

	public static void createAncestors(File file) throws Exception {
		File dir = file.getParentFile();
		if (dir != null) {
			if (!dir.isDirectory()) {
				createAncestors(dir);
				createDirectory(dir);
			}
		}
	}

	public static File createTemporaryDirectoryFromClassPath(Class<?> classInResourcesPackage) {
		URL resourceListUrl = classInResourcesPackage.getResource("_resourceList.txt");
		if (resourceListUrl == null) {
			throw new UnexpectedError("Classpath resource not found: "
					+ classInResourcesPackage.getPackage().getName().replace('.', '/') + "/_resourceList.txt");
		}
		InputStream resourceListInput = null;
		String resourceListLines;
		try {
			resourceListInput = resourceListUrl.openStream();
			resourceListLines = read(resourceListInput);
		} catch (Exception e) {
			throw new UnexpectedError(e);
		} finally {
			if (resourceListInput != null) {
				try {
					resourceListInput.close();
				} catch (IOException e) {
					throw new UnexpectedError(e);
				}
			}
		}
		File result;
		try {
			result = createTemporaryDirectory();
		} catch (Exception e) {
			throw new UnexpectedError(e);
		}
		Runtime.getRuntime().addShutdownHook(new Thread() {
			@Override
			public void run() {
				try {
					delete(result);
				} catch (Exception e) {
					throw new UnexpectedError(e);
				}
			}
		});
		for (String resourcePath : TextUtils.splitLines(resourceListLines)) {
			InputStream resourceStream = null;
			FileOutputStream fileStream = null;
			try {
				resourceStream = classInResourcesPackage.getResourceAsStream(resourcePath);
				if (resourceStream == null) {
					throw new UnexpectedError("Classpath resource not found: "
							+ classInResourcesPackage.getPackage().getName().replace('.', '/') + "/" + resourcePath);
				}
				File file = new File(result, resourcePath);
				createAncestors(file);
				fileStream = new FileOutputStream(file);
				MiscUtils.readAll(resourceStream, fileStream);
			} catch (Exception e) {
				throw new UnexpectedError(e);
			} finally {
				if (resourceStream != null) {
					try {
						resourceStream.close();
					} catch (IOException ignore) {
					}
				}
				if (fileStream != null) {
					try {
						fileStream.close();
					} catch (IOException ignore) {
					}
				}
			}
		}
		return result;
	}

	public static File createTemporaryDirectoryFromFilePack(File packFile, InputStream catalogInputStream) {
		try {
			File result = createTemporaryDirectory();
			Runtime.getRuntime().addShutdownHook(new Thread() {
				@Override
				public void run() {
					try {
						delete(result);
					} catch (Exception e) {
						throw new UnexpectedError(e);
					}
				}
			});
			Catalog catalog = new Catalog();
			catalog.load(read(catalogInputStream));
			FileInputStream packInputStream = null;
			try {
				packInputStream = new FileInputStream(packFile);
				FilePacker.recreateCatalogFiles(catalog, packInputStream, result);
			} finally {
				try {
					packInputStream.close();
				} catch (Throwable ignore) {
				}
			}
			return result;
		} catch (Exception e) {
			throw new UnexpectedError(e);
		}
	}

	public static File createTemporaryFileFromURL(URL url) throws Exception {
		InputStream inputStream = null;
		try {
			inputStream = url.openStream();
			byte[] content = readBinary(inputStream);
			File tmpFile = createTemporaryFile();
			writeBinary(tmpFile, content, false);
			return tmpFile;
		} finally {
			if (inputStream != null) {
				try {
					inputStream.close();
				} catch (Throwable ignore) {
				}
			}
		}
	}

	public static void requireDirectory(File dir) {
		if (!dir.exists()) {
			try {
				FileUtils.createDirectory(dir);
			} catch (Exception e) {
				throw new UnexpectedError();
			}
		}
	}

}
