package com.otk.application.util;

import java.awt.Component;
import java.awt.Font;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.text.DateFormat;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.otk.application.error.UnexpectedError;

public class TextUtils {

	public static final String[] NEW_LINE_SEQUENCES = new String[] { "\r\n", "\n", "\r" };
	public static final Pattern SPECIAL_REGEX_CHARS = Pattern.compile("[{}()\\[\\].+*?^$\\\\|]");

	public static String escapeRegex(String str) {
		return SPECIAL_REGEX_CHARS.matcher(str).replaceAll("\\\\$0");
	}

	public static String toText(Properties properties) {
		ByteArrayOutputStream out = new ByteArrayOutputStream();
		try {
			properties.store(out, null);
		} catch (IOException e) {
			throw new UnexpectedError(e);
		}
		return out.toString();
	}

	public static String[] splitLines(String s) {
		if (s.length() == 0) {
			return new String[0];
		}
		return s.split(getNewLineRegex(), -1);
	}

	public static String getNewLineRegex() {
		return stringJoin(Arrays.asList(TextUtils.NEW_LINE_SEQUENCES), "|");
	}

	public static String escapeHTML(String string) {
		return escapeHTML(string, false);
	}

	public static String escapeHTML(String string, boolean convertNewLinesToHTML) {
		string = string.replaceAll(getNewLineRegex(), "\n");
		StringBuffer sb = new StringBuffer(string.length());
		int len = string.length();
		char c;

		for (int i = 0; i < len; i++) {
			c = string.charAt(i);
			if (c == '"')
				sb.append("&quot;");
			else if (c == '&')
				sb.append("&amp;");
			else if (c == '<')
				sb.append("&lt;");
			else if (c == '>')
				sb.append("&gt;");
			else if (c == '\n')
				// Handle Newline
				if (convertNewLinesToHTML) {
					sb.append("<br/>");
				} else {
					sb.append(c);
				}
			else {
				int ci = 0xffff & c;
				if (ci < 160)
					// nothing special only 7 Bit
					sb.append(c);
				else {
					// Not 7 Bit use the unicode system
					sb.append("&#");
					sb.append(new Integer(ci).toString());
					sb.append(';');
				}
			}
		}
		return sb.toString();
	}

	public static String truncateNicely(String s, int length) {
		if (s.length() > length) {
			s = s.substring(0, length - 3) + "...";
		}
		return s;
	}

	public static DateFormat getDefaultDateFormat() {
		return DateFormat.getDateTimeInstance(DateFormat.SHORT, DateFormat.MEDIUM);
	}

	public static String escapeNewLines(String s) {
		StringBuilder result = new StringBuilder();
		char lastC = 0;
		for (char currentC : s.toCharArray()) {
			char standardC = standardizeNewLineSequences(lastC, currentC);
			try {
				if (standardC == 0) {
					continue;
				} else if (standardC == '\n') {
					result.append("\\n");
				} else if (standardC == '\\') {
					result.append("\\\\");
				} else {
					result.append(standardC);
				}
			} finally {
				lastC = currentC;
			}
		}
		return result.toString();
	}

	public static char standardizeNewLineSequences(char lastC, char c) {
		for (String newLineSequence : TextUtils.NEW_LINE_SEQUENCES) {
			if (newLineSequence.equals("" + lastC + c)) {
				return 0;
			}
		}
		for (String newLineSequence : TextUtils.NEW_LINE_SEQUENCES) {
			if (newLineSequence.startsWith("" + c)) {
				return '\n';
			}
		}
		return c;
	}

	public static String unescapeNewLine(String s) {
		StringBuilder result = new StringBuilder();
		boolean escaping = false;
		for (char c : s.toCharArray()) {
			if (escaping) {
				if (c == '\\') {
					result.append('\\');
				} else if (c == 'n') {
					result.append('\n');
				} else {
					result.append('\\');
					result.append(c);
				}
				escaping = false;
			} else {
				if (c == '\\') {
					escaping = true;
				} else {
					result.append(c);
				}
			}
		}
		return result.toString();
	}

	public static String stringJoin(Collection<?> col, String delim) {
		StringBuilder sb = new StringBuilder();
		Iterator<?> iter = col.iterator();
		if (iter.hasNext())
			sb.append(iter.next().toString());
		while (iter.hasNext()) {
			sb.append(delim);
			sb.append(iter.next().toString());
		}
		return sb.toString();
	}

	public static String padRight(String s, int len, char c) {
		StringBuilder result = new StringBuilder(s);
		for (int i = s.length(); i < len; i++) {
			result.append(c);
		}
		return result.toString();
	}

	public static String padLeft(String s, int len, char c) {
		StringBuilder result = new StringBuilder(s);
		for (int i = s.length(); i < len; i++) {
			result.insert(0, c);
		}
		return result.toString();
	}

	public static String padBothSides(String s, int len, char c) {
		int padLength = MathUtils.round((len - s.length()) / 2.0);
		s = TextUtils.padLeft(s, padLength + s.length(), ' ');
		s = TextUtils.padRight(s, len, ' ');
		return s;
	}

	public static void changeFont(Component c, Font f) {
		f = f.deriveFont(c.getFont().getSize2D());
		c.setFont(f);
	}

	public static void changeFontSize(Component c, int newSize) {
		c.setFont(c.getFont().deriveFont((float) newSize));
	}

	public static <C> String toString(C[] ts) {
		return Arrays.asList(ts).toString();
	}

	public static String fixNewLines(String string) {
		string = string.replace("\r\n", "\n");
		string = string.replace("\r", "\n");
		string = string.replace("\n", System.getProperty("line.separator", "\n"));
		return string;
	}

	public static String normalizeLineTerminators(String text) {
		return text.replace("\r\n", "\n").replace("\r", "\n");
	}

	public static String resolveEnvVars(String input) {
		if (null == input) {
			return null;
		}
		Pattern p = Pattern.compile("\\$\\{(\\w+)\\}|\\$(\\w+)");
		Matcher m = p.matcher(input);
		StringBuffer sb = new StringBuffer();
		while (m.find()) {
			String envVarName = null == m.group(1) ? m.group(2) : m.group(1);
			String envVarValue = System.getenv(envVarName);
			envVarValue = Matcher.quoteReplacement(envVarValue);
			m.appendReplacement(sb, null == envVarValue ? "" : envVarValue);
		}
		m.appendTail(sb);
		return sb.toString();
	}

	public static String extractCapturedGroup(String string, String regex, int capturedGroup) {
		Pattern p = Pattern.compile(regex);
		Matcher m = p.matcher(string);
		if (!m.find()) {
			return null;
		}
		return m.group(capturedGroup);
	}

	public static String escapeJSON(String input) {
		StringBuilder output = new StringBuilder();

		for (int i = 0; i < input.length(); i++) {
			char ch = input.charAt(i);
			int chx = (int) ch;

			// let's not put any nulls in our strings
			assert (chx != 0);

			if (ch == '\n') {
				output.append("\\n");
			} else if (ch == '\t') {
				output.append("\\t");
			} else if (ch == '\r') {
				output.append("\\r");
			} else if (ch == '\\') {
				output.append("\\\\");
			} else if (ch == '"') {
				output.append("\\\"");
			} else if (ch == '\b') {
				output.append("\\b");
			} else if (ch == '\f') {
				output.append("\\f");
			} else if (chx >= 0x10000) {
				assert false : "Java stores as u16, so it should never give us a character that's bigger than 2 bytes. It literally can't.";
			} else if (chx > 127) {
				output.append(String.format("\\u%04x", chx));
			} else {
				output.append(ch);
			}
		}

		return output.toString();
	}

	public static String unescapeJSON(String input) {
		StringBuilder builder = new StringBuilder();

		int i = 0;
		while (i < input.length()) {
			char delimiter = input.charAt(i);
			i++; // consume letter or backslash

			if (delimiter == '\\' && i < input.length()) {

				// consume first after backslash
				char ch = input.charAt(i);
				i++;

				if (ch == '\\' || ch == '/' || ch == '"' || ch == '\'') {
					builder.append(ch);
				} else if (ch == 'n')
					builder.append('\n');
				else if (ch == 'r')
					builder.append('\r');
				else if (ch == 't')
					builder.append('\t');
				else if (ch == 'b')
					builder.append('\b');
				else if (ch == 'f')
					builder.append('\f');
				else if (ch == 'u') {

					StringBuilder hex = new StringBuilder();

					// expect 4 digits
					if (i + 4 > input.length()) {
						throw new RuntimeException("Not enough unicode digits! ");
					}
					for (char x : input.substring(i, i + 4).toCharArray()) {
						if (!Character.isLetterOrDigit(x)) {
							throw new RuntimeException("Bad character in unicode escape.");
						}
						hex.append(Character.toLowerCase(x));
					}
					i += 4; // consume those four digits.

					int code = Integer.parseInt(hex.toString(), 16);
					builder.append((char) code);
				} else {
					throw new RuntimeException("Illegal escape sequence: \\" + ch);
				}
			} else { // it's not a backslash, or it's the last character.
				builder.append(delimiter);
			}
		}

		return builder.toString();
	}

	public static Font loadFont(InputStream in) {
		try {
			Font result = Font.createFont(Font.TRUETYPE_FONT, in);
			return result;
		} catch (Exception e) {
			throw new UnexpectedError(e);
		} finally {
			if (in != null) {
				try {
					in.close();
				} catch (IOException ignore) {
				}
			}
		}
	}

	public static String getPrettyErrorMessage(Throwable t) {
		String result = t.getMessage();
		if (result == null) {
			result = "Error";
		}
		if (t.getCause() != null) {
			String causeClassName = t.getCause().getClass().getName();
			if (result.contains(causeClassName)) {
				String causeClassCaption;
				if (NullPointerException.class.getName().equals(causeClassName)) {
					causeClassCaption = "Missing Value Error";
				} else {
					causeClassCaption = identifierToCaption(t.getCause().getClass().getSimpleName());
				}
				if (!causeClassCaption.contains("Error") && causeClassCaption.contains("Exception")) {
					causeClassCaption = causeClassCaption.replace("Exception", "Error");
				}
				result = result.replace(causeClassName, causeClassCaption);
			}
		}
		return result;
	}

	public static String identifierToCaption(String id) {
		StringBuilder result = new StringBuilder();
		int i = 0;
		char lastC = 0;
		for (char c : id.toCharArray()) {
			if (i == 0) {
				result.append(Character.toUpperCase(c));
			} else if (Character.isUpperCase(c) && !Character.isUpperCase(lastC)) {
				result.append(" " + c);
			} else if (Character.isDigit(c) && !Character.isDigit(lastC)) {
				result.append(" " + c);
			} else if (!Character.isLetterOrDigit(c) && Character.isLetterOrDigit(lastC)) {
				result.append(" " + c);
			} else {
				result.append(c);
			}
			lastC = c;
			i++;
		}
		return result.toString();
	}
}
