package com.otk.application.window.component.crud;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JColorChooser;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSlider;
import javax.swing.JSpinner;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.SpinnerNumberModel;
import javax.swing.SwingConstants;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;

import com.otk.application.error.UnexpectedError;
import com.otk.application.translation.T;
import com.otk.application.util.FileUtils;
import com.otk.application.util.ImageUtils;
import com.otk.application.util.MiscUtils;
import com.otk.application.window.component.JSlinner;

public abstract class AbstractCrudItem implements Cloneable {

	protected JPanel fieldsPanel;
	protected boolean fieldListenerEnabled = true;
	protected Map<String, Object> fieldRealTimeValueByName = new HashMap<String, Object>();
	protected Map<String, List<Component>> fieldControlsByName = new HashMap<String, List<Component>>();
	protected JButton applyButton;
	protected ICrudItemControlListener listener;

	protected abstract void createFieldControls(JPanel panel, ICrudItemControlListener listener);

	protected String[] getFieldNames() {
		List<String> result = new ArrayList<String>();
		for (Field field : getClass().getFields()) {
			if (Modifier.isStatic(field.getModifiers())) {
				continue;
			}
			result.add(field.getName());
		}
		return result.toArray(new String[result.size()]);
	}

	@Override
	public AbstractCrudItem clone() {
		try {
			return (AbstractCrudItem) super.clone();
		} catch (CloneNotSupportedException e) {
			throw new UnexpectedError(e);
		}
	}

	public boolean isModifying() {
		if (applyButton == null) {
			return false;
		}
		return applyButton.isEnabled();
	}

	protected void applyChanges() {
		for (String fieldName : fieldRealTimeValueByName.keySet()) {
			if (!Arrays.asList(getFieldNames()).contains(fieldName)) {
				continue;
			}
			Object filedValue = fieldRealTimeValueByName.get(fieldName);
			setFieldValue(fieldName, filedValue);
		}
	}

	protected void setFieldValue(String fieldName, Object filedValue) {
		MiscUtils.setFieldValue(this, fieldName, filedValue);
	}

	public void withRealTimeValues(Runnable runnable) {
		Map<String, Object> fieldValueByNameToRestore = new HashMap<String, Object>();
		for (String fieldName : fieldRealTimeValueByName.keySet()) {
			if (!Arrays.asList(getFieldNames()).contains(fieldName)) {
				continue;
			}
			fieldValueByNameToRestore.put(fieldName, getFieldValue(fieldName));
			Object filedValue = fieldRealTimeValueByName.get(fieldName);
			setFieldValue(fieldName, filedValue);
		}
		runnable.run();
		for (String fieldName : fieldRealTimeValueByName.keySet()) {
			if (!Arrays.asList(getFieldNames()).contains(fieldName)) {
				continue;
			}
			Object filedValue = fieldValueByNameToRestore.get(fieldName);
			setFieldValue(fieldName, filedValue);
		}
	}

	public String writeFields() {
		return MiscUtils.saveFieldValuesLine(getFieldNames(), this);
	}

	public void readFields(String line) {
		MiscUtils.loadFieldValuesLine(line, getFieldNames(), this);
	}

	@Override
	public String toString() {
		return getClass().getName() + "[" + writeFields() + "]";
	}

	protected Object getFieldValue(String name) {
		return MiscUtils.getFieldValue(this, name);
	}

	protected JSpinner createSpinnerFieldControl(JPanel fieldsPanel, final String name,
			SpinnerNumberModel spinnerNumberModel, String caption) {
		final JSpinner spinner = new JSpinner(spinnerNumberModel);
		labelFieldControl(spinner, caption);
		fieldsPanel.add(spinner);
		ChangeListener actionListener = new ChangeListener() {
			public void stateChanged(ChangeEvent e) {
				onFieldControlStateChange(name, spinner.getValue());
			}
		};
		spinner.addChangeListener(actionListener);
		spinner.setValue(getFieldValue(name));
		actionListener.stateChanged(null);
		fieldControlsByName.put(name, Arrays.asList(spinner));
		return spinner;
	}

	protected JSlider createSliderFieldControl(JPanel fieldsPanel, final String name, int minimum, int maximum,
			final String caption) {
		final JSlider slider = new JSlider();
		labelFieldControl(slider, caption);
		slider.setMaximum(maximum);
		slider.setMinimum(minimum);
		slider.setPaintTicks(true);
		slider.setPaintLabels(true);
		slider.setMajorTickSpacing(10);
		slider.setMinorTickSpacing(1);
		fieldsPanel.add(slider);
		ChangeListener actionListener = new ChangeListener() {
			public void stateChanged(ChangeEvent e) {
				labelFieldControl(slider, caption + ": " + slider.getValue());
				onFieldControlStateChange(name, slider.getValue());
			}
		};
		slider.addChangeListener(actionListener);
		slider.setValue((Integer) getFieldValue(name));
		actionListener.stateChanged(null);
		fieldControlsByName.put(name, Arrays.asList(slider));
		return slider;
	}

	protected JSlinner createSlinnerFieldControl(JPanel fieldsPanel, final String name, int minimum, int maximum,
			final String caption) {
		return createSlinnerFieldControl(fieldsPanel, name, new SpinnerNumberModel(minimum, minimum, maximum, 1),
				caption);
	}

	protected JSlinner createSlinnerFieldControl(JPanel fieldsPanel, final String name,
			SpinnerNumberModel spinnerNumberModel, String caption) {
		final JSlinner slinner = new JSlinner(spinnerNumberModel) {

			private static final long serialVersionUID = 1L;

			@Override
			protected JSpinner createSpinner() {
				JSpinner result = super.createSpinner();
				result.setPreferredSize(new Dimension(MiscUtils.getStandardCharacterWidth(result) * 9,
						result.getPreferredSize().height));
				return result;
			}

		};
		labelFieldControl(slinner, caption);
		slinner.setPaintTicks(true);
		fieldsPanel.add(slinner);
		ChangeListener actionListener = new ChangeListener() {
			public void stateChanged(ChangeEvent e) {
				onFieldControlStateChange(name, slinner.getValue());
			}
		};
		slinner.addChangeListener(actionListener);
		slinner.setValue(getFieldValue(name));
		actionListener.stateChanged(null);
		fieldControlsByName.put(name, Arrays.asList(slinner));
		return slinner;
	}

	protected JButton createColorPickerFieldControl(final JPanel fieldsPanel, final String name, String caption) {
		fieldControlsByName.put(name, new ArrayList<Component>());
		final JButton button = new JButton(" " + caption + "... ") {
			private static final long serialVersionUID = 1L;

			@Override
			public void setBackground(Color bg) {
				super.setBackground(bg);
				setForeground(ImageUtils.getDistantColor(bg));
				setBorder(BorderFactory.createTitledBorder(""));
			}
		};
		JLabel separator1 = new JLabel(" ");
		fieldsPanel.add(separator1);
		fieldControlsByName.get(name).add(separator1);
		fieldsPanel.add(button);
		fieldControlsByName.get(name).add(button);
		JLabel separator2 = new JLabel(" ");
		fieldsPanel.add(separator2);
		fieldControlsByName.get(name).add(separator2);
		final ChangeListener actionListener = new ChangeListener() {
			public void stateChanged(ChangeEvent e) {
				onFieldControlStateChange(name, button.getBackground());
			}
		};
		button.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				Color newColor = JColorChooser.showDialog(fieldsPanel, T.t("Choose a color"), button.getBackground());
				if (newColor != null) {
					MiscUtils.setCrossLookAndFeelBackground(button, newColor);
					actionListener.stateChanged(null);
				}
			}
		});
		MiscUtils.setCrossLookAndFeelBackground(button, (Color) getFieldValue(name));
		actionListener.stateChanged(null);
		return button;
	}

	protected JButton createFilePathFieldControl(final JPanel fieldsPanel, final String name, final String[] extensions,
			final File directory, String caption) {
		fieldControlsByName.put(name, new ArrayList<Component>());
		fieldControlsByName.get(name).add(addFieldLabel(caption, fieldsPanel));
		final String EMPTY = "                        ...                        ";
		final JButton button = new JButton(EMPTY);
		button.setMaximumSize(new Dimension(Integer.MAX_VALUE, button.getMaximumSize().height));
		fieldsPanel.add(button);
		fieldControlsByName.get(name).add(button);
		final ChangeListener actionListener = new ChangeListener() {
			public void stateChanged(ChangeEvent e) {
				String value = button.getText();
				if (value.equals(EMPTY)) {
					value = "";
				}
				onFieldControlStateChange(name, value);
			}
		};
		button.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				String currentFilePath = button.getText();
				if (EMPTY.equals(currentFilePath)) {
					currentFilePath = "";
				}
				currentFilePath = expandFilePath(currentFilePath);
				if (new File(currentFilePath).getParent() == null) {
					if (directory != null) {
						currentFilePath = new File(directory, currentFilePath).getPath();
					}
				}
				File currentFile = (currentFilePath.length() == 0) ? null : new File(currentFilePath);
				String filePath = FileUtils.askFilePath(fieldsPanel, T.t("Indicate"), extensions, currentFile);
				if (filePath != null) {
					filePath = abstractFilePath(filePath);
					button.setText(filePath);
					actionListener.stateChanged(null);
				}
			}
		});
		button.setText((String) getFieldValue(name));
		if (button.getText().length() == 0) {
			button.setText(EMPTY);
		}
		actionListener.stateChanged(null);
		return button;
	}

	protected String abstractFilePath(String filePath) {
		return filePath;
	}

	protected String expandFilePath(String filePath) {
		return filePath;
	}

	protected void labelFieldControl(JComponent c, String name) {
		c.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), name));
	}

	protected JCheckBox createCheckBoxFieldControl(JPanel fieldsPanel, final String name, String caption) {
		fieldControlsByName.put(name, new ArrayList<Component>());
		final JCheckBox checkBox = new JCheckBox(caption);
		ActionListener actionListener = new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				onFieldControlStateChange(name, checkBox.isSelected());
			}
		};
		checkBox.addActionListener(actionListener);
		fieldsPanel.add(checkBox);
		fieldControlsByName.get(name).add(checkBox);
		checkBox.setSelected((Boolean) getFieldValue(name));
		actionListener.actionPerformed(null);
		return checkBox;
	}

	protected JLabel createLabel(String string) {
		JLabel result = new JLabel(string);
		result.setAlignmentX(JButton.CENTER_ALIGNMENT);
		return result;
	}

	protected JLabel addFieldLabel(String caption, JPanel fieldsPanel) {
		JLabel label = createLabel(caption + ":");
		label.setHorizontalTextPosition(SwingConstants.LEFT);
		fieldsPanel.add(label);
		lockSize(label, false, true);
		return label;
	}

	protected void lockSize(Component c, boolean lockWidth, boolean lockHeight) {
		Dimension size = c.getPreferredSize();
		if (!lockWidth) {
			size.width = Integer.MAX_VALUE;
		}
		if (!lockHeight) {
			size.height = Integer.MAX_VALUE;
		}
		c.setMaximumSize(size);
	}

	protected JTextField createTextFieldControl(JPanel fieldsPanel, final String name, String caption) {
		fieldControlsByName.put(name, new ArrayList<Component>());
		fieldControlsByName.get(name).add(addFieldLabel(caption, fieldsPanel));
		final JTextField field = new JTextField();
		fieldsPanel.add(field);
		fieldControlsByName.get(name).add(field);
		lockSize(field, false, true);
		DocumentListener listener = new DocumentListener() {
			@Override
			public void removeUpdate(DocumentEvent e) {
				stateChanged();
			}

			@Override
			public void insertUpdate(DocumentEvent e) {
				stateChanged();
			}

			@Override
			public void changedUpdate(DocumentEvent e) {
				stateChanged();
			}

			private void stateChanged() {
				onFieldControlStateChange(name, field.getText());
			}
		};
		field.getDocument().addDocumentListener(listener);
		field.setText((String) getFieldValue(name));
		listener.changedUpdate(null);
		return field;
	}

	protected JTextArea createMultilineTextFieldControl(JPanel fieldsPanel, final String name, String caption) {
		fieldControlsByName.put(name, new ArrayList<Component>());
		fieldControlsByName.get(name).add(addFieldLabel(caption, fieldsPanel));
		final JTextArea field = new JTextArea();
		fieldsPanel.add(field);
		fieldControlsByName.get(name).add(field);
		DocumentListener listener = new DocumentListener() {
			@Override
			public void removeUpdate(DocumentEvent e) {
				stateChanged();
			}

			@Override
			public void insertUpdate(DocumentEvent e) {
				stateChanged();
			}

			@Override
			public void changedUpdate(DocumentEvent e) {
				stateChanged();
			}

			private void stateChanged() {
				onFieldControlStateChange(name, field.getText());
			}
		};
		field.getDocument().addDocumentListener(listener);
		field.setText((String) getFieldValue(name));
		listener.changedUpdate(null);
		return field;
	}

	protected JComboBox<Object> createChoiceFieldControl(JPanel fieldsPanel, final String name, final Object[] choices,
			String[] choiceCaptions, String caption) {
		fieldControlsByName.put(name, new ArrayList<Component>());
		fieldControlsByName.get(name).add(addFieldLabel(caption, fieldsPanel));
		final JComboBox<Object> comboBox = new JComboBox<Object>(choiceCaptions);
		fieldsPanel.add(comboBox);
		fieldControlsByName.get(name).add(comboBox);
		ActionListener actionListener = new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				Object value = choices[comboBox.getSelectedIndex()];
				onFieldControlStateChange(name, value);
			}
		};
		comboBox.addActionListener(actionListener);
		Object currentvalue = getFieldValue(name);
		int currentvalueIndex = Arrays.asList(choices).indexOf(currentvalue);
		comboBox.setSelectedIndex(currentvalueIndex);
		actionListener.actionPerformed(null);
		return comboBox;
	}

	protected void onApplyRequested(ICrudItemControlListener listener) {
		listener.beforeItemUpdate(AbstractCrudItem.this);
		applyChanges();
		applyButton.setEnabled(false);
		applyButton.setBorder(new JButton().getBorder());
		listener.itemUpdated(AbstractCrudItem.this);
	}

	protected void onFieldControlStateChange(String fieldName, Object value) {
		fieldRealTimeValueByName.put(fieldName, value);
		if (fieldListenerEnabled) {
			applyButton.setEnabled(true);
			applyButton.setBorder(BorderFactory.createCompoundBorder(BorderFactory.createLineBorder(Color.RED, 3),
					new JButton().getBorder()));
		}
	}

	public Component createComponent(final ICrudItemControlListener listener) {
		this.listener = listener;

		fieldListenerEnabled = false;

		fieldsPanel = new JPanel();
		fieldsPanel.setLayout(new BoxLayout(fieldsPanel, BoxLayout.Y_AXIS));

		applyButton = new JButton(T.t("Apply"));
		fieldsPanel.add(applyButton);
		applyButton.setEnabled(false);
		applyButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				onApplyRequested(listener);
			}
		});

		JPanel customFieldsPanel = new JPanel();
		fieldsPanel.add(new JScrollPane(customFieldsPanel));
		customFieldsPanel.setLayout(new BoxLayout(customFieldsPanel, BoxLayout.Y_AXIS));
		customFieldsPanel.setBorder(BorderFactory.createCompoundBorder(BorderFactory.createLoweredBevelBorder(),
				BorderFactory.createEmptyBorder(10, 10, 10, 10)));

		createFieldControls(customFieldsPanel, listener);
		fieldListenerEnabled = true;

		return fieldsPanel;
	}

}
