package com.otk.application.window.component.crud;

import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.DefaultListModel;
import javax.swing.JList;

import com.otk.application.error.StandardError;
import com.otk.application.error.UnexpectedError;
import com.otk.application.util.TextUtils;

public abstract class AbstractCrudModel {

	private DefaultListModel<Object> listModel = new DefaultListModel<Object>();
	private List<IListener> listeners = new ArrayList<IListener>();
	private IListener notifier = new IListener() {

		@Override
		public void added(int index) {
			for (IListener l : listeners) {
				l.added(index);
			}
		}

		@Override
		public void batchChange() {
			for (IListener l : listeners) {
				l.batchChange();
			}
		}

		@Override
		public void cleared() {
			for (IListener l : listeners) {
				l.cleared();
			}
		}

		@Override
		public void deleted(int index) {
			for (IListener l : listeners) {
				l.deleted(index);
			}
		}

		@Override
		public void duplicated(int i) {
			for (IListener l : listeners) {
				l.duplicated(i);
			}
		}

		@Override
		public void movedDown(int i) {
			for (IListener l : listeners) {
				l.movedDown(i);
			}
		}

		@Override
		public void movedUp(int i) {
			for (IListener l : listeners) {
				l.movedUp(i);
			}
		}

		@Override
		public void updated(int index) {
			for (IListener l : listeners) {
				l.updated(index);
			}
		}

	};

	public List<IListener> getListeners() {
		return listeners;
	}

	public List<AbstractCrudItem> getList() {
		List<AbstractCrudItem> list = new ArrayList<AbstractCrudItem>();
		for (int i = 0; i < listModel.getSize(); i++) {
			list.add((AbstractCrudItem) listModel.get(i));
		}
		return list;
	}

	@SuppressWarnings("unchecked")
	public Class<? extends AbstractCrudItem> getItemType(String typeName) {
		try {
			return (Class<? extends AbstractCrudItem>) Class.forName(typeName.replace('-', '$'));
		} catch (ClassNotFoundException e) {
			throw new StandardError(e.toString(), e);
		}
	}

	public String getItemName(Class<? extends AbstractCrudItem> c) {
		return c.getName().replace('$', '-');
	}

	public String getItemTypeCaption(Class<? extends AbstractCrudItem> c) {
		return getItemName(c);
	}

	protected abstract Class<? extends AbstractCrudItem>[] getItemTypes();

	public void load(String s) {
		load(s, false, 0);
	}

	public void load(String s, boolean append, int index) {
		List<AbstractCrudItem> items = new ArrayList<AbstractCrudItem>();
		for (String line : TextUtils.splitLines(s)) {
			if (line.length() > 0) {
				items.add(deserialize(line));
			}
		}
		if (!append) {
			listModel.clear();
		}
		if (index == -1) {
			index = listModel.getSize();
		}
		for (AbstractCrudItem item : items) {
			listModel.add(index, item);
			index++;
		}
		notifier.batchChange();
	}

	public String save() {
		List<AbstractCrudItem> list = getList();
		StringBuilder result = new StringBuilder();
		for (int i = 0; i < list.size(); i++) {
			AbstractCrudItem f = list.get(i);
			if (i > 0) {
				result.append("\n");
			}
			result.append(serialize(f));
		}
		return result.toString();
	}

	public void setList(List<AbstractCrudItem> items) {
		listModel.clear();
		for (AbstractCrudItem item : items) {
			listModel.addElement(item);
		}
		notifier.batchChange();
	}

	public String serialize(AbstractCrudItem item) {
		String s = "[" + getItemName(item.getClass()) + "]: " + item.writeFields();
		return s;
	}

	public AbstractCrudItem deserialize(String line) {
		StringBuilder patternBuilder = new StringBuilder();
		patternBuilder.append("\\s*\\[(");
		int iType = 0;
		for (Class<? extends AbstractCrudItem> t : getItemTypes()) {
			if (iType > 0) {
				patternBuilder.append("|");
			}
			patternBuilder.append(getItemName(t));
			iType++;
		}
		patternBuilder.append(")\\]:\\s*(.+)");
		Pattern pattern = Pattern.compile(patternBuilder.toString());
		Matcher matcher = pattern.matcher(line);
		if (!matcher.matches()) {
			throw new StandardError("Invalid " + getClass().getSimpleName() + " line format: '" + line
					+ "'. Does not match '" + patternBuilder.toString() + "'");
		}
		String typeName = matcher.group(1);
		String valuesString = matcher.group(2);
		Class<? extends AbstractCrudItem> c = getItemType(typeName);
		AbstractCrudItem item = createItem(c);
		item.readFields(valuesString);
		return item;
	}

	protected <T extends AbstractCrudItem> T createItem(Class<T> itemType) {
		try {
			return itemType.newInstance();
		} catch (Exception e) {
			throw new UnexpectedError(e);
		}
	}

	static interface IListener {
		void added(int index);

		void batchChange();

		void cleared();

		void deleted(int index);

		void duplicated(int i);

		void movedDown(int i);

		void movedUp(int i);

		void updated(int index);
	}

	public void add(AbstractCrudItem newItem) {
		add(-1, newItem);
	}

	public void add(int index, AbstractCrudItem newItem) {
		if (index == -1) {
			index = listModel.getSize();
		}
		listModel.add(index, newItem);
		notifier.added(index);
	}

	public void clear() {
		listModel.clear();
		notifier.cleared();
	}

	public void delete(int index) {
		listModel.remove(index);
		notifier.deleted(index);
	}

	public void duplicate(int index) {
		AbstractCrudItem sourceItem = (AbstractCrudItem) listModel.get(index);
		AbstractCrudItem newItem = (AbstractCrudItem) sourceItem.clone();
		listModel.add(index + 1, newItem);
		notifier.duplicated(index + 1);
	}

	public void moveDown(int index) {
		int moveFrom = index;
		Object toMove = listModel.get(moveFrom);
		listModel.remove(moveFrom);
		listModel.add(moveFrom + 1, toMove);
		notifier.movedDown(moveFrom + 1);
	}

	public void moveUp(int index) {
		int moveFrom = index;
		Object toMove = listModel.get(moveFrom);
		listModel.remove(moveFrom);
		listModel.add(moveFrom - 1, toMove);
		notifier.movedUp(moveFrom - 1);
	}

	public void set(int index, AbstractCrudItem item) {
		listModel.set(index, item);
		notifier.updated(index);
	}

	public void notifyUpdate(AbstractCrudItem item) {
		set(getList().indexOf(item), item);
	}

	public void conectTo(JList<Object> listControl) {
		listControl.setModel(listModel);
	}

}
